#!/usr/bin/env perl
# Description: Analyse network traffic (dump file).
#
# Copyright:
# Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

# mapwocanalyse.pl - Analayse network traffic from dump file.
# Usage:
#   mapwocanalyse -p <dump_file> -a <antivir cfg> -b [working_directory] -j [json_file]
# working_directory defaults to cwd.

# Result:
# Creates working_directory/Log and working_directory/Bin.

use strict;
use warnings;

BEGIN {

    $ENV{ LC_ALL } = "en_US.UTF-8";
}

use Cwd;
use File::Spec;
use File::Basename;
use Getopt::Std;
use Config::General qw(ParseConfig);
use Proc::Killall;

use MapWOCAnalyser::MapWOCAnalyser qw( @filename ); # import @filename from MapWOCAnalyser.pm
use Data::Dumper;

use constant { 
    TRUE  => 1,
    FALSE => 0
};      

my %opts;

getopt('b:p:j:a:', \%opts); # get arguments

my $BASE_DIR = cwd; # Set BASE_DIR to current dir.

if ( defined $opts{ 'b' } ) { # Use another BASE_DIR?

    if( ! -d $opts{ 'b' } ) {
        if( ! mkdir ("$opts{ 'b' }", 0777) ) {
            print "ERROR: Could not create base-dir '" . $opts{ 'b' } . "'Exit now!\n";
            exit 252;
        }
    }

    $BASE_DIR = $opts{ 'b' };
}

if ( ! chdir( $BASE_DIR ) ) {
    print "ERROR: Could not change into dir $BASE_DIR\n";
    exit 254;
}

if( ! defined $opts{ 'p' } || ! defined $opts{ 'a' } ) {
    print "Usage: mapwocanalyse -p <dump_file> -a <antivir_cfg> -b [working_directory] -j [json_file]\n";
    exit 255;
}

if( ! File::Spec->file_name_is_absolute( $opts{ 'p' } ) ) { # path to pcap file is absolut?
    print "ERROR: Path to dump file must be absolute.\n";
    exit 243;
}

my $err;

my $LOG_DIR    = $BASE_DIR . '/Log';
my $BINARY_DIR = $BASE_DIR . '/Binary';
my $xmlfile    = $LOG_DIR  . '/Analyse.xml';

if( ! -d $LOG_DIR ) { # Create dir for logs if it does not exist.
    if( ! mkdir ("$LOG_DIR", 0777) ) {
        print "ERROR: Could nor create logdir $LOG_DIR. Exit now!\n";
        exit 255;
    }
}

if( ! -d $BINARY_DIR ) { # Create dir for binarys if it does not exist.
    if( ! mkdir ("$BINARY_DIR", 0777) ) {
        print "ERROR: Could nor create binarydir $BINARY_DIR. Exit now!\n";
        exit 255;
    }
}

my $pcapfile = $opts{ 'p' };

if ( ! -e $pcapfile ) {
    print "ERROR: PCAP File $pcapfile not found. Exit now!\n";
    exit 255;
}

my $antivir_cfg = $opts{ 'a' };

if ( ! -e $antivir_cfg ) {
    print "ERROR: Antivir Configuration $antivir_cfg not found. Exit now!\n";
    exit 255;
}

$/="\n";
my $avconf = new Config::General( -ConfigFile =>  $antivir_cfg, -CComments => 0 ); # read antivir.cfg
my %antivir_config = $avconf->getall; # get values from antivir.cfg
$/ = undef;

my $json_file = undef;

if ( defined ( $opts{ 'j' } ) ) { # json file

    $json_file = $opts{ 'j' };

    if ( ! -e $json_file ) {
        print "ERROR: JSON File $json_file not found. Exit now!\n";
        exit 255;
    }

}

# Create dump dir from pcap filename.
my @path = split( /\//, $pcapfile );
my @work = split( /\./, $path[( scalar @path) -1] );
$work[ ( scalar @work ) -2 ] =~ s/:/_/mg;

my $work_dir = $work[( scalar @work ) -2 ]; # workdir based on pcap file name

if( ! -d $work_dir ) {
    if( ! mkdir ("$work_dir", 0777) ) {
        print "ERROR: Could nor create workdir $work_dir. Exit now!\n";
        exit 255;
    }
}

#########################################################################

# create a new analyser object.

my $analyser = MapWOCAnalyser::MapWOCAnalyser->_new(
    BINARY_DIR     => $BINARY_DIR,
    pcapfile       => $pcapfile,
    work_dir       => $work_dir,
    LOG_DIR        => $LOG_DIR,
    xmlfile        => $xmlfile,
    json_file      => $json_file,
    antivir_config => \%antivir_config
  );

{

    $/ = undef;
    $\ = "";
    $| = 0;

}

$analyser->{ 'XML' }->_start_xml( basename ( $pcapfile ) );

foreach my $name ( @filename ) { # @filename is defined in callback function 'syn_packets()' in MapWOCAnalyser.pm

    print "INFO: Evaluating $BASE_DIR/$name\n";

    if ( ! $analyser->_read_traffic( $BASE_DIR . "/" . $name ) ) { # read traffic

        print "ERROR: read_traffic() for  $BASE_DIR . "/" . $name failed. Try to continue...\n";
        next;

    }
}

if( defined ( $json_file ) ) {
    $analyser->_read_json( $json_file );
}

my @antivir_tmp = ( $BINARY_DIR, $BASE_DIR . "/" . $work_dir ); # Check both. Binary and Dump dir...

my ( $scanned_file, $antivir, $error_code );
my $antivir_timeout = 300; # max runtime for antivirus scanner in seconds (just to be on the safe side...)

foreach my $dir ( @antivir_tmp )  {

    while ( my ( $avscanner, $avoptions ) = each ( %{ $antivir_config{ 'Antivir' } } ) ) {

        my @filesFound = glob( $dir . "/*" );
        my $file_count = @filesFound;

        if ( $file_count > 0 ) { # no need to run the scanner if there are no files

            my $scanner = $avoptions->{ 'scanner' }; # get options for specific scanner

            if ( -e $scanner && -x $scanner ) { # check if read and exec of scanner possible

                print "Scanning files in dir $dir for Viruses. Scanner is: $scanner\n";

                if( $avoptions->{ 'use_basename' } > 0 ) { 
                    $scanner = basename( $scanner ); # some scanners (e.g. avg) wants to be called without the full path.
                }

                eval { # for timeout check...

                    local $SIG{ ALRM } = sub { # what to do if scanner reached timeout

                        $analyser->{ 'XML' }->_add_antivir_result_to_xml( $avoptions->{'name' }, undef, "scanner error (timeout)" );
                        killall('KILL', basename( $scanner )); # kill scanner if timeout reached
                        { next; } #  next scanner

                      };

                    alarm $antivir_timeout; # set timeout for scanner
                    $antivir     = qx/$scanner $avoptions->{ 'options' } $dir/; # execute antivirus scanner
                    $error_code  = $? >> 8; # get the exit code from scanner
                    alarm 0; # unset timeout

                  };

                if( &_check_antivir_error_code( $error_code, $avoptions->{ 'error_codes' } ) ) { # check if scanner exit with error
                    $analyser->{ 'XML' }->_add_antivir_result_to_xml( $avoptions->{ 'name' }, undef, "scanner error ($error_code)" );
                    next;
                }

                my @antivir_out = split( /\n/, $antivir ); # split antivir output into lines

                foreach my $av_line ( @antivir_out ) {

                    $av_line =~ s/[\r\n]+//g; # remove newlines
                    $av_line =~ s/[[:cntrl:]]//g; # remove ascii contr. chars

                    if( $av_line =~ m/$avoptions->{ 'regex' }/ ) { # if line matches the regex from antivir.cfg

                        if( eval( $avoptions->{ 'result_code' } ) && ( defined( $scanned_file ) && defined ( $av_line ) ) ) { # eval code from antivir.cfg to get $scanned_file and $av_line

                            $analyser->{ 'XML' }->_add_antivir_result_to_xml( $avoptions->{ 'name' }, $scanned_file, $av_line ); # update antivir placeholders with antivir result

                        } else {

                            $analyser->{ 'XML' }->_add_antivir_result_to_xml( $avoptions->{ 'name' }, $scanned_file, "scanner error (eval error)" ); # there is an error in the evaled code

                        }

                    }

                }

            } else {

                print "Scanner $scanner does not exist. Try to continue.\n";
                next;

            }
        }
    }
}

$analyser->{ 'XML' }->_close_xml;

sub _check_antivir_error_code {

    my ($error_code, $error_codes) = @_;
    my( $cmd, $code);

    my %cmds=(
        '<'  => sub{ $_[0] <  $_[1] },
        '='  => sub{ $_[0] == $_[1] },
        '>'  => sub{ $_[0] >  $_[1] },
        '>=' => sub{ $_[0] >= $_[1] },
        '<=' => sub{ $_[0] <= $_[1] }
      );

    my @codes = split( /,/, $error_codes);

    foreach( @codes ) {

        $_ =~ s/[\s\t]+//g;

        if( $_ =~ m/([><=]+)([0-9]+)/g ) {
            $cmd  = $1;
            $code = $2;
        }

        if( exists( $cmds{ $cmd } ) ) {

            if( $cmds{ $cmd }->( $error_code, $code ) ) {

                return TRUE;

            }
        }

    }

    return FALSE;

}

__END__

=head1 NAME

mapwocanalyse.pl - Perl Script for mapwoc

=head1 DESCRIPTION

Script for running the mapwoc analyser. 

=head1 LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2,
or, at your option, any later version as published by the Free
Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

=head1 AUTHOR

Michael Meyer <michael.meyer@intevation.de>

=head1 COPYRIGHT

Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik

=cut
