# Copyright:
# Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

package MapWOCAnalyser::MapWOCAnalyser;

use strict;
use warnings;

use Exporter;
our @ISA = qw( Exporter );
our @EXPORT_OK = qw( @filename );

use Net::Pcap;
use NetPacket::Ethernet;
use NetPacket::IP qw( IP_PROTO_TCP );
use NetPacket::TCP;

use Digest::MD5 qw( md5 md5_hex );
use IO::Uncompress::Gunzip qw( gunzip $GunzipError );
use XML::Simple;
use File::Basename;
use Digest::SHA1  qw(sha1_hex);

use JSON;
use MIME::Base64;
use MapWOCAnalyser::XML;

use Data::Dumper;

use constant { 
    TRUE  => 1,
    FALSE => 0
};

$/ = undef;
$\ = "";
$| = 0;

our @filename;
our @jsonfiles;

my $interesting_methods = "GET|POST|CONNECT|TRACE|OPTIONS|HEAD"; # these methods are interesting for us

sub _new {

    my ( $this, %args ) = @_;
    my $self = {};

    my $err;

    bless $self, $this;

    $self->{ 'work_dir'    }    = $args{ 'work_dir'       };
    $self->{ 'BINARY_DIR'  }    = $args{ 'BINARY_DIR'     };
    $self->{ 'LOG_DIR'     }    = $args{ 'LOG_DIR'        };
    $self->{ 'pcapfile'    }    = $args{ 'pcapfile'       };
    $self->{ 'xmlfile'     }    = $args{ 'xmlfile'        };
    $self->{ 'json_file'   }    = $args{ 'json_file'      };
    $self->{ 'antivir_config' } = $args{ 'antivir_config' };

    my $XML = MapWOCAnalyser::XML->_new( xmlfile => $self->{ 'xmlfile'} ); # initiate xml

    $self->{ 'XML' } = $XML;

    # Read PCAP-file.
    my $pcap = Net::Pcap::open_offline(  $self->{ 'pcapfile' }, \$err ); # open pcap file

    unless ( defined $pcap ) {
        print 'ERROR: Unable to create packet capture on file ',  $self->{ 'pcapfile' }, ' - ', $err . "\n";
        exit 255;
    }

    # Call syn_packets on all packets.
    Net::Pcap::loop( $pcap, -1, \&_syn_packets, $self ) ; # loop over all pakets in pcap file
    Net::Pcap::close( $pcap ); # close pcap file

    return $self;

}

sub _read_traffic {

    my $self = shift;
    my ( $infile ) = @_;

    my ( $xml, $host, $fname, $browser, $referer, $request_is_redirector_request  );

    if( ! open( OFH, "<" . $infile ) ) { # open traffic files

        print  "ERROR: Could not open $infile for data evaluation.\n";
        return FALSE;

    }

    binmode( OFH );
    seek   ( OFH, 0, 0 );
    my $data = <OFH>;
    close  ( OFH );

    my @data_split = split( /SPLITMWOC/, $data ); # split traffic dumps into separate connections

    foreach my $r ( @data_split ) {

        $request_is_redirector_request = undef;

        if( length( $r ) > 0 ) {

            my ( $request_header, $response_header, $response_body ) = split( /\n\n|\r\n\r\n/, $r ,3); # split traffic

            if( defined ( $response_body ) && $response_body =~ m/MAPWOC_FASTMODE_URL/g) { # fastmode
                $self->{ 'traffic_is_fastmode' } = TRUE;
                $self->_parse_fastmode_urls( $response_body );
            }

            if ( exists ( $self->{ basename ( $infile ) } ) && ( defined ( $request_header ) && defined ( $response_header ) ) ) {

                my ( $return_code, $location ) = $self->_get_http_return_code( $response_header );

                if( $self->{ basename ( $infile ) }{ 'ip' } eq "10.0.0.1" && $return_code eq "302" ) { # check if request is redirector request
                    $request_is_redirector_request = TRUE;
                }

                if( $request_header =~ m/($interesting_methods)\s(http[s]*:\/\/)?[^\/]*\/(.*) HTTP\/1\../ig ) { # extract filename
                    $fname = $3;
                    $fname = $self->_clean_string( $fname );
                }

                if( $request_header =~ m/Referer:\s(http[s]*:\/\/[^\n\r]+)/ig ) { # extract referer
                    $referer = $1;
                    $referer = $self->_clean_string( $referer );
                }

                if( $request_header =~ m/User-Agent:\s([^\n\r]+)/ig ) { # extract browser
                    $browser = substr( $1, 0, 12 );
                    $browser = $self->_clean_string( $browser );
                }

                if( $request_header =~ m/Host:\s([^\n\r]+)/i ) { # extract host
                    $host = $1;
                    $host = $self->_clean_string( $host );
                    if( $host eq "10.0.0.1" && $return_code eq "302" ) { # check if request is redirector request through proxy
                        $request_is_redirector_request = TRUE;
                    }
                }

                if( ! ( defined $browser ) ) { $browser = "NODEF"; }
                if( ! ( defined $fname   ) ) { $fname   = "";      }
                if( ! ( defined $referer ) ) { $referer = "NODEF";      }

                # store some informastion in session
                $self->{ basename ( $infile ) }{ 'host'      } = $host    ? $host    : "NO_HOST";
                $self->{ basename ( $infile ) }{ 'fname'     } = $fname   ? $fname   : "";
                $self->{ basename ( $infile ) }{ 'referer'   } = $referer ? $referer : "NO_REFERER";
                $self->{ basename ( $infile ) }{ 'browser'   } = $browser;

                if( ! defined ( $self->{ 'XML' }->{ 'pcap_start' } ) ) { # for duration

                    $self->{ 'XML' }->{ 'pcap_start' } = $self->{ basename ( $infile ) }{ 'real_time' }; # set start time

                    my $xml = '<time>' . $self->{'XML'}->{ 'pcap_start' } . '</time>';

                    $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                            RootName      => undef,
                            NoAttr        => 0,
                            SuppressEmpty => 0,
                            NoSort        => 1,
                            NoEscape      => 0
                          )
                      );

                } else {

                    $self->{ 'XML' }->{ 'pcap_stop' } = $self->{ basename ( $infile ) }{ 'real_time' };

                }

                if ( defined ( $request_is_redirector_request ) || defined ( $self->{ 'traffic_is_fastmode' } ) ) {

                    if( defined ( $self->{ basename ( $infile ) }{ 'host' } && defined ( $self->{ basename ( $infile ) }{ 'fname' } ) ) ) {

                        my ( $return_code, $location ) = $self->_get_http_return_code( $response_header );
                        my $url = 'http://' . $self->{ basename ( $infile ) }{ 'host' } . '/' .  $self->{ basename ( $infile ) }{ 'fname' };

                        if( ( defined ( $self->{ 'traffic_is_fastmode' } ) && $self->_url_is_fastmode_url( $url ) ) ||
                            ( defined ( $request_is_redirector_request ) && ! defined ( $self->{ 'traffic_is_fastmode' } ) ) )  {

                            $xml = {
                                url         => [ $url ],
                                ip          => [ $self->{ basename ( $infile ) }{ 'ip' } ],
                                browser     => [ $self->{ basename ( $infile ) }{ 'browser' } ],
                                location    => [ $location ]
                              };

                            $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                                    NoAttr        => 0,
                                    RootName      => 'target',
                                    SuppressEmpty => 1,
                                    NoSort        => 1,
                                    NoEscape      => 0
                                  )
                              );

                        }
                    }
                }
            }

            if( defined ( $request_header ) && defined ( $response_header ) ) {

                # Read HTTP Traffic.

                $self->_read_http( $request_header, $response_header, $response_body, $infile );
            }

            else {

                # Read Non HTTP Traffic.

                $self->_read_non_http( $r, $infile );
            }

        }
    }

    return TRUE;

}

sub _syn_packets {

    my ($self, $header, $packet ) = @_;

    my ( $fname, $browser, $host, $referer );

    #    our @filename;

    # Strip ethernet encapsulation of captured packet.

    my $ether_dec  = NetPacket::Ethernet->decode( $packet );
    my $ether_data = NetPacket::Ethernet::strip( $packet );

   # Decode contents of TCP/IP packet contained within captured ethernet packet.

    my $ip  = NetPacket::IP->decode( $ether_data );

    return if( $ip->{ 'proto' } != IP_PROTO_TCP ); # ignore other than TCP (e.g. UDP)

    my $tcp = NetPacket::TCP->decode( $ip->{ 'data' } );

    my $peer_string =  $self->{ 'work_dir' } . "/" .$ip->{ 'src_ip' } . "_" . $tcp->{ 'src_port' } . "_" . $tcp->{ 'dest_port' };

    my $tcpdata = $tcp->{ 'data' };

    # Open File according to string that should represent this communication (it doesnt, as dst-port is not unique).
    if ( ( -e $peer_string ) && ( length ( $tcp->{ 'data' }) > 0 ) ) {

        open (AUS ,">>" . $peer_string);

    } else {

        $peer_string =  $self->{ 'work_dir' } . "/" .$ip->{ 'dest_ip' } . "_" . $tcp->{ 'dest_port' } . "_" . $tcp->{ 'src_port' };
        if ( ( -e $peer_string )  && ( length ( $tcp->{ 'data' }) > 0 ) ) {

            # What does this mean? It is the second packet of this peer combo?

            open ( AUS , ">>" . $peer_string );

        } else {

            if ( length ( $tcp->{'data'} ) >1 ) {

                open ( AUS , ">>" . $peer_string );
                push ( @filename, $peer_string );

            }
        }
    }

    if (length ( $tcpdata ) > 1) {

        my $real_time = $self->_ts2time( $header->{ 'tv_sec' } ); # build time from timestamp

        $self->{ basename ( $peer_string ) }{ 'time'      } = $header->{ 'tv_sec' };
        $self->{ basename ( $peer_string ) }{ 'sequence'  } = $header->{ 'tv_sec' };
        $self->{ basename ( $peer_string ) }{ 'real_time' } = $real_time;

        if (  $tcpdata =~ /^($interesting_methods)/ ) {

            print AUS "SPLITMWOC";
            print AUS "TIME:" . $header->{ 'tv_sec' } . "\n";

            my $h1 = substr( $tcpdata, 0, 500 );

            if( defined $h1 ){

                $h1 =~ s/(GET|POST) \/(.*) HTTP\/1\../$fname = $2;/egi;
                $h1 =~ s/Referer: (http[s]*:\/\/[^\r\n]+)/$referer = $1;/egi;
                $h1 =~ s/Host: ([^\r\n]+)./$host = $1;/egi;
                $h1 =~ s/User-Agent: ([^\r\n]+)/$browser = substr( $1, 0, 12 );/egi;

                if( ! ( defined $browser ) ) { $browser = "NODEF"; }
                if( ! ( defined $fname   ) ) { $fname   = "";}

            }

        }

        print AUS  "$tcpdata"; # write complete traffic dump to file
        close(AUS);

    }

    $self->{ basename ( $peer_string ) }{ 'ip' } = $ip->{ 'dest_ip' };
}

sub _read_http {

    my $self = shift;

    my ($request_header, $response_header, $response_body, $infile) = @_;
    my ($request, $requestfile);

    if ( ( defined $response_body ) && ( $response_body =~ /^\x1f\x8b\x08/ ) ) {
        my $uncompress;
        gunzip \$response_body => \$uncompress; # unzip content if it is gziped
        $response_body = $uncompress;
    }

    if( defined $response_body ) {

        $self->{ basename ( $infile ) }{ 'data' } .= $request_header . "\n" .  $response_body . "\n###split###\n"; # store data and headers in session to act on it later

        my $f_type = $self->_file_type( substr( $response_body, 0, 300 ) ); # get the file type for content

        if( defined ( $f_type ) && $self->_is_interesting_filetype( $f_type ) ) {

            my ( $outfile1, $filename ) = $self->_make_outfile1( $response_header, $request_header, $infile );

            if( ! defined ( $outfile1 ) ) {
                return FALSE;
            }

            my $md   = md5_hex(  $response_body );
            my $sha1 = sha1_hex( $response_body );

            if ( open (DA, ">> $self->{ 'BINARY_DIR' }/" .  $outfile1 . "-" . $md ) ) {

                binmode(DA);
                print DA "$response_body";
                close(DA);

                my $data2 = `file -b $self->{ 'BINARY_DIR' }/$outfile1-$md`;
                $data2 =~ s/[\r\n]+//g;

                my $mime_typ = `file -bi $self->{ 'BINARY_DIR' }/$outfile1-$md`;
                $mime_typ =~ s/[\r\n]+//g;

                my $ssdeep = $self->_ssdeep ( $self->{ 'BINARY_DIR' } . "/" . $outfile1 . "-" . $md );

                ( $request, $requestfile ) = $self->_parse_request_header( $request_header );
                my ($parent_request, $parent_sequence) = $self->_parent_request( $requestfile, $infile, @filename );

                my $file_type = $self->_get_filetypestring( $f_type );

                my ( $x, $z );

                my $xml = {

                    file => {

                        source          => [ $request ],
                        filename        => [ $requestfile ],
                        extractname     => [ $self->_basepath( $self->{ 'BINARY_DIR' } . "/" .  $outfile1 . "-" . $md ) ],
                        parent_request  => [ $parent_request ],
                        parent_sequence => [ $parent_sequence ],
                        analysis        => {

                          }
                      },

                    dumpfile => {

                        file     =>  [ $self->_basepath( $infile ) ],
                        analysis => {

                          }

                      }

                  };

                $xml->{ 'file' }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "mapwocanalyse.pl/type_classifier",
                    output => [ $file_type ]
                  };

                $xml->{ 'file' }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:file -b",
                    output => [ $data2 ]
                  };

                $xml->{ 'file' }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:file -bi",
                    output => [ $mime_typ ]
                  };

                $xml->{ 'file' }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:sha1sum",
                    output => [ $sha1 ]
                  };

                $xml->{ 'file' }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:ssdeep",
                    output => [ $ssdeep ]
                  };

                while ( my ( undef, $avoptions ) = each (%{ $self->{ 'antivir_config' }{ 'Antivir' } } ) ) {

                    $xml->{ 'file' }{ 'analysis' }{ 'method' }[ $x++ ] = {
                        name   =>  "tool:" . $avoptions->{ 'name' },
                        output => [ "###" . $avoptions->{ 'name' } .  "###" . $outfile1 . "-" . $md . "###" ]
                      };

                    $xml->{ 'dumpfile' }{ 'analysis' }{ 'method' }[ $z++ ] = {
                        name   =>  "tool:" . $avoptions->{ 'name' },
                        output => [ "###" . $avoptions->{ 'name' } . "###" . basename ( $infile ) . "###" ]
                      };
                }

                $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                        RootName      => "files",
                        NoAttr        => 0,
                        SuppressEmpty => 0,
                        NoSort        => 1,
                        NoEscape      => 0
                      )
                  );

            }

        }
    }

    my $x;

    if( defined ( $response_header ) && defined ( $request_header ) ) {

        my ( $return_code, $loc ) = $self->_get_http_return_code( $response_header );

        my $xml = {

            connection => {

                sequence  => $self->{ basename ( $infile ) }{ 'time' },
                time      => $self->{ basename ( $infile ) }{ 'real_time' },
                protocol  => [ "HTTP" ],
                ressource => [ $request ],
                code      => [ $return_code ],
                ip        => [ $self->{ basename ( $infile ) }{ 'ip' } ],
                location  => [ $loc ],
                findings  => {

                  }
              }
          };

        if( defined ( $response_body ) ) {

            if ( my $google_referer = $self->_strip_google( $response_header, $request_header, $response_body, $infile ) ) {

                $xml->{ 'connection'  }{ 'findings' }{ 'method' }[ $x++ ] =
                  {
                    name    => "GoogleAnalyticsParser",
                    output  => [ "Referrer: "  . $google_referer ]
                  };

            }

            if ( my ($parent_request, $parent_sequence ) = $self->_strip_javascript( $response_header, $request_header, $response_body, $infile, @filename ) ) {

                if ( defined ( $parent_request ) && defined ( $parent_sequence ) ) {

                    $xml->{ 'connection'  }{ 'findings' }{ 'method' }[ $x++ ] =
                      {
                        name            => "JSAnalyser",
                        output          => [ "JavaScript found" ],
                        parent_request  => [ $parent_request    ],
                        parent_sequence => [ $parent_sequence   ]
                      };

                }
            }

            if ( $self->_strip_google_safebrowsing( $response_header, $request_header, $response_body, $infile ) ) {

                $xml->{ 'connection'  }{ 'findings' }{ 'method' }[ $x++ ] =
                  {
                    name    => "GoogleSafebrowsingParser",
                    output  => [ "Google Safebrowsing found" ]
                  };

            }

        }

        $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                RootName      => "connections",
                NoAttr        => 0,
                SuppressEmpty => 1,
                NoSort        => 1,
                NoEscape      => 0
              )
          );

    }

}

sub _read_non_http {

    my $self = shift;
    my ($r, $infile) = @_;

    my $f_type = $self->_file_type( substr( $r, 0, 300 ) );

    if( defined ( $f_type ) &&  $self->_is_interesting_filetype( $f_type ) ) {

        if ($r =~ /^\x1f\x8b\x08/ )  {
            my $uncompress;
            gunzip \$r => \$uncompress; # unzip content if it is gziped
            $r = $uncompress;
        }

        my $md   = md5_hex(  $r );
        my $sha1 = sha1_hex( $r );

        if ( open (DA, ">>$self->{ 'BINARY_DIR' }/NON_HTTP_dump-" . $md ) ) {
            binmode(DA);
            print DA "$r";
            close(DA);

        }

        my $data2 = `file -b $self->{ 'BINARY_DIR' }/NON_HTTP_dump-$md`;
        $data2 =~ s/[\r\n]+//g;

        my $mime_typ = `file -bi $self->{ 'BINARY_DIR' }/NON_HTTP_dump-$md`;
        $mime_typ =~ s/[\r\n]+//g;

        my $ssdeep = $self->_ssdeep (  $self->{ 'BINARY_DIR'} . '/NON_HTTP_dump-' . $md );

        my $logdata = $infile . "\t0\tNO_HOST\tNO_FNAME\tNO_REFERER\tNO_BROWSER\n";

        my $file_type = $self->_get_filetypestring( $f_type );
        my ( $traffic_type, $traffic_file, $traffic_source ) = $self->_guess_traffic_type( substr( $r, 0, 800 ) );

        my ( $x, $z );

        my $xml = {

            file => {

                source      => [ $traffic_source ],
                filename    => [ $traffic_file ],
                extractname => [ $self->_basepath( $self->{ 'BINARY_DIR' } . "/NON_HTTP_dump-" . $md ) ],
                analysis    => {

                  }
              },

            dumpfile => {

                file     =>  [ $self->_basepath( $infile ) ],
                analysis => {

                  }

              }
          };

        $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
            name   =>  "mapwocanalyse.pl/type_classifier",
            output => [ $file_type ]
          };

        $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
            name   =>  "tool:file -b",
            output => [ $data2 ]
          };

        $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
            name   =>  "tool:file -bi",
            output => [ $mime_typ ]
          };

        $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
            name   =>  "tool:sha1sum",
            output => [ $sha1 ]
          };

        $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
            name   =>  "tool:ssdeep",
            output => [ $ssdeep ]
          };

        while ( my ( undef, $avoptions ) = each ( %{ $self->{ 'antivir_config' }{ 'Antivir' } } ) ) {

            $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] = {
                name   =>  "tool:" . $avoptions->{ 'name' },
                output => [ "###" . $avoptions->{ 'name' } .  "###NON_HTTP_dump-" . $md . "###" ]
              };

            $xml->{ 'dumpfile'  }{ 'analysis' }{ 'method' }[ $z++ ] = {
                name   =>  "tool:" . $avoptions->{ 'name' },
                output => [ "###" . $avoptions->{ 'name' } . "###" . basename ( $infile ) . "###" ]
              };
        }

        $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                RootName      => "files",
                NoAttr        => 0,
                SuppressEmpty => 0,
                NoSort        => 1,
                NoEscape      => 0
              )
          );

        $xml = {

            connection => {

                sequence  => $self->{ basename ( $infile ) }{ 'time' },
                time      => $self->{ basename ( $infile ) }{ 'real_time' },
                protocol  => [ $traffic_type ],

                #                            ressource => [ $request ],
                code      => [ "none" ],
                ip        => [ $self->{ basename ( $infile ) }{ 'ip' } ],
                findings  => {

                  }
              }
          };

        $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                RootName      => "connections",
                NoAttr        => 0,
                SuppressEmpty => 1,
                NoSort        => 1,
                NoEscape      => 0
              )
          )

    }

}

sub _file_type {

    my $self = shift;
    my $data = shift;

    if( length( $data ) < 1 ) {
        return "empty_body";
    }

    my $type = "";

    if ($data =~ /^GIF8[7-9]/)                                                            { $type = "gif";    }
    elsif ($data =~ /^\377.....(JPEG|JFIF)/)                                              { $type = "jpeg";   }
    elsif ($data =~ /^.PNG/)                                                              { $type = "png";    }
    elsif ($data =~ /^PK\003\004/)                                                        { $type = "zip";    }
    elsif ($data =~ /^\%PDF/)                                                             { $type = "pdf";    }
    elsif ($data =~ /^\037\213/)                                                          { $type = "gz";     }
    elsif ($data =~ /^BZh/)                                                               { $type = "bz2";    }
    elsif ($data =~ /^\177ELF/)                                                           { $type = "elf";    }
    elsif ($data =~ /^\%!/)                                                               { $type = "ps";     }
    elsif ($data =~ /^\x4D\x5A.\x00.\x00\x00\x00.\x00\x00\x00\xff\xff\x00\x00/)           { $type = "exe";    }
    elsif ($data =~ /^\xca\xfe\xba\xbe\x00\x00\x00/)                                      { $type = "mach-o"; }
    elsif ($data =~ /<html>/i)                                                            { $type = "html";   }
    elsif ($data =~ /^\xD0\xCF\x11\xE0\xA1\xB1\x1A\xE1/)                                  { $type = "ms_doc"; }
    elsif ($data =~ /ustar/)                                                              { $type = "tar";    }
    elsif ($data =~ /^\x43\x57\x53/ || $data =~ /^\x46\x57\x53/)                          { $type = "swf";    }
    elsif ($data =~ /^\x46\x4c\x56\x01/)                                                  { $type = "flv";    }
    else { $type = "data"; }

    return $type;
}

sub _is_interesting_filetype {

    my $self = shift;
    my $file_type = shift;

    return TRUE; # atm we are interested in all filetypes.

    if( $self->_is_image( $file_type ) ) { # If we are not interested in images.
        return FALSE;
    }

    my %intersting;

    $intersting{ 'pdf' }     = 1;
    $intersting{ 'exe' }     = 1;
    $intersting{ 'tar' }     = 1;
    $intersting{ 'elf' }     = 1;
    $intersting{ 'zip' }     = 1;
    $intersting{ 'ms_doc'}   = 1;
    $intersting{ 'mach-o' }  = 1;
    $intersting{ 'gz' }      = 1;
    $intersting{ 'bz2' }     = 1;
    $intersting{ 'swf' }     = 1;
    $intersting{ 'flv' }     = 1;

    if( exists ( $intersting{ $file_type } ) ) {
        return TRUE;
    }

    return FALSE;

}

sub _get_filetypestring {

    my $self = shift;
    my $file_type = shift;

    my %type;

    $type{ 'pdf'    }  = "Portable Document Format";
    $type{ 'exe'    }  = "Windows Executable/PE";
    $type{ 'tar'    }  = "Tape Archive";
    $type{ 'elf'    }  = "Executable and Linkable Format";
    $type{ 'zip'    }  = "Compressed ZIP";
    $type{ 'ms_doc' }  = "MS Office Dokument";
    $type{ 'mach-o' }  = "Mach object file";
    $type{ 'gz'     }  = "GNU zip";
    $type{ 'bz2'    }  = "Compressed Bzip2";
    $type{ 'gif'    }  = "Graphics Interchange Format";
    $type{ 'png'    }  = "Portable Network Graphics";
    $type{ 'jpeg'   }  = "JPEG";
    $type{ 'html'   }  = "HTML";
    $type{ 'swf'   }   = "Shockwave Flash file";
    $type{ 'flv'   }   = "Flash video file";

    if ( exists ( $type{ $file_type } ) ) {

        return $type{ $file_type };

    } else {

        return "unknnown filetype";

    }

}

sub _is_image {

    my $self = shift;
    my $file_type = shift;

    if( $file_type eq "gif" || $file_type eq "png" || $file_type eq "jpeg" ) {
        return TRUE;
    }

    return FALSE;

}

sub _get_http_return_code {

    my $self = shift;
    my $response_header = shift;

    my ($http_code, $location);

    $response_header =~ s/HTTP\/1\.(0|1) ([0-9]+)/$http_code = $2;/eg if defined ( $response_header  );
    $response_header =~ s/location: ([^\r\n]+)/$location = $1;/egi if defined ( $response_header  );

    if ( ! defined ( $location ) ) {
        $location = undef;
    } else {
        $location =~ s/\x0D//;
        $location = $self->_clean_string( $location );
        $self->{ 'location' } = $location;
    }

    if ( ! defined ( $http_code ) ) {
        $http_code = "NO_HTTP_CODE";
    }

    return ($http_code, $location);

}

sub _parent_request {

    my $self = shift;
    my ( $fname, $infile, @files ) = @_;

    my @check_files = @files;

    my $parent_request  = "unknown";
    my $parent_sequence = "unknown";
    my $host            = "unknown";

    my $found = undef;

    foreach my $name ( @check_files ) {

        last if defined ( $found );

        if( defined ( $self->{ basename ( $name ) }{ 'data' } ) ) {

            if( basename ( $name ) ne basename ( $infile )) { # ignore self

                my $data = $self->{ basename ( $name ) }{ 'data' };
                my @data_split = split( /###split###/, $data) ;

                foreach my $check_data ( @data_split ) {

                    my $pos = index( $check_data, $fname );

                    if( $pos >= 0 ) {

                        if( $check_data =~ m/([hH][rR][eE][fF]|[sS][rR][cC]).*\Q$fname\E/g) {

                            my ( $req, $seq );

                            $found = TRUE;

                            $check_data =~ s/(GET|POST)\s(.*)\sHTTP\/1\..[\x0a\x0d]/$req = $2;/egi;
                            $check_data =~ s/Host: ([^\n\r]+)/$host = $1;/egi;
                            $check_data =~ s/TIME:([0-9]+)/$seq = $1;/egi;

                            $parent_request  = $host.$req;
                            $parent_sequence = $seq;

                            last;

                        }
                    }
                }

            }

        }

    }

    return ( $parent_request, $parent_sequence );

}

sub _strip_javascript {

    my $self = shift;
    my ( $response_header, $request_header, $response_body, $infile, @files ) = @_;

    my @check_files = @files;

    my $fname           = "unknown";
    my $host            = "unknown";
    my $parent_request  = "unknown";
    my $parent_sequence = "unknown";

    if ( $response_header =~ /[jJ]avascript/) {

        $request_header =~ s/(GET|POST)[\x00-\xff]{1,}\/(.[\x00-\xff]{1,}) HTTP\/1\..[\x0a\x0d]/$fname = $2;/egi;

        my ($parent_request, $parent_sequence) = $self->_parent_request( $fname, $infile, @check_files );

        my ( $outfile1 ) = $self->_make_outfile1( $response_header, $request_header, $infile );

        if( ! defined ( $outfile1 ) ) {
            return FALSE;
        }

        if ( open(NFH, "> " . $self->{ 'BINARY_DIR' } . "/" . "$outfile1") ) {
            binmode(NFH);
            print NFH "$response_body";
            close(NFH);
        }

        return ( $parent_request, $parent_sequence );
    }

    return ( undef, undef);

}

sub _strip_google {

    my $self = shift;
    my ( $response_header, $request_header, $response_body, $infile ) = @_;

    my $header1 = $request_header . $response_header;
    my ( $logheader, $ftime, $fname, $referer, $host, $browser, $location, $outfile1 );

    if( defined $header1 ) {

        $header1=~s/TIME:(.[a-zA-Z0-9 :]{1,})/$ftime=$1;/eg;

        if( defined $ftime ) {

            $header1=~s/Referer: http[s]*:\/\/(.[\x10-\xff]{1,})[\x0a\x0d]/$referer=$1;/egi;
            $header1=~s/Host: (.[\x10-\xff]{1,})[\x0a\x0d]/$host=$1;/egi;

            if( ! defined $referer ) { $referer = "NO_REFERER"; }
            if( ! defined $host )    { $host    = "NO_HOST"; }

            $logheader =$infile."\t" . $ftime . "\tHost:" . $host . "\tReferer:" . $referer . "\n";

        }

        if( ( defined $host ) && ( $host =~ /[Gg][Oo][Oo][Gg][Ll][Ee]-[Aa][Nn][Aa]/) ) {

            my ( $outfile1 ) = $self->_make_outfile1( $response_header, $request_header, $infile );

            if( ! defined ( $outfile1 ) ) {
                return FALSE;
            }

            my $md = md5_hex( $response_body );

            open(NFH, "> $self->{ 'BINARY_DIR' }/GOOGLE-" . $outfile1 . "-" .$md ) or die "Could not open $outfile1";
            binmode(NFH);
            print NFH "$response_body";
            close(NFH);

            return $referer;
        }

        return FALSE;
    }

    return FALSE;

}

sub _strip_google_safebrowsing {

    my $self = shift;
    my ( $response_header, $request_header, $response_body, $infile ) = @_;

    if ( ( $response_header =~ /safebrowsing/ ) && ( $response_header =~ /gethash/ ) ) {

        my ( $outfile1 ) = $self->_make_outfile1( $response_header, $request_header, $infile );

        if( ! defined ( $outfile1 ) ) {
            return FALSE;
        }

        if ( open(NFH, "> $self->{ 'BINARY_DIR' }/"."$outfile1") ) {
            binmode(NFH);
            print NFH "$response_header\n$response_body";
            close(NFH);
        }

        return TRUE;

    }

    return FALSE;

}

sub _make_outfile1 {

    my $self = shift;
    my ($response_header, $request_header, $infile) = @_;

    my $header = $request_header . $response_header;

    my ($ftime, $fname, $referer, $host, $browser, $location, $outfile1);

    if( defined $header ) {

        $header =~ s/TIME:(.[a-zA-Z0-9 :]{1,})/$ftime = $1;/eg;
        $header =~ s/(GET|POST)[\x00-\xff]{1,}\/(.[\x00-\xff]{1,}) HTTP\/1\..[\x0a\x0d]/$fname = $2;/eg;
        $header =~ s/Referer: http[s]*:\/\/(.[\x10-\xff]{1,})[\x0a\x0d]/$referer = $1;/egi;
        $header =~ s/Host: (.[\x10-\xff]{1,})[\x0a\x0d]/$host = $1;/egi;
        $header =~ s/User-Agent: (.[\x10-\xff]{1,})[\x0a\x0d]/$browser = $1;/egi;

        $header =~ s/Location: ([^\r\n]+)/$location = $1;/egi;

        if( ! defined $fname)   { $fname   = "NO_NAME";    }
        if( ! defined $referer) { $referer = "NO_REFERER"; }
        if( ! defined $host)    { $host    = "NO_HOST";    }

        if (length $host    > 15)  {   $host    = substr( $host, 0, 13 );    }
        if (length $referer > 15)  {   $referer = substr( $referer, 0, 13 ); }
        if (length $fname   > 15)  {   $fname   = substr( $fname, 0, 13 );   }

        $outfile1 = $host . "_" . $referer . "_" . $fname;
        $outfile1 = $self->_clean_filename( $outfile1 );

    }
    else {
        my $rand = int( rand( 1000000 ) );
        $outfile1="NO_HEADERINFO_" . $rand;
    }

    if( defined ( $outfile1 ) ) {
        return ( $outfile1, $fname );
    } else {
        return FALSE;
    }

}

sub _parse_request_header {

    my $self   = shift;
    my $header = shift;

    my ( $host, $fname, $request );

    if( $header =~ m/(GET|POST) (.*) HTTP\/1\..[\x0a\x0d]/i ) {
        $request = $2;
    }

    $header =~ s/($interesting_methods)[\x00-\xff]{1,}\/(.[\x00-\xff]{1,}) HTTP\/1\..[\x0a\x0d]/$fname = $2;/egi;
    $header =~ s/Host: (.[\x10-\xff]{1,})[\x0a\x0d]/$host = $1;/egi;

    if( ! defined ( $request ) ) { $request = ""; }
    if( ! defined ( $fname   ) ) { $fname   = ""; }
    if( ! defined ( $host    ) ) { $host    = "NO_HOST"; }

    $request = $self->_clean_string( $request );

    if( $request =~ m/^HTTP[s]*:\/\//i) {
        return ($request, $fname);
    } else {
        return ( "http://" . $host . $request, $fname );
    }

}

sub _ssdeep {

    my $self = shift;
    my $file = shift;

    my $ssdeep = `ssdeep -sb $file 2>/dev/null`;
    my @sd = split(/\n/, $ssdeep);

    if ( defined ( $sd[1] ) ) {

        return $sd[1];

    }  else {

        return "NO SSDEEP";

    }

}

sub _guess_traffic_type { # At the moment only FTP

    my $self    = shift;
    my $traffic = shift;

    my $traffic_file   = "UNKNOWN";
    my $traffic_type   = "NON_HTTP/UNKNOWN";
    my $traffic_source = "UNKNOWN";

    if( defined ( $self->{ 'next_is_ftp_data' }  ) ) {

        $self->{ 'next_is_ftp_data' } = undef;

        $traffic_type = "FTP";

        if ( defined ( $self->{ 'location' } ) ) {

            if( $self->{ 'location' } =~ m/^ftp:\/\//i ) {
                $traffic_source = $self->{ 'location' };
            }

            $self->{ 'location' } = undef;
        }

        if( defined ( $self->{ 'next_ftp_data_trafficfile' } ) ) {

            $traffic_file = $self->{ 'next_ftp_data_trafficfile' };
            $self->{ 'next_ftp_data_trafficfile' } = undef;

        }
    }

    if ( $traffic =~ m/^220(.*)FTP/i ) {

        $traffic_type = "FTP";
        $self->{ 'next_is_ftp_data' } = 1;

        if ( $traffic =~ m/(get|recv|retr)\s([^\r\n]+)/i ) {

            $traffic_file = $2;
            $traffic_file =~ s/\x0D//;
            $self->{ 'next_ftp_data_trafficfile' } = $traffic_file;
            $traffic_file = "FTP-Header";

            if ( defined ( $self->{ 'location' } ) ) {

                if( $self->{ 'location' } =~ m/^ftp:\/\//i ) {

                    $traffic_source = $self->{ 'location' };

                }

            }

        }
    }

    return ( $traffic_type, $traffic_file, $traffic_source );

}

sub _parse_fastmode_urls {

    my $self = shift;
    my $body = shift;

    if ( $body  =~ m/MAPWOC_FASTMODE_URL\s"(.*)"/g ) {

        my @urls = split( /","/, $1 );

        foreach my $url ( @urls ) {

            $url =~ s/"//g;
            $url =~ s/\/$//;

            $self->{ 'fastmode_urls' }{ $url } = 1;

        }

    }

}

sub _url_is_fastmode_url {

    my $self = shift;
    my $url  = shift;

    $url =~ s/\/$//;

    if ( exists ( $self->{ 'fastmode_urls' }{ $url } ) ) {

        return TRUE;

    }

    return undef;
}

sub _read_json {

    my $self      = shift;
    my $json_file = shift;

    if( ! open( JSFILE, '<', $json_file ) ) { # open json file
        print "ERROR: Could not read $json_file in _read_json(). Exit now\n";
        exit 255;
    }

    my $json = new JSON;

    my $json_text = <JSFILE>;
    close(JSFILE);

    my @json_split = split ( /[0-9]+:{/, $json_text); # content (from mitmdump) is JSON, encapsed in Netstring but Text::Netstring fails to decode it. Guess we need one Netstring per line for Text::Netstring while we get all netstrings in a single line. Therefore, we decode it yourself...

    foreach my $content ( @json_split ) {

        $content = '{' . $content;
        chop( $content );

        my $valid_json = eval { $json->decode( $content ); 1 }; # do we have valid json

        if ( $valid_json ) {

            my $json_content = $json->decode( $content ); # decode json
            $self->_read_json_traffic( $json_content );

        }

    }
}

sub _read_json_traffic {

    my $self = shift;
    my $json_content = shift;

    my $referer       = "NO_REFERER";
    my $is_javascript = undef;

    my ( $location, $response_content, $browser, $req_resp, $response_header );

    my $host           =  $json_content->{ 'request' }->{ 'host' };
    my $request        =  $json_content->{ 'request' }->{ 'path' };
    my $request_file   =  basename( $request );
    my $target_port    =  $json_content->{ 'request' }->{ 'port' };

    my $timestamp      =  $json_content->{ 'request' }->{ 'timestamp' };
    my @timestamp      =  split(/\./, $timestamp);
    $timestamp         =  $timestamp[0];

    my $request_typ    =  $json_content->{ 'request' }->{ 'scheme' };
    my $local_port     =  $json_content->{ 'request' }->{ 'client_conn' }[1];
    my $target_ip      =  $self->_hostname2ip( $host );
    my $method         =  $json_content->{ 'request' }->{ 'method' };

    my $request_header = "TIME:" . $timestamp . ' ' . $method . ' ' . $request . " HTTP/1.1\n";
    my $response_code  = $json_content->{ 'response' }->{ 'code' };

    my $real_time = $self->_ts2time( $timestamp );

    if( defined ( $json_content->{ 'response' }->{ 'content' } ) && length ( $json_content->{ 'response' }->{ 'content' } ) > 1 ) {

        $response_content = decode_base64( $json_content->{ 'response' }->{ 'content' } );

        if ( $response_content =~ /^\x1f\x8b\x08/  ) {
            my $uncompress;
            gunzip \$response_content => \$uncompress;
            $response_content = $uncompress;
        }

        foreach ( @{ $json_content->{ 'request' }->{ 'headers' } } ) {

            if( $_->[ 0 ] eq "referer" ) {
                $referer = $_->[ 1 ];
            }

            if( $_->[ 0 ] eq "user-agent" ) {
                $browser = $_->[ 1 ];
            }

            $request_header .= $_->[ 0 ] . ': ' . $_->[ 1 ] . "\n"; # build complete request header

        }

        $request_header .= "\n";

        $response_header = "HTTP/1.1 " . $response_code . "\n";

        foreach ( @{ $json_content->{ 'response' }->{ 'headers' } } ) {

            if( $_->[ 1 ] =~ m/[jJ]avascript/ ) {
                $is_javascript = TRUE;
            }

            if( $_->[ 0 ] eq "location" ) {
                $location = $_->[ 1 ];
            }

            $response_header .= $_->[ 0 ] . ': ' . $_->[ 1 ] . "\n"; # build complete response header

        }

        $response_header .= "\n";

        $req_resp = $request_header . $response_header . $response_content . "\n\n"; # build complete traffic dump. Without that we can't use all the functions written for pcap-based traffic.

        my $md   = md5_hex( $response_content );
        my $sha1 = sha1_hex( $response_content );
        my $rand = int( rand ( 1000000 ) );

        my $peer_string =  $self->{ 'work_dir' } . "/json_" . $target_ip . "_" . $local_port . "_" . $target_port . "_" . $rand;;

        if ( open (AUS , ">>" . $peer_string ) ) {

            binmode(AUS);
            print AUS  "$req_resp";
            close( AUS );
            push ( @jsonfiles, $peer_string );
            $self->{ basename ( $peer_string ) }{ 'data' } = $req_resp;

        }

        my $outfile1 = $host . "_" . $referer . "_" . $request_file;
        $outfile1 = $self->_clean_filename( $outfile1 );

        my $f_type = $self->_file_type( substr( $response_content, 0, 300 ) );

        if( defined ( $f_type ) && $self->_is_interesting_filetype( $f_type ) ) {

            if ( open (DA, ">> $self->{ 'BINARY_DIR' }/" .  $outfile1 . "-" . $md ) ) {

                my $file_type = $self->_get_filetypestring( $f_type );

                binmode(DA);
                print DA "$response_content";
                close(DA);

                my $data2 = `file -b $self->{ 'BINARY_DIR' }/$outfile1-$md`;
                $data2 =~ s/[\r\n]+//g;

                my $mime_typ = `file -bi $self->{ 'BINARY_DIR' }/$outfile1-$md`;
                $mime_typ =~ s/[\r\n]+//g;

                my $ssdeep = $self->_ssdeep ( $self->{ 'BINARY_DIR' } . "/" . $outfile1 . "-" . $md );

                my ( $x, $z );

                my $xml = {

                    file => {

                        source          => [ $request_typ . "://" . $host . $request ],
                        filename        => [ $request_file ],
                        extractname     => [ $self->_basepath( $self->{ 'BINARY_DIR' } . "/" .  $outfile1 . "-" . $md ) ],
                        analysis        => {

                          }

                      },

                    dumpfile => {

                        file     =>  [ $self->_basepath( $peer_string ) ],
                        analysis => {

                          }

                      }

                  };

                $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "mapwocanalyse.pl/type_classifier",
                    output => [ $file_type ]
                  };

                $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:file -b",
                    output => [ $data2 ]
                  };

                $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:file -bi",
                    output => [ $mime_typ ]
                  };

                $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:sha1sum",
                    output => [ $sha1 ]
                  };

                $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =  {
                    name   =>  "tool:ssdeep",
                    output => [ $ssdeep ]
                  };

                while ( my ( undef, $avoptions ) = each ( %{ $self->{ 'antivir_config' }{ 'Antivir' } } ) ) {

                    $xml->{ 'file'  }{ 'analysis' }{ 'method' }[ $x++ ] =

                      {
                        name   =>  "tool:" . $avoptions->{ 'name' },
                        output => [ "###" . $avoptions->{ 'name' } . "###" . $outfile1 . "-" . $md . "###" ]
                      };

                    $xml->{ 'dumpfile'  }{ 'analysis' }{ 'method' }[ $z++ ] =

                      {
                        name   =>  "tool:" . $avoptions->{ 'name' },
                        output => [ "###" . $avoptions->{ 'name' } . "###" . basename ( $peer_string ) . "###" ]
                      }

                }

                $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                        RootName      => "files",
                        NoAttr        => 0,
                        SuppressEmpty => 0,
                        NoSort        => 1,
                        NoEscape      => 0
                      )
                  );

            }

        }

    }

    my $x;

    my $xml = {

        connection => {

            sequence  => $timestamp,
            time      => $real_time,
            protocol  => [ $request_typ ],
            ressource => [ $request_typ . "://" . $host . $request ],
            code      => [ $response_code ],
            ip        => [ $target_ip ],
            location  => [ $location ],
            findings  => {

              }
          }
      };

    if( $host =~ /[Gg][Oo][Oo][Gg][Ll][Ee]-[Aa][Nn][Aa]/ ) {

        $xml->{ 'connection'  }{ 'findings' }{ 'method' }[ $x++ ] =
          {
            name    => "GoogleAnalyticsParser",
            output  => [ "Referrer: " . $referer ]
          }

    }

    if( defined ( $response_header ) && defined ( $request_header ) ) {

        if( defined ( $is_javascript ) ) {

            if ( my ($parent_request, $parent_sequence ) = $self->_strip_javascript( $response_header, $request_header, $response_content, 'dummy', @jsonfiles ) ) {

                $xml->{ 'connection'  }{ 'findings' }{ 'method' }[ $x++ ] =
                  {
                    name            => "JSAnalyser",
                    output          => [ "JavaScript found" ],
                    parent_request  => [ $parent_request    ],
                    parent_sequence => [ $parent_sequence   ]
                  }

            }

        }

        if ( $self->_strip_google_safebrowsing( $response_header, $request_header, $response_content, 'dummy' ) ) {

            $xml->{ 'connection'  }{ 'findings' }{ 'method' }[ $x++ ] =
              {
                name    => "GoogleSafebrowsingParser",
                output  => [ "Google Safebrowsing found" ]
              }

        }

        $self->{ 'XML' }->_add_xml( XMLout ( $xml,
                RootName      => "connections",
                NoAttr        => 0,
                SuppressEmpty => 1,
                NoSort        => 1,
                NoEscape      => 0
              )
          );

    }

}

sub _hostname2ip  {

    my $self = shift;
    my $host = shift;

    my $hostname = $host;

    my $addr = ( gethostbyname( $host ) ) [4];
    if( defined ( $addr ) ) {
        my @oct   = unpack( "C4", $addr );
        $hostname = join( ".", @oct );
    }

    return( $hostname );
}

sub _ts2time {

    my $self      = shift;
    my $timestamp = shift;

    my ( $sec, $min, $hour, $mday, $mon, $year, undef, undef, undef) = gmtime( $timestamp );

    $mon++;
    $sec  = "0" . $sec  if ( $sec  < 10 );
    $min  = "0" . $min  if ( $min  < 10 );
    $hour = "0" . $hour if ( $hour < 10 );
    $mday = "0" . $mday if ( $mday < 10 );
    $mon  = "0" . $mon  if ( $mon  < 10 );

    my $real_time = $year + 1900 . "-" . $mon . "-" . ${mday} ."-" . $hour .":" . $min . ":" . $sec;

    return $real_time;

}

sub _clean_filename {

    my $self   = shift;
    my $string = shift;

    $string =~ s/[^a-zA-Z0-9-_.]/_/mg;

    return $string;

}

sub _clean_string {

    my $self = shift;
    my $string = shift;

    $string =~ s/[[:cntrl:]]//g;

    return $string;

}

sub _basepath {

    my $self = shift; 
    my $path = shift;

    if ( $path =~ m/\/.*\/(.*\/.*)$/) {
    
        return $1;

    } else {

        return $path;

    }   
}

1;

__END__

=head1 NAME

MapWOCAnalyser::MapWOCAnalyser - Perl module for mapwoc 

=head1 SYNOPSIS

use MapWOCAnalyser::MapWOCAnalyser qw( @filename )

my $analyser = MapWOCAnalyser::MapWOCAnalyser->_new(
                                                    BINARY_DIR  => $BINARY_DIR,
                                                    pcapfile    => $pcapfile,
                                                    work_dir    => $work_dir,
                                                    LOG_DIR     => $LOG_DIR,
                                                    xmlfile     => $xmlfile,
                                                    json_file   => $json_file
                                              )  

=head1 DESCRIPTION

This module is a collection of functions for mapwoc (massive automated passive Web Observation Center). 

=over

=item _syn_packets()

Callback funktion for Net::Pcap::loop(). Read the pcapfile and dump the traffic into files. No return value.

=item _read_traffic()

Open and read the files created in _syn_packets(). Calls _read_http or _read_non_http. Return always TRUE.

=item _file_type()

Return file typ for given data, else just "data".

=item _is_interesting_filetype()

Check if file typ is interesting for us. Return TRUE or FALSE.

=item _read_http()

Read the HTTP-Traffic, do some checks and store information in XML. No return Value.

=item _read_non_http()

Same as _read_http() but for non http traffic.

=item _get_filetypestring()

Return full filetypestring from fileextension. Return e.g."Windows Executable/PE"

=item _is_image()

Check if File is an image.

=item _log_http_code()

Log the HTTP response code. Return TRUE or FALSE.

=item _get_http_return_code()

Return HTTP response code and, if available the Location.

=item _parent_request()

Determine the parent request for given file. Return the parent request and the parent sequence.

=item _strip_javascript()

Determine if given file is JA. Return parent request and parent sequence.

=item _strip_google()

Determine ig request is google analytics request. Return the referer.

=item _strip_google_safebrowsing()

Determine ig request is googlesafebrowsing request. Return TRUE or FALSE.

=item _make_outfile1()

Create filename for given data. Return a filename.

=item _parse_request_header()

Parse request header. Return full request and requested filename.

=item ssdeep()

Call ssdeep for given data. Return ssdeep output.

=item _guess_traffic_type()

Try to determine the traffic type. Only HTTP and FTP are supportet at the moment. Return Traffictype, Trafficfile and Trafficsource.

=item _parse_fastmode_urls()

Get all fastmode URLs from response body if request is fastmode request. Store all URLs in session. No return value.

=item _url_is_fastmode_url()

Check if given URL is a fastmode URL. Return TRUE or FALSE.

=item _read_json()

Open a JSON file decode the Netstring and call _read_json_traffic(). No return value.

=item _read_json_traffic()

Same as _read_http() but for JSON data. No return value.

=item _hostname2ip()

Get the ip for given hostname. Return the IP.

=item _ts2time()

Convert a timestamp into human readeable format.

=item _clean_filename()

Remove unwanted signs from filename. Return a "clean" filename.  

=item _clean_string()

Remove control chars from string  

=item _basepath

Return the relative path to dumpfile and extractname.

=back

=head1 LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2,
or, at your option, any later version as published by the Free
Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

=head1 AUTHOR

Michael Meyer <michael.meyer@intevation.de>

=head1 COPYRIGHT

Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik

=cut
