# Copyright:
# Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

# XML functions for MapWOCAnalyser

package MapWOCAnalyser::XML;

use strict;
use warnings;

use Exporter;
our @ISA = qw( Exporter );
our @EXPORT_OK = qw( _start_xml _close_xml _clean_xml _add_xml _add_antivir_result_to_xml );

use XML::Simple;
use File::Basename;
use Date::Calc qw(Normalize_DHMS Delta_DHMS);
use utf8;
use HTML::Entities;

use Data::Dumper;

use constant { 
    TRUE  => 1,
    FALSE => 0
};

sub _new {

    my ( $this, %args ) = @_;
    my $self = {};

    my $err;

    bless $self, $this;

    $self->{ 'xmlfile' }  = $args{ 'xmlfile' };

    return $self;

}

sub _start_xml {

    my $self = shift;

    my $hcsid = shift;

    my $xml = '<?xml version="1.0" encoding="UTF-8"?>' . "\n"  . '<visit visitor="' . $hcsid  . '">' . "\n";
    $self->{'current_xml'} = $xml;

    return TRUE;
}

sub _close_xml {

    my $self = shift;

    my $duration = "unknown";

    $self->_clean_xml; # Remove unneeded antivir placeholders

    if( defined ( $self->{ 'pcap_start' } ) && defined ( $self->{ 'pcap_stop' } ) ) {

        my $start = $self->{ 'pcap_start' };
        my $stop  = $self->{ 'pcap_stop' };

        $start =~ s/:/-/g;
        $stop  =~ s/:/-/g;

        my @start = split( /-/, $start );
        my @stop  = split( /-/, $stop  );

        my ( undef, $Dh, $Dm, $Ds) = Delta_DHMS( $start[0], $start[1], $start[2], $start[3], $start[4], $start[5], $stop[0], $stop[1], $stop[2], $stop[3], $stop[4], $stop[5]);

        if( $Dh < 10 ) { $Dh = "0" . $Dh; }
        if( $Dm < 10 ) { $Dm = "0" . $Dm; }
        if( $Ds < 10 ) { $Ds = "0" . $Ds; }

        $duration = $Dh . ":" . $Dm . ":" . $Ds;

    }

    my $xml = "<duration>" . $duration . "</duration>\n";

    $xml .= "</visit>";
    $self->{'current_xml'} .= $xml;

    utf8::encode( $self->{ 'current_xml' } );

    if( ! open( XML, ">> $self->{ 'xmlfile' }" ) ) {
        print "ERROR: Could not open XML-File $self->{ 'xmlfile' } for writing in Function close_xml(). Exit now!\n";
        exit 255;
    }

    print XML "$self->{'current_xml'}";
    close( XML );

    return TRUE;

}

sub _add_xml {

    my $self = shift;
    my $xml = "@_";
    $self->{'current_xml'} .= $xml;

    return TRUE;

}

sub _add_antivir_result_to_xml {

    my $self = shift;
    my ( $avscanner, $scanned_file, $av_line ) = @_;

    if( $av_line =~ m/^scanner error/) {
        $self->{'current_xml'} =~ s/###$avscanner###.*###/$av_line/g;
        return TRUE;
    }

    my $antivir_file = basename( $scanned_file );
    $antivir_file =~ s/\r\n//g;

    my $antivir_result = $av_line;
    $antivir_result =~ s/^\s*//g;
    encode_entities($antivir_result);

    $self->{'current_xml'} =~ s/\Q###$avscanner###$antivir_file###\E/$antivir_result/g;

    return TRUE;

}

sub _clean_xml {

    my $self = shift;

    $self->{'current_xml'} =~ s/###(.*)###/no virus/g;

    return TRUE;
}

1;

__END__

=head1 NAME

MapWOCAnalyser::XML - XML Functions for MapWOC

=head1 SYNOPSIS

use MapWOCAnalyser::XML;

my $XML = MapWOCAnalyser::XML->_new( xmlfile => $self->{ 'xmlfile'} );

=head1 DESCRIPTION

This module is a collection of XML functions for mapwoc (massive automated passive Web Observation Center).

=over

=item _start_xml()

Initiate the XML. Return TRUE or FALSE.

=item _close_xml()

Close the XML.  No return value.

=item _add_xml()

Add values to XML. Return TRUE.

=item _add_antivir_result_to_xml()

Replace placeholders with results from antivir scanner. Return TRUE.

=item _clean_xml()

Remove unneeded placeholders from XML. Return TRUE

=back

=head1 LICENSE

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2,
or, at your option, any later version as published by the Free
Software Foundation

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-13

=head1 AUTHOR

Michael Meyer <michael.meyer@intevation.de>

=head1 COPYRIGHT

Copyright (C) 2011 Bundesamt fuer Sicherheit in der Informationstechnik

=cut



