# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Information about controlling a certain HoneyClient.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.


import ConfigParser

from hc import HC

class InvalidHCControlFileException(Exception):
    """Exception raised when a hccontrol file is considered invalid."""

    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class HCControl(HC):
    """
    Bundles HC meta- and control-information.
    Allowed types of HCs are:
     - kvm
     - kvm-mac
     - native
    """

    TYPE_KVM = 'kvm'
    TYPE_KVM_MAC = 'kvm-mac'
    TYPE_NATIVE = 'native'
    TYPES = [TYPE_KVM, TYPE_KVM_MAC, TYPE_NATIVE]

    def __init__(self, filepath=None):
        """
        Initializes a HCControl.
        @param filepath: Reads control information from this file if given.
                         Otherwise, initializes values to empties.
        """

        HC.__init__(self, filepath)

        STR_HOMEPAGE = 'homepage'
        STR_CLOSE_WINDOW = 'close_window'
        STR_ENTER = 'enter'

        shortcut = {STR_HOMEPAGE : 'alt-home',
                    STR_CLOSE_WINDOW : 'alt-f4',
                    STR_ENTER : 'ret'}

        CONFIG_GROUP_CONTROL = 'control'
        CONFIG_GROUP_NETWORK = 'network'
        CONFIG_GROUP_KEYBOARD_SHORTCUTS = 'keyboard-shortcuts'

        if filepath is not None:
            config = ConfigParser.RawConfigParser(shortcut)
            config.read(filepath)
            self.hctype = config.get(CONFIG_GROUP_CONTROL, 'type')

            if self.hctype not in HCControl.TYPES:
                raise InvalidHCControlFileException("Unknown HC Type %s" % self.hctype)

            if self.hctype == 'kvm' or self.hctype == 'kvm-mac':
                self.image_file = config.get(CONFIG_GROUP_CONTROL,
                                            'image_file')
                if config.has_option(CONFIG_GROUP_CONTROL, 'boots'):
                    self.boots = config.getboolean(CONFIG_GROUP_CONTROL, 'boots')
                else:
                    self.boots = False
                    self.image_snapshot = config.get(CONFIG_GROUP_CONTROL,
                                                    'image_snapshot')
                if config.has_option(CONFIG_GROUP_CONTROL, 'kvm_kernel'):
                    self.kvmkernel = config.get(CONFIG_GROUP_CONTROL, 'kvm_kernel')
                else:
                    self.kvmkernel = None
                self.own_ip = config.get(CONFIG_GROUP_NETWORK, 'own_ip')
                self.dns_ip = config.get(CONFIG_GROUP_NETWORK, 'dns_ip')
                if config.has_option(CONFIG_GROUP_NETWORK, 'proxy_ip'):
                    p_ip = config.get(CONFIG_GROUP_NETWORK, 'proxy_ip')
                    if len(p_ip.split(":")) == 2:
                        self.proxy_ip = p_ip.split(":")[0]
                        self.proxy_port = int(p_ip.split(":")[1])
                    else:
                        self.proxy_ip = p_ip
                        self.proxy_port = None
                else:
                        self.proxy_ip = None
                        self.proxy_port = None

                self.gateway_ip = config.get(CONFIG_GROUP_NETWORK,
                                             'gateway_ip')
                self.homepage = config.get(CONFIG_GROUP_NETWORK, 'homepage')

            if self.hctype == HCControl.TYPE_NATIVE:
                self.interface = config.get(CONFIG_GROUP_CONTROL, 'interface')
                self.mac_address = config.get(CONFIG_GROUP_NETWORK, 'client_mac')
                self.ssid = config.get(CONFIG_GROUP_NETWORK, 'wlan_ssid')
                if config.has_option(CONFIG_GROUP_NETWORK, 'proxy_ip'):
                    p_ip = config.get(CONFIG_GROUP_NETWORK, 'proxy_ip')
                    if len(p_ip.split(":")) == 2:
                        self.proxy_ip = p_ip.split(":")[0]
                        self.proxy_port = int(p_ip.split(":")[1])
                    else:
                        self.proxy_ip = p_ip
                        self.proxy_port = None
                if config.has_option(CONFIG_GROUP_NETWORK, 'gateway_ip'):
                    self.gateway_ip = config.get(CONFIG_GROUP_NETWORK,
                                                 'gateway_ip')
                if config.has_option(CONFIG_GROUP_NETWORK, 'homepage'):
                    self.homepage = config.get(CONFIG_GROUP_NETWORK, 'homepage')
            else:
                self.ssid = None

            self.image_memory = config.getint(CONFIG_GROUP_CONTROL,
                                              'image_memory')
            if config.has_option(CONFIG_GROUP_CONTROL, 'vmclock'):
                self.vmclock = config.getboolean(CONFIG_GROUP_CONTROL,
                                                 'vmclock')

            if not config.has_section(CONFIG_GROUP_KEYBOARD_SHORTCUTS):
                config.add_section(CONFIG_GROUP_KEYBOARD_SHORTCUTS)
            for key in shortcut.keys():
                shortcut[key] = config.get(CONFIG_GROUP_KEYBOARD_SHORTCUTS,
                                           key)
            if config.has_option(CONFIG_GROUP_NETWORK, 'nic'):
                self.nic = config.get(CONFIG_GROUP_NETWORK, 'nic')

        else:
            self.image_file = None
            self.image_snapshot = None
            self.image_memory = None
            self.hctype = None
            self.own_ip = None
            self.dns_ip = None
            self.proxy_ip = None
            self.gateway_ip = None
            self.homepage = None

        self.SHORTCUT_HOMEPAGE = shortcut[STR_HOMEPAGE]
        self.SHORTCUT_CLOSE_WINDOW = shortcut[STR_CLOSE_WINDOW]
        self.SHORTCUT_ENTER = shortcut[STR_ENTER]

