# -*- coding: utf-8 -*-

# mapWOC
#
# Description: JSON De/Encoders.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import json

from mapwoc.lib.hc import HC
from hccontrol import HCControl
from message import Message
from redirectorvisit import RedirectorVisit
from scanconfiguration import ScanConfiguration, VisitorConfiguration
from scanconfiguration import VisitorBundle, TransportableSC
from trivialresponse import TrivialResponse
from node import Node
from url import URL, URLList

"""
Shortcuts to create mapping to/from json.
Easy access given by:
for (mclass,name,props) in meta_map_map:
    pass

Note that ordering is only partially alphabetical, in order to
match the 'deepest' class first (concrete: HCControl is subclass of HC;
an HCControl instance is also an HC instance).

A clear improvement would be to generate the first entry in the tuple
from the second (or vice versa).
"""
meta_map_map = (
               (HCControl, 'HCControl', ['hcsid', 'name', 'description',
                                         'active_browser', 'on_nodes',
                                         'image_memory',
                                         'installed_software', 'os', 'hctype',
                                         'ssid']),
               (HC, 'HC', ['hcsid', 'name', 'description',
                           'active_browser', 'on_nodes', 'max_parallel']),
               (Node, 'Node', ['name', 'key', 'description', 'ip', 'port',
                               'rsync_address', 'data_path' ]),
               (Message, 'Message', ['level', 'text', 'heading', 'timestamp']),
               (RedirectorVisit, 'RedirectorVisit', ['date', 'url',
                                                      'hcsid', 'peer']),
               (TransportableSC, 'TransportableSC', ['name',
                                                     'comment',
                                                     'urllistIDs',
                                                     'visitor_configuration',
                                                     'description',
                                                     'enable_vnc',
                                                     'id']),
               (TrivialResponse, 'TrivialResponse', ['error', 'message']),
               (URL, 'URL', ['url']),
               (VisitorBundle, 'VisitorBundle', ['hcsid', 'number', 'node']),
               (VisitorConfiguration, 'VisitorConfiguration', ['active_window',
                                                               'visitors',
                                                               'num_urls',
                                                               'eventmap'])
               )

class MapWOCJSONEncoder(json.JSONEncoder):
    """ Encode MapWOC objects to the JSON format """
    def default(self, python_object):
        """
        Overridden default encoding.
        @param python_object: Object to be jsonized.
        """

        # Use fancy, but probably slow, meta-fu.
        obj = python_object
        for (mclass,name,props) in meta_map_map:
            if isinstance(obj,mclass):
                obj_dict = {'__class__' : 'mapWOC_%s' % name}
                for prop in props:
                    obj_dict[prop] = obj.__getattribute__(prop)
                return obj_dict
        # Else, use handcraft.
        if isinstance(python_object, URLList):
            assert("URLList is not meant for transport")
        elif isinstance(python_object, ScanConfiguration):
            assert("ScanConfiguration is not meant for transport")
        else:
            return json.JSONEncoder.default(self, python_object)

def mapwoc_from_json_hook(json_object):
    """
    Object Hook to construct mapWOC objects from JSON.
    @param json_object: JSON-string of which to create a python object.
    @return: An object that is created from the json string.
    """

    objClass = json_object.get('__class__')
    if not objClass: return json_object

    for (mclass,name,props) in meta_map_map:
        if objClass == 'mapWOC_%s' % name:
            obj = mclass()
            for prop in props:
                obj.__setattr__(prop, json_object[prop])
            return obj

    # Otherwise, could do it by hand.
    return json_object

def mapwoc_dejson(string):
    """
    Convinience shortcut to decode a json string.
    @param string: JSON-string of which to create a python object.
    @return: An object that is created from the json string.
    """
    return json.loads(string, object_hook=mapwoc_from_json_hook)
