# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Database Integrity checks.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import traceback

from sqlalchemy import *
from sqlalchemy.orm import create_session

from mapwoc.lib.model.metatable import meta_table

from mapwoc.lib import *
from mapwoc.lib.result import ResultType, RESULT_TYPES
from mapwoc.lib.url import *
from mapwoc.lib.model import *
from mapwoc.lib.model.result import *
from mapwoc.lib.model.url import *


class DBCheck(object):
    """
    Basic Database Check with tons of space for optimizations and completions.
    """
    def __init__(self, dbfile, fix=True):
        self.dbfile = dbfile
        self.fix = fix

    def all_urllists_correct_numurls(self):
        """count(Visit.urls) constraint."""
        engine = create_engine('sqlite:///%s' % self.dbfile)
        c = engine.connect()
        # TODO update metadata
        dbsession = create_session(bind=engine)
        print "Checking for correct number of numurls."
        for urllist in dbsession.query(URLList).all():
            if urllist.numurls != len(urllist.urls):
                print "URLList '%s' has incorrect numurls set (%d, correct would be: %d)." % (urllist.name, urllist.numurls, len(urllist.urls))
                if self.fix:
                    print "Fixing."
                    urllist.numurls = len(urllist.urls)
                    dbsession.flush()
        print " - done."

    def all_visit_numresults_correct(self):
        """count(Visit.results) constraints."""
        engine = create_engine('sqlite:///%s' % self.dbfile)
        c = engine.connect()
        # TODO update metadata
        dbsession = create_session(bind=engine)
        print "Checking for correct number of numresults."
        cnt = 0
        # TODO The plain ORM-approach is way too slow here.
        #      Custom sql statements or improved sqla-queries are needed.
        for visit in dbsession.query(Visit).yield_per(20):
            lenresults = len(visit.results)
            r = dbsession.execute("select count() from result where result.visit_id = %d; " % visit.id).first()
            print r[0]
            if visit.numresults != lenresults:
                print "Visit '%d' has incorrect numresults set (%d, correct would be: %d)." % (visit.id, visit.numresults, lenresults)
                if self.fix:
                    cnt += 1
                    print "Fixing."
                    visit.numresults = lenresults
                    #if cnt == 100:
                    #    print "commit"
                    #    #dbsession.commit()
                    #    dnt = 0
            dbsession.flush()
        dbsession.flush()

    def all_visit_highest_sev_correct(self):
        """Visit.highest_severity constraints."""
        engine = create_engine('sqlite:///%s' % self.dbfile)
        c = engine.connect()
        # TODO update metadata
        print "Checking for correct highest severities."
        dbsession = create_session(bind=engine)
        for visit in dbsession.query(Visit).all():
            r = dbsession.execute("select result_type.severity, result_type.id from result_type, result where result_type.id = result.result_type_id and result.visit_id = %d  order by result_type.severity desc;" % visit.id).first()
            if r and r[0] is not None and visit.severest_result_type_id != r[1]:
                print "Visit '%d' has incorrect highest severity set (%s, correct would be: %d)." % (visit.id, visit.severest_result_type_id, r[1])
                if self.fix:
                    print "Fixing."
                    visit.severest_result_type_id = r[1]
                    dbsession.flush()

    def migrate(self, dbfrom, dbto):
        """
        Migrate database from version A to version B.
        only supports migrations for versions >= 4.
        @param dbfrom: Version from which to migrate.
        @param dbto: Version to which to migrate.
        @return: True if successfull, False otherwise.
        """
        if dbfrom == dbto:
            return True
        # Raise to version 5.
        if dbfrom == 4:
            engine = create_engine('sqlite:///%s' % self.dbfile)
            c = engine.connect()
            dbconn = engine.connect()
            print("New Column 'numurls' in urllist table.")
            p = dbconn.execute("ALTER TABLE urllist ADD numurls INTEGER DEFAULT 0 NOT NULL; ")
            # Increase version
            # let exceptions pass.
            session = create_session(bind=engine)
            stmt = meta_table.update(bind=session.bind)
            stmt.execute(version=5)
            return self.migrate(5, dbto)
        elif dbfrom == 5:
            engine = create_engine('sqlite:///%s' % self.dbfile)
            c = engine.connect()
            dbconn = engine.connect()
            session = create_session(bind=engine)
            print "New columns 'numresults' and 'severest_result_type_id' in urllist table."
            sql="""
            BEGIN TRANSACTION;
            CREATE TEMPORARY TABLE visit_backup(id,hcsid,node,filepath,outgoing_ip,date,scanrun_id,resolved_ips);
            INSERT INTO visit_backup SELECT id,hcsid,node,filepath,outgoing_ip,date,scanrun_id,resolved_ips FROM visit;
            DROP TABLE visit;
            CREATE TABLE visit (
                                id INTEGER NOT NULL,
                                hcsid VARCHAR NOT NULL,
                                node VARCHAR NOT NULL,
                                filepath VARCHAR,
                                outgoing_ip VARCHAR,
                                date DATETIME,
                                scanrun_id INTEGER NOT NULL,
                                severest_result_type_id INTEGER,
                                numresults INTEGER,
                                resolved_ips VARCHAR,
                                PRIMARY KEY (id),
                                FOREIGN KEY(scanrun_id) REFERENCES scanrun (id),
                                FOREIGN KEY(severest_result_type_id) REFERENCES result_type (id));
            INSERT INTO visit SELECT id,hcsid,node,filepath,outgoing_ip,date,scanrun_id,0,0,resolved_ips FROM visit_backup;
            DROP TABLE visit_backup;
            COMMIT;"""
            try:
                #dbconn.execute("BEGIN TRANSACTION;")
                dbconn.execute("CREATE TEMPORARY TABLE visit_backup(id,hcsid,node,filepath,outgoing_ip,date,scanrun_id,resolved_ips);")
                dbconn.execute("INSERT INTO visit_backup SELECT id,hcsid,node,filepath,outgoing_ip,date,scanrun_id,resolved_ips FROM visit;")
                dbconn.execute("DROP TABLE visit;")
                dbconn.execute("""CREATE TABLE visit (
                                id INTEGER NOT NULL,
                                hcsid VARCHAR NOT NULL,
                                node VARCHAR NOT NULL,
                                filepath VARCHAR,
                                outgoing_ip VARCHAR,
                                date DATETIME,
                                scanrun_id INTEGER NOT NULL,
                                severest_result_type_id INTEGER,
                                numresults INTEGER,
                                resolved_ips VARCHAR,
                                PRIMARY KEY (id),
                                FOREIGN KEY(scanrun_id) REFERENCES scanrun (id),
                                FOREIGN KEY(severest_result_type_id) REFERENCES result_type (id));""")
                dbconn.execute("INSERT INTO visit SELECT id,hcsid,node,filepath,outgoing_ip,date,scanrun_id,0,0,resolved_ips FROM visit_backup;")
                dbconn.execute("DROP TABLE visit_backup;")
                #dbconn.execute("COMMIT;")
                #p = dbconn.execute(sql)
            except Exception as e:
                print "Error happened while migrating database."
                print e
                # raise?
                return False
            # Populate new columns.
            # TODO
            self.all_visit_numresults_correct()
            self.all_visit_highest_sev_correct()
            # Increase db version to 6.
            stmt = meta_table.update(bind=session.bind)
            stmt.execute(version=6)
            return self.migrate(6, dbto)
        elif dbfrom == 6:
            engine = create_engine('sqlite:///%s' % self.dbfile)
            c = engine.connect()
            dbconn = engine.connect()
            session = create_session(bind=engine)
            print "New column 'enable_vnc' in scanconfiguration table."
            p = dbconn.execute("ALTER TABLE scanconfiguration ADD enable_vnc INTEGER DEFAULT 0 NOT NULL; ")
            # Increase db version to 6.
            stmt = meta_table.update(bind=session.bind)
            stmt.execute(version=7)
            return self.migrate(7, dbto)

        print "Do not know how to migrate database from version %d to version %d" % (dbfrom, dbto)
        return False

    def is_current_version(self, migrate=False):
        """Check weather the database is up to date
           @param migrate: Set True to call the migration instead of returning false
        """
        engine = create_engine('sqlite:///%s' % self.dbfile)
        c = engine.connect()
        session = create_session(bind=engine)
        dbinformation = meta_table.select(bind=session.bind).execute().fetchone()
        db_version = dbinformation.version
        from mapwocgui.websetup import DB_VERSION as current_version
        if db_version == current_version:
            print "Database is at current version (%d)" % db_version
            return True
        else:
            print "Database is not at current version (%d, most recent: %d)" % (db_version, current_version)
            if migrate:
                return self.migrate(db_version, current_version)
            return False

def check_migration(dbpath, migrate):
    dbc = DBCheck(dbpath, True)
    return dbc.is_current_version(True)

def check_db_integrity(dbpath):
    """Open a database and check whether the contained data is integer."""
    dbc = DBCheck(dbpath, False)
    if not dbc.is_current_version():
        # No sense in maintaining integrety checks for outdated
        # databases
        return False
    dbc.all_urllists_correct_numurls()
    dbc.all_visit_highest_sev_correct()
    dbc.all_visit_numresults_correct()
    return True

def check_quickscan_exists(session):
    quickscan = session.query(ScanConfiguration).filter_by(name = "QuickScan").first()
    return not (not quickscan)

def check_resulttypes_exist(session):
    return False

def check_no_open_scanruns(session):
    return False
