# -*- coding: utf-8 -*-

# mapWOC
#
# Description: SQLAlchemy/ORM-Mapper for Results.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from sqlalchemy import Table, Column, ForeignKey, Integer, String, Text, DateTime, Boolean
from sqlalchemy.orm import relationship, mapper

from mapwocgui.model.meta import Base, metadata

from mapwoc.lib.url import URL
import mapwoc.lib.model.url

from mapwoc.lib.scanconfiguration import ScanConfiguration, ScanRun
from mapwoc.lib.result import Result, ResultType, Visit, ResultFile

# M:N Relational Table.
visit_url_table = Table('visit_url', metadata,
                    Column('visit_id', ForeignKey('visit.id')),
                    Column('url_id', Integer, ForeignKey('url.id')))

# Visit Table.
visit_table = Table('visit', metadata,
                    Column('id', Integer, primary_key=True),
                    Column('hcsid', String, nullable=False),
                    Column('node', String, nullable=False),
                    Column('filepath', String),
                    Column('outgoing_ip', String),
                    Column('date', DateTime),
                    Column('scanrun_id', Integer,
                        ForeignKey('scanrun.id'), nullable=False),
                    Column('severest_result_type_id', Integer, ForeignKey('result_type.id'), nullable=True),
                    Column('numresults', Integer),
                    Column('resolved_ips', String))

# Result Type Table.
result_type_table = Table('result_type', metadata,
                        Column('id', Integer, primary_key=True),
                        Column('type_name', String, nullable=False),
                        Column('severity', Integer))

# Result File Table.
result_file_table = Table('result_file', metadata,
                        Column('id', Integer, primary_key=True),
                        Column('filepath', String, nullable=False),
                        Column('filetype', String),
                        Column('analysed_type', String),
                        Column('mimetype', String),
                        Column('sha1sum', String, nullable=False),
                        Column('ssdeep', String))

# Result Table.
result_table = Table('result', metadata,
                  Column('id', Integer, primary_key=True),
                  Column('url', String),
                  Column('iframe', String),
                  Column('date', DateTime, nullable=False),
                  Column('result_type_id', Integer, ForeignKey('result_type.id'), nullable=False),
                  Column('result_file_id', Integer, ForeignKey('result_file.id'), nullable=True),
                  Column('visit_id', Integer, ForeignKey('visit.id'), nullable=False),
                  Column('text', Text),
                  Column('method', String),
                  Column('ip', String))

# The Mappers
mapper(ResultType, result_type_table)
mapper(ResultFile, result_file_table)

result_props = {'result_type':relationship(ResultType),
                'result_file':relationship(ResultFile),
                'visit':relationship(Visit)}
mapper(Result, result_table, properties=result_props)

visit_properties = {'results':relationship(Result,
                                           uselist=True),
                    'scan_run':relationship(ScanRun),
                    'severest_result_type':relationship(ResultType),
                    'urls':relationship(URL,
                                        secondary=visit_url_table,
                                        backref='visit',
                                        uselist=True)}
mapper(Visit, visit_table, properties=visit_properties)

