# -*- coding: utf-8 -*-

# mapWOC
#
# Description: SQLAlchemy/ORM-Mapper for URLs and URLLists.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer
from sqlalchemy import String, Table, Text
from sqlalchemy import func
from sqlalchemy.orm import relationship, mapper

from mapwocgui.model.meta import Base, metadata

from mapwoc.lib.url import URL, URLList

# M:N Relational Table.
urllist_url = Table('urllist_url', metadata,
                    Column('urllist_id', ForeignKey('urllist.id')),
                    Column('url_id', Integer, ForeignKey('url.id')))

# URL Table.
url_table = Table('url', metadata,
                  Column('id', Integer, primary_key=True),
                  Column('url', String, nullable=False))

# URL-List Table
urllist_table = Table('urllist', metadata,
                      Column('id', Integer, primary_key=True),
                      Column('name', String, nullable=False),
                      Column('comment', Text, nullable=True),
                      Column('date', DateTime, nullable=True),
                      Column('numurls', Integer, nullable=False),
                      Column('deleted', Boolean, nullable=True))

# The Mappers
mapper(URL, url_table)
# TODO The backreffed 'urllist' is a list, thus 'urllists' (plural) would be
# a better name.
urllist_map_properties = {'urls':relationship(URL, secondary=urllist_url,
                                              backref='urllist')}
mapper(URLList, urllist_table, properties=urllist_map_properties)


def insert_list(Session, url_set, urllist_name, urllist_comment=''):
    """
    Create big URLList.

    @param Session: The Session of which to use the bind.
    @param url_set: Set of valid URLs.
    @param urllist_name: Name of the URLList to create.
    @param urllist_comment: Optional comment to give to the created URLList.
    """

    # TODO There is some interest in debugging this
    #      part to allow speed improvements. Consider
    #      logging.

    # Create an (empty) URL-List.
    # Dissociate already existing and new urls.
    # Insert urls and retrieve IDs of inserted rows.
    # Populate the many-to-many "secondary" table with connections
    # between the urllist and the respective urls.

    urllist = URLList(urllist_name)
    urllist.comment = urllist_comment
    urllist.deleted = 1
    urllist.numurls = 0

    mapped_urls = []
    unmapped_urls = []

    # Dissociate already existing and 'new' URLs.
    # It is unfair, but we are fastest by loading all URLs into mem and
    # index them with a python dictionary.

    # TODO To improve scalability e.g. an in-mem berkleydb could be used.
    stmt = Session.execute(url_table.select())

    # We assume that more URLs are in DB than in uploaded file.
    dix = {}
    while True:
        chunk = stmt.fetchmany(2048)
        if not chunk: break
        for r in chunk:
            dix[r[1]] = r[0]

    idx = 0
    for urlstr in url_set:
        # TODO benchmark url in dix
        urlid = dix.get(urlstr)
        if urlid:
            mapped_urls.append((idx,urlid))
        else:
            unmapped_urls.append((idx,urlstr))
        idx = idx + 1

    # Add the URLList.
    Session.add(urllist)
    Session.commit()

    unmapped_ids = []

    # Add URLs.
    # Unfortunately, we cannot retrieve the added IDs at the same time, thus
    # we have to rely on heuristics.
    url_dicts = []
    for (__uidx,u_url) in unmapped_urls:
        url_dicts.append(dict(url=u_url))
    if len(url_dicts):
        # Last id before inserts
        lastid_before = Session.execute(func.max(URL.id).select()).first()[0]

        # TODO Insert in batches.
        st = url_table.insert(bind=Session.bind)
        r = st.execute(url_dicts)

        lastid_after = Session.execute(func.max(URL.id).select()).first()[0]
        if len(unmapped_urls) == lastid_after - lastid_before:
            unmapped_ids = range(lastid_before + 1, lastid_after + 1)
        else:
            # New IDs are not in one continues block.
            # TODO optimize this case (e.g. bisect and check where the "wholes" are)
            pass

    # Uncomfortable backup-solution, when ids of inserted urls do not
    # form a continous block.
    # Due to impossibility to find ids with mass-inserts.
    if not len(unmapped_ids):
        stmt = Session.execute(url_table.select())

        dix = {}
        # We assume that more URLs are in DB than in uploaded file.
        while True:
            chunk = stmt.fetchmany(2048)
            if not chunk: break
            for r in chunk:
                dix[r[1]] = r[0]

        for uidx,urlstr in unmapped_urls:
            unmapped_ids.append(dix.get(urlstr))
        # Other possibility is too slow (query with "in")
        #  r = Session.execute(url_table.select(bind=Session.bind, whereclause=url_table.c.url.in_(_slice)))

    # Add m:n linkage.
    # Unfortunately st.execute does not behave with generators.
    u_uid_dicts = []
    #for u_url_id in mapped_urls:
    #    u_uid_dicts.append(dict(url_id=u_url_id,urllist_id=urllist.id))
    #for urlid in unmapped_ids:
    #    u_uid_dicts.append(dict(url_id=urlid,urllist_id=urllist.id))

    # "Merge" the two sets again according to initial order.
    idx = 0
    ud_iter = iter(unmapped_ids)
    un_iter = iter(unmapped_urls)
    ma_iter = iter(mapped_urls)
    try:
        (idx_un,un_currur) = un_iter.next()
        un_currid = ud_iter.next()
    except StopIteration:
        idx_un = len(url_set) + 1
    try:
        (idx_ma,ma_currid) = ma_iter.next()
    except StopIteration:
        idx_ma = len(url_set) + 1

    while idx < len(url_set):
        if idx_un < idx_ma:
            u_uid_dicts.append(dict(url_id=un_currid,urllist_id=urllist.id))
            try:
                (idx_un,un_currur) = un_iter.next()
                un_currid = ud_iter.next()
            except:
                idx_un = len(url_set) + 1
        else:
            u_uid_dicts.append(dict(url_id=ma_currid,urllist_id=urllist.id))
            try:
                (idx_ma,ma_currid) = ma_iter.next()
            except:
                idx_ma = len(url_set) + 1
        idx = idx + 1

    st = urllist_url.insert(bind=Session.bind)

    # TODO Insert in batches.
    r = st.execute(u_uid_dicts)
    urllist.numurls = len(u_uid_dicts)

    # Undelete URL-List
    urllist.deleted = 0
    Session.flush()
    Session.commit()

