# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Specialized Queues for mapwoc.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechni
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import multiprocessing
from multiprocessing import Queue

class TaskQueuePair(object):
    """
    TaskQueuePair is used in situations where a single producer
    hands tasks to possible many consumers and checks for results
    from time to time.
    If a consumer puts an exception in its results queue, it is raised
    in get_result (think of raising exceptions through queues).
    """

    def __init__(self):
        self.__commands = multiprocessing.Queue()
        self.__results = multiprocessing.Queue()

    def pending(self):
        """
        """
        raise NotImplemented

    def add_task(self, command, timeout=-1):
        """
        Add an entry to the commands Queue.
        Can raise Errors.
        @param command: Command to add to commands Queue.
        @param timeout: Time to wait maximally.
        """
        if timeout == -1:
            self.__commands.put(command)
        else:
            self.__commands.put(command, True, timeout)

    def get_command(self, timeout=-1):
        """
        Get an entry from the commands Queue.
        Can raise Errors.
        @param timeout: Time to wait maximally (None to wait forever, -1 be quick).
        @return: Entry from commands Queue.
        """
        c = None
        if timeout == -1:
            c = self.__commands.get_nowait()
        else:
            c = self.__commands.get(True, timeout)
        return c

    def add_result(self, result, timeout=-1):
        """
        Adds a result to Queue, blocking if timeout >= 0/None.
        @param result: Entry to add to results queue.
        @param timeout: Timeout (if any).
        """
        if timeout == -1:
            self.__results.put_nowait(result)
        else:
            self.__results.put(result, True, timeout)

    def add_err(self, err, timeout=0):
        """
        Adds an error to the result queue.
        This call will block by default (adjust timeout parameter to avoid).
        @param err: The error to put to the result queue.
        @param timeout: Timeout if any.
        """
        self.add_result(err, timeout)

    def get_result(self, timeout=-1):
        """
        Get an entry from the results Queue.
        Can raise Errors.
        Raises errors that were put in the result queue.
        @param timeout: Time to wait maximally.
        @return: Entry from results Queue.
        """
        r = None
        if timeout == -1:
            r = self.__results.get_nowait()
        else:
            r = self.__results.get(True, timeout)
        if isinstance(r, BaseException):
            raise r
        else:
            return r

    def get_all_results(self, timeout=-1):
        """
        Get all results from result queue, raising exceptions as they cross
        our way.
        @param timeout: Timeout, if any.
        @return: List of contents of result queue.
        """
        rs = []
        while not self.__results.empty():
            rs.append(self.get_result(timeout))
        return rs


    def all_empty(self):
        """
        @return: True if both queues empty.
        """
        return self.__results.empty() and self.__commands.empty()


    def len_cmds(self):
        """
        @return: Number of unprocessed commands.
        """
        return self.__commands.qsize()

    def len_results(self):
        """
        @return: Number of unfetched results.
        """
        return self.__results.qsize()
