# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process to do something regularly and communicate results
#              through a queue.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from multiprocessing import Process, Queue
from time import sleep


class TickProcess(Process):
    """
    Call tick(), sleep(), tick(), sleep().
    """

    def __init__(self, queue, tick_interval=60):
        """
        Sets up (but doesnt start) a Process to periodically call tick(),
        which is not implemented in this class.


        @param queue: A Queue that can be filled up in tick() and
                      (asynchronously) been queried from "outside".
        @param tick_interval: The interval in seconds to wait between the
                              end of one tick()-call and the next.
        """
        Process.__init__(self)
        self.__tick_interval = tick_interval
        self.queue = queue

    def tick(self):
        """
        Place to put the meat.
        Note that we live in a forked environment here, so
        changes made to self are only visible from within tick().
        """
        raise NotImplementedError()

    def put(self, stuff):
        """
        Put stuff into queue (blocking).
        @param stuff: Stuff to put into queue.
        """
        self.queue.put(stuff)

    def run(self):
        """
        Calls tick(), sleep(tick_interval), tick()...
        """
        while True:
            self.tick()
            sleep(self.__tick_interval)
