# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Convenience- methods and constants for mapwoc applications.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import logging
from optparse import OptionParser
from optparse import SUPPRESS_HELP
from os.path import isfile

BSI_COPYRIGHT = ("Copyright (C) 2011 Bundesamt für"
                 " Sicherheit in der Informationstechnik, BSI")
VERSION_STATEMENT = ("mapwoc comes with ABSOLUTELY NO WARRANTY;\n"
                     "This is free software, and you are welcome to\n"
                     "redistribute it under certain conditions (GPLv2+).")

def print_version_exit(tool, version):
    """
    Print tool name, version, copyright, free software
    statement and exit(0).
    """
    print "%s version %s, %s\n%s" % (tool, version, BSI_COPYRIGHT,
                                     VERSION_STATEMENT)
    exit(0)

def setup_logger(quiet, verbose, loggername, logfile):
    """Setup logging."""

    logger = logging.getLogger(loggername)
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")

    if verbose:
        min_level = logging.DEBUG
    elif quiet:
        min_level = logging.WARNING
    else:
        min_level = logging.INFO

    lfh = logging.FileHandler(logfile)
    lsh = logging.StreamHandler()
    lfh.setFormatter(formatter)
    lsh.setFormatter(formatter)

    logger.addHandler(lfh)
    logger.addHandler(lsh)
    logger.setLevel(min_level)

"""
To ease cross-application consistent behaviour with respect to
configuration files and command line (configuration) options,
the mapwoc.lib.toolutil- module provides two classes to enable
the applications to be configured.

An option is specified like:
(short_option, long_option, description, default, type) .
"""


class MapwocOptionParser(OptionParser):
    COMBIOPT_TYPE_INT = 'int'
    COMBIOPT_TYPE_FLOAT = 'float'
    COMBIOPT_TYPE_STRING = 'string'
    COMBIOPT_TYPE_MAKETRUE = 'settrue'

    def __init__(self, version, description,
                 default_config_file_path,
                 combiopts=[], allow_gen=True):
        """
        Initialalizes internal OptionParser.
        @param version: Version String to display when queried (--version).
        @param default_config_file_path: Path to the (default) configuration
                                         file.
        @param combiopts: list of combiopts.
        @param allow_gen: Whether or not to allow to create a config file
                          with --generate-config.
        """
        OptionParser.__init__(self,
                              version=version,
                              description=description)
        self.combiopts = combiopts
        if not self.combiopts:
            self.combiopts = []
        self.default_config_file_path = default_config_file_path
        self.combiopts.append(('', 'config', 'Path to configuration file to read configuration from (or write to if called with --generate-config).',
                               default_config_file_path, self.COMBIOPT_TYPE_STRING),)
        if allow_gen:
            self.combiopts.append(('', 'generate-config', 'Generate a configuration file.',
                                   False , self.COMBIOPT_TYPE_MAKETRUE),)

        self.inject_combiopts(self.combiopts)

    def inject_combiopts(self, combiopts):
        """
        Inject options into the OptionParser.
        @param combiopts: List of option-tuples.
        """
        for (oshort,olong,odesc,odef,otype) in combiopts:
            short_opt = oshort
            if short_opt and short_opt != '':
                short_opt = '-%s' % short_opt
            long_opt = '--%s' % olong
            if otype == self.COMBIOPT_TYPE_INT:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc,
                                type=int)
            elif otype == self.COMBIOPT_TYPE_FLOAT:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc,
                                type=float)
            elif otype == self.COMBIOPT_TYPE_MAKETRUE:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc,
                                action="store_true")
            else:
                self.add_option(short_opt, long_opt, dest=olong, help=odesc)

    def parse_args(self):
        self.add_option('', '--man', dest="man", help=SUPPRESS_HELP)
        (options, args) = OptionParser.parse_args(self)

        if getattr(options,"man"):
            if options.man == 'conf':
                print self.to_nroff_config()
            else:
                print self.to_nroff_cmd()
            exit(0)

        # Config items given at cmd override config options given in file,
        # which override defaults.
        conf_defaults = {}
        for (__oshort,olong,__odesc,odef,__otype) in self.combiopts:
            conf_defaults[olong] = odef

        config = ConfigParser.RawConfigParser(defaults=conf_defaults)
        if options.config == None:
            options.config = self.default_config_file_path
        if not isfile(options.config):
            if getattr(options, 'generate-config'):
                print "Generating a config in %s" % options.config
                if not config.has_section('config'):
                    config.add_section('config')
                for (oshort,olong,odesc,odef,otype) in self.combiopts:
                    config.set('config', olong, odef)
                optf = open(options.config, 'w')
                config.write(optf)
                print "Config file written."
                exit(4)
            else:
                print "Configuration file %s does not exist." % options.config
                exit(3)
        try:
            config.read(options.config)
            if not config.has_section('config'):
                config.add_section('config')

            for (oshort,olong,odesc,odef,otype) in self.combiopts:
                cmd_opt_val = getattr(options, olong)
                if not cmd_opt_val:
                    if otype == self.COMBIOPT_TYPE_INT:
                        cfg_opt_val = config.getint('config', olong)
                    elif otype == self.COMBIOPT_TYPE_FLOAT:
                        cfg_opt_val = config.getfloat('config', olong)
                    else:
                        cfg_opt_val = config.get('config', olong)
                    setattr(options, olong, cfg_opt_val)
        except ConfigParser.NoSectionError:
            print "Given configuration file %s is invalid or missing option." % options.config
            exit(2)
        except TypeError:
            print "Error parsing the options, does the config file exist?"
            exit(3)

        return (options, args)

    def to_nroff_config(self):
        """
        Returns a basic man page for the configuration file.
        """
        out = ".TH %s.config 5" % self.cmd
        out += ".SH NAME\n"
        out += "%s.config \- configuration file for %s\n" % (self.cmd, self.cmd)

        out += ".SH DESCRIPTION\n"
        out += "All command-line options available to the %s program can also be specified in a configuration file.\n" % self.cmd
        out += "The format of the file is as follows.\n"
        out += "\n"
        out += " [config]\n"
        out += " option1=value1\n"
        out += " option2=value2\n"

        out += ".SH OPTIONS\n"
        for (oshort,olong,odesc,odef,__otype) in self.combiopts:
            if oshort:
                out += ".IP -%s,--%s\n%s\n" % (oshort, olong, odesc)
            else:
                out += ".IP --%s\n%s\n" % (olong, odesc)

        out += ".SH SEE ALSO\n"
        out += ".BR mapwoc-master(1)\n.BR mapwoc-redirector(1)\n.BR mapwoc-node(1)\n"
        out += ".BR mapwoc-master.config(5)\n.BR mapwoc-redirector.config(5)\n.BR mapwoc-node.config(5)\n"

        out += ".SH MORE INFORMATION ABOUT THE MAPWOC PROJECT\n"
        out += "The canonical places where you will find more information about the mapwoc project are: \fBhttp://www.mapwoc.org/\f1 (Official site)\n"

        out += ".SH COPYRIGHT\n"
        out += "mapwoc is released under the GNU GPL, version 2, or, at your option, any later version.\n"
        out += "Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik"
        return out

    def to_nroff_cmd(self):
        """
        Returns a basic man page, based on combiopts, the cmd, description
        and summary fields.
        @return: Basic manpage (use e.g. with groff).
        """
        out = ".TH %s 1 \"User Manuals\"\n" % self.cmd
        out += ".SH NAME\n"
        out += "%s \- %s\n" % (self.cmd, self.summary)

        out += ".SH SYNOPSIS\n"
        out += ".B mapwoc-node [OPTIONS]\n"

        out += ".SH DESCRIPTION\n"
        out += "%s\n" % self.description

        out += ".SH OPTIONS\n"
        for (oshort,olong,odesc,odef,__otype) in self.combiopts:
            if oshort:
                out += ".IP -%s,--%s\n%s\n" % (oshort, olong, odesc)
            else:
                out += ".IP --%s\n%s\n" % (olong, odesc)
        out += ".IP -h,--help\nShow help and exit.\n"
        out += ".IP --version\nShow version and exit.\n"

        out += ".SH FILES\n"
        out += ".I /etc/mapwoc/%s.config\n" % self.cmd
        out += ".RS\n"
        out += "The configuration file for %s, in grouped key=value-format.\n" % self.cmd
        out += "e.g.\n"
        out += " [config]\n"
        out += " verbose=True\n"
        out += ".RE\n"
        out += ".RS\nAll of the above options (except for --help, ---version, --config and --generate-config)  might be present in their long form in a configuration file for %s.\n" % self.cmd
        out += "In case of conflicts, information given at command line wins.\n.RE\n"


        out += ".SH EXAMPLES\n"
        out += "%s --config /home/user/.mapwoc.config --verbose\n" % self.cmd
        out += "Start with options given in the respective configuration file, be verbose..\n"

        out += ".SH SEE ALSO\n"
        out += ".BR mapwoc-master(1)\n.BR mapwoc-redirector(1)\n.BR mapwoc-node(1)\n"
        out += ".BR mapwoc-master.config(5)\n.BR mapwoc-redirector.config(5)\n.BR mapwoc-node.config(5)\n"

        out += ".SH MORE INFORMATION ABOUT THE MAPWOC PROJECT\n"
        out += "The canonical places where you will find more information about the mapwoc project are: \fBhttp://www.mapwoc.org/\f1 (Official site)\n"

        out += ".SH COPYRIGHT\n"
        out += "mapwoc is released under the GNU GPL, version 2, or, at your option, any later version.\n"
        out += "Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik"
        return out
