# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Master-centric classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import datetime
import json
import logging
from multiprocessing import Process, Queue
from os import listdir, path, makedirs
from Queue import Empty
import subprocess
import socket
import time
from time import sleep

from sqlalchemy import *

from mapwocgui.model.meta import Session

from mapwoc.lib.trivialresponse import TrivialResponse
from mapwoc.lib.message import Message
from mapwoc.lib.mailhelper import MailDispatcher
from mapwoc.lib.users import UserData
from mapwoc.lib.hc import HC
from mapwoc.lib.node import Node
from mapwoc.lib.scanconfiguration import ScanConfiguration
from mapwoc.lib.scanconfiguration import ScanRun
from mapwoc.lib.scanconfiguration import TransportableSC
from mapwoc.lib.scanconfiguration import VisitorConfiguration
from mapwoc.lib.scanconfiguration import VisitorBundle

from nodeconnection import NodeConnection
from nodepollprocess import NodePollProcess
from redirectorpollprocess import RedirectorPollProcess
from resultpollprocess import ResultPollProcess
from mapwoc.masterctrl.masterstate import MasterState

from mapwoc.lib.listdict import listdict

def handleError(self, record):
  raise
logging.Handler.handleError = handleError

logger = logging.getLogger('master')

class CannotStartScanError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


def load_node_cfgs(confdir, cred):
    logger.info("Loading nodes from %s." % confdir)

    # Expand and normalize all paths.
    abs_file_paths = [ path.normpath(path.join(confdir, file_rel))
                       for file_rel in listdir(confdir) ]
    # Just keep the ones ending in .conf .
    abs_conf_files = [ file_abs for file_abs in abs_file_paths
                       if file_abs.endswith(".config") ]

    nodes = {}

    # Parse the remaining ones and add to list of nodes.
    for node_conf_file in abs_conf_files:
        node = NodeConnection(node_conf_file, cred)
        if nodes.get(node.key) != None:
            logger.error("Node with name %s is defined multiple times, taking %s." % (node.key, node_conf_file))
        nodes[node.key] = node
        node.data_path = node.cmd_get_resultdir()
        if not node.data_path.startswith('/'):
            logger.error("Could not get resultsdir from node %s " % node.key)
            node.data_path = None

    logger.info("Loaded %d nodes." % len(nodes))
    return nodes

class Master:
    """A Master."""

    VERSION = "1.3.3"

    def __init__(self, configfile, credentials_path,
                 work_dir, db, redirectorproxy, nodes, dry=False,
                 severity_tolerance=0):
        """
        Init fields to empty lists and reads configuration file.
        @param configfile: Path to the configuration file.
        @param nodes: Dictionary with NodeConnections.
        @param credentials_path: Path to the credentials to use when polling
                                 results from nodes.
        @param work_dir: Path to working directory.
        @param db: Path to database.
        @param redirectorproxy: RedirectorProyx (for commands).
        @param dry: Set to True to keep subprocesses unstarted (to debug).
        @param severity_tolerance: Mimumum value of results to store.
        """

        self.__nodes = nodes # Mapping node keys (names) to NodeConnections
        self.messages = [] # List of Messages
        self.__hcs = [] # List of HC's
        self.__scan = None # ScanRun Instance

        self.__read_conf(configfile)
        self.credentials_path = credentials_path
        self.__work_dir = work_dir
        self.__db_file = db
        self.redirectorproxy = redirectorproxy
        self.__sync_hcs()

        # Setup and start child processes that poll the other services.
        self.node_queue = Queue()
        self.node_poll_process = NodePollProcess(self.node_queue,
                                                 self.__nodes,
                                                 5)
        if not dry:
            self.node_poll_process.start()

        self.redirector_queue = Queue()
        self.redirector_poll_process = RedirectorPollProcess(
                                            self.redirector_queue,
                                            5,
                                            self.redirectorproxy
                                            )
        if not dry:
            self.redirector_poll_process.start()

        if not path.isdir(self.__work_dir):
            makedirs(self.__work_dir)

        engine = create_engine('sqlite:///%s' % self.__db_file)
        __c = engine.connect()
        Session.configure(bind=engine)

        self.resultsyncpoller = ResultPollProcess(self.__work_dir,
                                                  self.__nodes, Queue(),
                                                  self.credentials_path,
                                                  self.__db_file,
                                                  severity_tolerance)
        if not dry:
            self.resultsyncpoller.start()
        self.__state = MasterState()

    def __del__(self):
        """Deconstructor. Take care of child processes, open handles etc."""

        #TODO terminate and join the poll processes.
        pass

    def __update_redirector_state(self):
        """Sets state_set['redirector'] and places a Message."""
        try:
            # Get last answer to 'state' question of redirector.
            redirector_answer = self.redirector_queue.get_nowait()
            while not self.redirector_queue.empty():
                redirector_answer = self.redirector_queue.get_nowait()

            if isinstance(redirector_answer,Exception):
                raise redirector_answer
            else:
                self.__state.of_redirector_set(redirector_answer)
        except Empty:
            logger.debug("No update from redirector state poll.")
            self.__state.of_redirector_set(None)
        except Exception as e:
            logger.warning("Exception while getting the redirector state (%s)." % e)
            self.__state.of_redirector_set(None)

    def __update_node_state(self):
        """Sets state_set['nodes']."""
        try:
           node_answer = self.node_queue.get_nowait()
           while not self.node_queue.empty():
               node_answer = self.node_queue.get_nowait()
           self.__state.of_nodes_set(node_answer)
        except Empty:
            logger.debug("No update from nodes state poll yet.")


    def tick(self):
        """
        Heart of a mapwoc-master.
        Checks the Queues which are communication channels to the processes
        that ask related services about their state etc.
        Update state and act according to state(s).
        This method is to be called regularly (e.g. by a twisted engine).
        """

        self.__update_redirector_state()
        self.__update_node_state()

        self.__update_state()

    def __update_status_scan(self):
        """Set self.__state.scanname."""

        if self.__scan:
            self.__state.scanname = self.__scan.scanconfiguration.name
        else:
            self.__state.scanname = None

    def __stopall_smooth(self):
        """
        Ask all nodes to stop gracefully.
        """
        for _name,node in self.__nodes.items():
            node.cmd_stopsmooth()
        self.__state.set(MasterState.STOPPING)

    def __go_idle(self):
        """
        Set scan to None and state to IDLE.
        """
        # TODO Tell redirector to idle. Listens to what it says.
        try:
            ra = self.redirectorproxy.cmd_idle()
        except:
            logger.error("Couldnt ask redirector to go idle.")
        self.__scan = None
        self.__state.set(MasterState.IDLE)

    def __scan_done(self):
        """
        Called when a scan should be set to done (successfull finish).
        """
        scanname = self.__scan.scanconfiguration.name
        msg = Message(Message.LEVEL_FUNCTIONAL_LOG,
                      "Scan %s done" % scanname,
                      heading="Scan")
        self.messages.append(msg)
        self.__send_scan_finished(scanname)
        # Update ScanRun in database.
        self.__scan.stoptime = datetime.datetime.now()
        self.__scan.status = "done"
        Session.add(self.__scan)
        Session.commit()

    def __update_state_scanning(self):
        """
        Act upon remote states and update own state.
        """

        is_time_scan = self.__scan.scanconfiguration.is_timed_scan()
	if self.__state.redirectorstate:
	    is_redirector_done = self.__state.redirectorstate.get('state') == 'done'
        else:
            is_redirector_done = False
        nodes_idle = self.__state.are_nodes_idle()

        # Timed scan (ignorant wrt redirector).
        if is_time_scan:
            started = self.__scan.starttime
            duration = datetime.datetime.now() - started
            max_duration = datetime.timedelta(seconds=abs(
                            self.__scan.scanconfiguration.visitor_configuration.active_window))

            if duration > max_duration:
                logger.debug("Stopping timed scan.")
                self.__stopall_smooth()
            if self.__state.of_nodes_is_valid and nodes_idle:
                self.__scan_done()
                self.__go_idle()
                logger.debug("Go idle (nodes are idle)")
        elif is_redirector_done:
            # Not Time Scan.
            if nodes_idle:
                self.__scan_done()
                self.__go_idle()
                logger.debug("Go idle (r. is idle, nodes are idle)")
            else:
                for _name,node in self.__nodes.items():
                    node.cmd_stopsmooth()
                self.__state.set(MasterState.STOPPING)
        elif self.__state.state == MasterState.STOPPING and nodes_idle:
            # Tell world that scan is aborted.
            self.messages.append(Message(Message.LEVEL_FUNCTIONAL_LOG,
                                 "Scan %s aborted" % self.__scan.scanconfiguration.name,
                                 heading="Scan"))
            self.__scan.stoptime = datetime.datetime.now()
            self.__scan.status = "aborted"
            Session.add(self.__scan)
            Session.commit()
            self.__go_idle()
        elif self.__state.state != MasterState.STOPPING and not nodes_idle:
            # Nodes are scanning and we are not stopping
            self.__state.set(MasterState.SCANNING)
        elif self.__state.state == MasterState.STOPPING and not nodes_idle:
            # Nodes are running but we are stopping.
            logger.debug("Gently pushing nodes to stop (again).")
        else:
            logger.warning("State not associated with action (%s)." % str(self.state()))

    def __stop_all(self):
        """
        Tell all known components to stop/idle.
        Do not report errors in log, fo not throw exceptions.
        """
        try:
            self.redirector.cmd_idle()
        except:
            pass
        for _key, node in self.__nodes.items():
            try:
                node.cmd_stoprun()
            except:
                pass


    def __update_state(self):
        """
        Act upon state of all components.
        Note that the state information might be outdated.
        """

        if not self.__state.of_nodes_is_valid():
            logger.info("Invalidated node states.")
            #TODO Should we wait for next poll?
            #return

        # If we are in error, everybody else should idle (or error itself).
        if MasterState.ERROR == self.__state.state:
            self.__update_status_scan()
            self.__stop_all()
            # Wait for humans to help.
            return
        elif MasterState.IDLE == self.__state.state:
            self.__state.scanname = None
            err_list = self.__state.node_has_error()
            # If somebody is not idle, idle it (if its error, error ourselves).
            if self.__state.are_nodes_idle():
                if not self.__state.is_redirector_idle():
                    self.redirectorproxy.cmd_idle()
                #else: Everything is fine, really.
                return
            elif err_list:
                for (node,errmsg) in err_list:
                    msgtext = "Node %s is in error state:\n%s" % (node, errmsg)
                    self.messages.append(Message(Message.LEVEL_ERROR, msgtext,
                                                 heading="Node has error"))
                    self.__send_problem(msgtext)
                    logger.critical("Master idle, but node %s in error (%s)." % (node, errmsg))
                    self.__state.set(MasterState.ERROR)
                self.redirectorproxy.cmd_idle()
                return
            else:
                # Probably nodes are running. That should not happen.
                logger.warning("Master is in idle state, but nodes are not.")
                for _name,node in self.__nodes.items():
                    node.cmd_stoprun()
                self.redirectorproxy.cmd_idle()
                return
        elif MasterState.SCANNING == self.__state.state or MasterState.STOPPING == self.__state.state:
            err_list = self.__state.node_has_error()
            if err_list:
                self.__state.set(MasterState.ERROR)
                logger.critical("At least one node in error state, pressing red button.")
                for (node,errmsg) in err_list:
                    msgtext = "Node %s is in error state:\n%s" % (node, errmsg)
                    self.messages.append(Message(Message.LEVEL_ERROR, msgtext,
                                                 heading="Node has error"))
                    self.__send_problem(msgtext)
                # TODO or self.stop_scan()?
                for _name,node in self.__nodes.items():
                    node.cmd_stoprun()
                self.redirectorproxy.cmd_idle()
                self.__update_status_scan()
                return
            elif self.__scan is None:
                # Not Scanning anymore.
                if self.__state.of_nodes_is_valid() and not self.__state.are_nodes_idle():
                    logger.warning("Node does not idle but master has no scan!")
                if not self.__state.state == MasterState.IDLE:
                    # Become idle.
                    self.__state.set(MasterState.IDLE)
                    logger.info("Setting master state to idle.")
                self.__update_status_scan()
                logger.warning("(scan is none) Other components should idle, too.")
            else:
                # Scanning.
                self.__update_status_scan()
                self.__update_state_scanning()
        else:
            logger.critical("Masters own state is undefined.")
            self.__state.set(MasterState.ERROR)
            return

    def check_scan(self, scanconf):
        """
        Checks whether a scan could be started by the nodes.
        @param scanconf: The Scan Configuration to let the nodes check.
        @return: Triple, first value being True if all nodes could start that
                 scan.  Otherwise (False,nameofnode,messge).
        """

        nodetasks = listdict()

        # Sort single HCs into correct nodetasks pocket.
        for visitor in scanconf.visitor_configuration.visitors:
            node = self.__nodes.get(visitor.node)
            if not node:
                logger.warning("Node %s for which jobs exist not found." % \
                             visitor.node)
                continue
            nodetasks.lappend(visitor.node, visitor)

        for node, visitorlist in nodetasks.items():
            node_vconf = VisitorConfiguration()
            node_vconf.visitors = visitorlist
            node_vconf.active_window = scanconf.visitor_configuration.active_window
            n = self.__nodes.get(node)
            ok,tr = n.cmd_checkjob(node_vconf)
            if not ok:
                logger.warning("Node will not be able to do the job: %s." % tr.message)
                return (False,node,tr.message)

        return (True,None,None)

    def __node_start_scan(self, nodename, visitor_configuration,
                          visitorlist, scanconf):
        """
        Start a Scan on a node, return its response.
        @param nodename: Name of node to start scan on.
        @param visitor_configuration: Configuration to use.
        @param visitorlist: Visitorbundles to use.
        @param scanconf: The scanconfiguration to use.
        @return: Answer of node.
        """
        # TODO Further refactoring might make some of the parameters obsolete
        #      (e.g. visitor_conf and visitorlist can be taken from scanconf)
        node = self.__nodes.get(nodename)
        vconf = VisitorConfiguration()
        vconf.eventmap = visitor_configuration.eventmap
        vconf.active_window = visitor_configuration.active_window
        vconf.visitors = visitorlist
        job = TransportableSC(scanconf)
        job.visitor_configuration = vconf
        return node.cmd_startrun(self.__scan, job)

    def start_scan(self, scanconf_id):
        """
        Start a ScanRun with a given ScanConfiguration.

        Raises CannotStartScanError s.

        @param scanconf_id: ID of the ScanConfiguration to distribute to nodes
                            and redirector.
        @return: False in case of errors, True otherwise.
        """

        if self.__scan:
            logger.warning("Master can only have one scan running at a time.")
            raise CannotStartScanError("Already running a scan.")

        # TODO Respect own state.

        scanconf = Session.query(ScanConfiguration).get(scanconf_id)
        if not scanconf:
            logger.warning("Master did not find scanconf %d in database." % scanconf_id)
            raise CannotStartScanError("Scan %d not found in database." % scanconf_id)

        if scanconf is None:
            logger.warning("Given no ScanConfiguration.")
            return False
        vc = scanconf.visitor_configuration
        if vc is None:
            logger.warning("Given Scan Configuration lacks visitor configuration.")
            return False
        if not vc.visitors:
            logger.warning("Given Vistor Configuration does not have Visitor entries.")
            return False

        # Check whether all HCs are known and whether the types are not mixed.
        nr_native = 0
        nr_virtual = 0
        for visitor in vc.visitors:
            hc = self.get_hc(visitor.hcsid)
            if hc is None:
                raise CannotStartScanError("Scan contains an unknown HC (%s)." % visitor.hcsid)
            if hc.hctype == 'native':
                if nr_virtual > 0:
                    raise CannotStartScanError("Scan cannot have mixed types of HCs (native/virtual).")
                nr_native = nr_native + 1
            else:
                if nr_native > 0:
                    raise CannotStartScanError("Scan cannot have mixed types of HCs (native/virtual).")
                nr_virtual = nr_virtual + 1

        (can,node,msg) = self.check_scan(scanconf)
        if not can:
            raise CannotStartScanError(msg)

        nodetasks = listdict()
        self.__scan = ScanRun()
        self.__scan.scanconfiguration_id = scanconf.id
        self.__scan.status = "start prepared"

        Session.add(self.__scan)
        Session.commit()

        if scanconf.is_timed_scan():
            logger.debug("Start a timed scan.")
        else:
            logger.debug("Start a regular scan.")

        # TODO further unify redirector chatter.
        # Give order to redirector.
        try:
            urllistids = [u.id for u in scanconf.urlcollection]
            r_response = self.redirectorproxy.cmd_loaddb(urllistids,
                                                    path.abspath(self.__db_file))
        except Exception as e:
            logger.warning("Could not ask or decode redirectors answer"
                           " to loaddb question")
            logger.error(str(e))
            r_response = ''

        if r_response == None:
            logger.warning("panic: Redirector did not answer to loaddb question")
        if r_response.error != 0 and not self.__scan.scanconfiguration.is_timed_scan():
            logger.warning("Redirector answered %s " % r_response.message)
            return False

        # Communicate amout of URLs to scan at once.
        try:
            r_response = self.redirectorproxy.cmd_fastmode(vc.num_urls)
        except Exception as e:
            logger.warning("Could not ask redirector to go to fast mode (%s)." % e)
        logger.debug("Asking redirector to move to fastmode %d " % vc.num_urls)
        if r_response != 'ok':
            logger.warning("panic: Redirector did not acknowledged fastmode question")

        # Sort single HCs into correct nodetasks pocket.
        for visitor in vc.visitors:
            node = self.__nodes.get(visitor.node)
            if node is None:
                logger.info("Node %s for which jobs exist not found." % \
                             visitor.node)
            else:
                nodetasks.lappend(visitor.node, visitor)

        # Create visitorconfiguration, add to scanconf, send to node.
        for node, visitorlist in nodetasks.items():
            response = self.__node_start_scan(node, vc,
                                              visitorlist, scanconf)
            if response.error != 0:
                # TODO which state to go into?
                logger.warning("Node answered: %s" % response.message)
                return False
            else:
                logger.debug("Node accepted to start scan.")

        self.__state.of_nodes_invalidate()
        logger.info("Starting scan.")
        self.messages += [Message(Message.LEVEL_FUNCTIONAL_LOG,
                                  "Starting Scan %s" % scanconf.name,
                                  heading="Scan")]
        # Set State.
        self.__state.set(MasterState.SCANNING)
        self.__state.scanname = scanconf.name
        return True

    def stop_scan(self, graceful=False):
        """Stops a running scan."""

        if not self.__scan:
            logger.warning("No Scan found to stop.")
            return False

        for v in self.__scan.scanconfiguration.visitor_configuration.visitors:
            if graceful:
                response = self.__nodes[v.node].cmd_stopsmooth()
            else:
                response = self.__nodes[v.node].cmd_stoprun()

        try:
            self.redirectorproxy.cmd_idle()
        except:
            logger.warning("Couldnt ask redirector to go idle.")

        # Collect the answers from the stop questions, remove the scan config if
        # successfull. (TODO)

        self.__state.set(MasterState.STOPPING)
        return True

    def state(self):
        """Returns string representation of state."""
        return self.__state.export()

    def __sync_hcs(self):
        """
        Syncs HCs with configured nodes.
        Calling might block for a longer time.
        """
        for key, node in self.__nodes.items():
            logger.debug("Getting hc's from: %s" % key)
            hcs = node.cmd_get_hcs()
            if not hcs:
                logger.warning("Node %s reports no HC's." % key)
                continue
            for hc in hcs:
                # Check if hcid is already known
                for known in self.__hcs:
                    if known.hcsid == hc.hcsid:
                        logger.debug("Adding known hc: %s " % hc.hcsid)
                        known.on_nodes.append(key)
                        break
                else:
                    hc.on_nodes = [key]
                    self.__hcs.append(hc)

    def __read_conf(self, config_file):
        """
        Reads in configuration from file.
        @param config_file: The configuration file to read from.
        """

        config = ConfigParser.RawConfigParser()
        config.read(config_file)
        try:
            self.__maildispatcher = MailDispatcher(
                    config.get('notifications', 'smtp_server'),
                    config.get('notifications', 'smtp_port'),
                    config.get('notifications', 'admin_mail'),
                    config.get('notifications', 'sender_mail'))
        except:
            logger.info("No valid mail configuration found. "
                        "Disabling Notifications.")
            self.__maildispatcher = MailDispatcher(None, None, None, None)

    def get_node_by_name(self, name):
        """Get a Node object by it's name."""

        return self.__nodes[name]

    def get_nodes(self):
        """Accessor for nodes."""

        return self.__nodes

    def get_hcsid_on_nodes(self):
        """
        @return: A mapping from hcsid to list of nodes.
        """

        answer = {}
        for hc in self.__hcs:
            for node in hc.on_nodes:
                if answer.get(hc.hcsid):
                    answer[hc.hcsid].append(node)
                else:
                    answer[hc.hcsid] = [node]
        return answer

    def get_hc(self, hcsid):
        """
        @return: HCControl belonging to hcsid.
        """
        for hc in self.__hcs:
            if hc.hcsid == hcsid:
                return hc
        return None

    def get_last_msgs(self, num=1):
        """
        Get the num last messages as a list.

        @param num: Number of messages to fetch.
        @return: The last num messages (first item in list is last message).
        """

        msgs = self.messages[-num:]
        msgs.reverse()
        return msgs

    def __send_problem(self, problem_text, err_no=0):
        """
        Sends a email to all users who wanted to be notified on problems.
        @param problem_text: Text describing the problem.
        """
        for user_data in Session.query(UserData).all():
            if (user_data.notify_on_error == True and
                user_data.e_mail != None):
                self.__maildispatcher.send_error(
                        errornr=err_no,
                        reciever=str(user_data.e_mail),
                        errormessage=str(problem_text))

    def __send_scan_finished(self, scan_name):
        """
        Sends a email to all users who wanted to be notified on success.
        @param scan_name: Name of finished scan.
        """
        for user_data in Session.query(UserData).all():
            if ( user_data.notify_on_success == True and
                 user_data.e_mail != None ):
                self.__maildispatcher.send_scan_finished(
                        scanname=str(scan_name),
                        reciever=str(user_data.e_mail))
