# -*- coding: utf-8 -*-

# mapWOC
#
# Description: MasterCtrl with twisted.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from ConfigParser import RawConfigParser, NoSectionError
import logging
import json
import re

from twisted.internet import reactor, task
from twisted.internet.protocol import Factory, Protocol
from twisted.protocols.basic import LineReceiver

from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook, mapwoc_dejson
from mapwoc.lib.protocolutil import json_trsuccess, json_trerror
from mapwoc.lib.toolutil import MapwocOptionParser

from mapwoc.masterctrl.master import Master, CannotStartScanError

logger = logging.getLogger('masterctrlsrv')

class Requests:
    """ Request Type 'enum'."""
    NODESTATES = 1001
    STATE = 1002
    MESSAGES = 1003
    HCS = 1005
    HC = 1006
    START_SCAN = 1007
    STOP_SCAN = 1008
    NODE = 1009
    VERSION = 1010
    CHECK_SCAN = 1011
    STOP_SMOOTH = 1012

cmd_request_map = {'state': Requests.STATE,
                   'nodestates': Requests.NODESTATES,
                   'messages':Requests.MESSAGES,
                   'hcs':Requests.HCS,
                   'hc':Requests.HC,
                   'startscan':Requests.START_SCAN,
                   'stopscan':Requests.STOP_SCAN,
                   'stopsmooth':Requests.STOP_SMOOTH,
                   'version':Requests.VERSION,
                   'checkscan':Requests.CHECK_SCAN,
                   'node':Requests.NODE
                   }

def which_cmd(text):
    """Returns the Request field or None."""

    simple_cmd = cmd_request_map.get(text)

    # Try to parse more complicated command.
    if simple_cmd is None:
        # TODO why not store e.g.
        # ('startscan', 'startscan \d', Requests.START_SCAN) in map above?
        tokens = text.split(" ")
        if len(tokens) == 2 and tokens[0] == 'hc':
            return Requests.HC,[tokens[1]]
        elif len(tokens) == 2 and re.match('\d', tokens[1]) and tokens[0] == 'messages':
            return Requests.MESSAGES,[int(tokens[1])]
        elif len(tokens) == 2 and tokens[0] == 'node':
            return Requests.NODE,[tokens[1]]
        elif len(tokens) == 2 and re.match('\d', tokens[1]) and tokens[0] == 'startscan':
            return Requests.START_SCAN,[int(tokens[1])]
        else:
            return None,[]
    else:
        return simple_cmd,[]


class MasterRequest(LineReceiver):

    def connectionMade(self):
        """Inits empty buffer and command."""

        self.read_data = ''
        self.cmd = None

        session = self.transport.socket
        try:
            peer_name = session.peer_certificate.subject
        except AttributeError as err:
            peer_name = 'Unknown'
        try:
            if hasattr(session, "verify_peer"):
                session.verify_peer()
        except Exception as e:
            logger.warning("Could not verify peers certificate: %s" % str(e))
        logger.debug('Connection to master from peer "%s" (%s).' % (
                                                 peer_name, self.transport.getPeer()))

    def lineReceived(self, line):
        """React on a received line."""

        self.read_data += line

        if not self.cmd:
            self.cmd,args =  which_cmd(self.read_data.strip())
            if self.cmd is Requests.NODESTATES:
                answer = self.factory.master.query_nodes_state()
                self.__answer_json(answer)
            elif self. cmd is Requests.STATE:
                answer = self.factory.master.state()
                self.__answer_json(answer)
            elif self.cmd is Requests.MESSAGES:
                if len(args) == 0:
                    args.append(5)
                answer = self.factory.master.get_last_msgs(args[0])
                self.__answer_json(answer)
            elif self.cmd is Requests.HCS:
                answer = self.factory.master.get_hcsid_on_nodes()
                self.__answer_json(answer)
            elif self.cmd is Requests.HC:
                answer = self.factory.master.get_hc(args[0])
                self.__answer_json(answer)
            elif self.cmd is Requests.START_SCAN:
                try:
                    if self.factory.master.start_scan(args[0]):
                        self.__answer(json_trsuccess())
                    else:
                        self.__answer(json_trerror(333, "Did not start scan"))
                except CannotStartScanError as e:
                    self.__answer(json_trerror(333, e.value))
            elif self.cmd is Requests.STOP_SCAN:
                self.factory.master.stop_scan()
                self.__answer(json_trsuccess())
            elif self.cmd is Requests.STOP_SMOOTH:
                self.factory.master.stop_scan(True)
                self.__answer(json_trsuccess())
            elif self.cmd is Requests.VERSION:
                self.__answer(self.factory.master.VERSION)
            elif self.cmd is Requests.CHECK_SCAN:
                self.read_data = ''
            elif self.cmd is Requests.NODE:
                answer = self.factory.master.get_node_by_name(args[0])
                self.__answer_json(answer)
            else:
                logger.error("Not (yet) a command: _%s_" % self.read_data)
                self.__answer('')
                self.transport.loseConnection()
        else:
            if self.cmd is Requests.CHECK_SCAN:
                scanconfig = mapwoc_dejson(self.read_data)
                (ok,node,msg) = self.factory.master.check_scan(scanconfig)
                if ok:
                    self.__answer(json_trsuccess())
                else:
                    self.__answer(json_trerror(444, "Node %s can not start scan: %s" % (node,msg)))
            else:
                logger.error("Unexpected input: _%s_" % self.read_data)
                self.transport.loseConnection()


    def __answer(self, answer):
        """
        Sends answer through the wire and closes connection afterwards.
        @param answer: The answer to send.
        """

        self.transport.write(answer)
        self.transport.loseConnection()


    def __answer_json(self, answer):
        """
        Sends json encoded answer through the wire and closes the
        connection afterwards.
        @param answer: The answer to send.
        """

        json_str = MapWOCJSONEncoder().encode(answer)
        self.__answer(json_str)


class MasterResponseFactory(Factory):
    """Factory producing MasterRequest Protocols."""

    protocol = MasterRequest

    def __init__(self, master):
        self.master = master

