# -*- coding: utf-8 -*-

# mapWOC
#
# Description: State of a master.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from datetime import datetime

class MasterState(object):
    """Representation of the state of a master."""

    IDLE,SCANNING,STOPPING,ERROR = range(1,5)
    strstate = {IDLE : 'idle',
                SCANNING : 'scanning',
                STOPPING : 'stopping',
                ERROR : 'error'}

    def __init__(self):
        """Initialiases Master State Set."""

        self.state = self.IDLE
        self.scanname = None
        self.nodestates = {}
        self.nodestates_are_valid = True
        self.redirectorstate = None

        self.state_stamp = None
        self.nodestate_stamp = None
        self.redirectorstate_stamp = None

    def set(self, newstate):
        """Sets state of master."""

        self.state = newstate
        self.state_stamp = datetime.now()

    def export(self):
        """Returns dictionary of stateset as expected by mapwocgui."""

        return {'master' : self.strstate[self.state],
                'scanname' : self.scanname,
                'nodes' : self.nodestates,
                'redirector' : self.redirectorstate}

    def of_nodes_invalidate(self):
        """Invalidates the stateset of the nodes until they are updated."""
        self.nodestates_are_valid = False

    def of_nodes_is_valid(self):
        """Returns true if the nodestates are valid."""
        return self.nodestates_are_valid

    def of_nodes_set(self, newstates):
        """Sets states of nodes."""
        self.nodestates = newstates
        self.nodestates_are_valid = True
        self.nodestate_stamp = datetime.now()

    def of_redirector_set(self, newstate):
        """Sets state of redirector."""
        self.redirectorstate = newstate
        self.redirectorstate_stamp = datetime.now()

    def are_nodes_idle(self):
        """True if all nodes state is 'idle'."""

        if not self.of_nodes_is_valid():
            return False

        for _node,state in self.nodestates.items():
            if not state or state.get('') != 'idle':
                return False
        return True

    def node_has_error(self):
        err_list = []
        for node, state in self.nodestates.items():
            if state and state.get('') == 'error':
                err_list.append( (node, state.get('err_msg')) )
        return err_list

    def is_redirector_done(self):
        return self.redirectorstate.get('state') == 'done'

    def is_redirector_idle(self):
        return self.redirectorstate and self.redirectorstate.get('state') == 'idle'

