# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Weak abstractions of connections of a master to nodes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import logging
import socket

import gnutls.crypto
from gnutls.crypto import *
from gnutls.connection import *

from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook
from mapwoc.lib.node import Node
from mapwoc.lib.trivialresponse import TrivialResponse
from mapwoc.lib.gnutls_util import recvall

logger = logging.getLogger('master')

class NodeConnection(Node):
    """Object to query a Node."""

    def __init__ (self, filename, credentials=None):
        """Read information about how to connecto to a node from a keyfile."""

        Node.__init__(self, filename)
        self.credentials = credentials

    def ask(self, question):
        """
        Connect to node, send a string and receive answer.
        @param question: Question to be asked.
        @return: Answer received.
        """

        try:
            # Create a socket (SOCK_STREAM means a TCP socket).
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            if self.credentials:
                session = ClientSession(sock, self.credentials)
                session.connect((self.ip, int(self.port)))
                session.handshake()
                session.sendall(question + "\r\n")
                received = recvall(session)
                logger.debug("(%s) connected to master with certificate: %s" %
                             (str(session.socket.getpeername()),str(session.peer_certificate.subject)))
                try:
                    session.bye()
                    session.close()
                except:
                    # Likely, session was invalidated by recvall-call.
                    pass
            else:
                # Connect to server and send the question.
                sock.connect((self.ip, int(self.port)))
                sock.sendall(question + "\r\n")
                # TODO lessen assumption about amount of data transmitted (twisted?)
                # Receive data from the server and shut down.
                received = sock.recv(2*16384, socket.MSG_WAITALL)
                sock.close()
        except socket.error as err:
            logger.error("Exception: while talking to node %s:%s Error: %s" % \
                         (self.ip, self.port, str(err)))
            logger.warning("Question was: %s " % question)
            received = "Connection Error: " + str(err)
        except Exception as gnerr:
            logger.error("GnuTLS-Exception: while talking to node %s:%s Error: %s" % \
                        (self.ip, self.port, str(gnerr)))
            logger.warning("Question was: %s " % question)
            received = "Connection Error: " + str(gnerr)
        return received

    def ask_decode(self, question):
        """
        Asks question to node, returns deserialized answer.
        @param question: Question understood by node.
        @return: Answer, raises exception in error cases.
        """
        answer = self.ask(question)
        try:
            dec = json.loads(answer, object_hook=mapwoc_from_json_hook)
            return dec
        except Exception as e:
            logger.warning("Communication with node '%s' failed: (%s) %s "\
                           % (self.name, str(answer),str(e)))
            # TODO handle error
            # raise
            return None

    def ask_ok_tr(self, question):
        """
        Asks node for a trivial response and returns answer.
        @param question: Question to ask.
        @return: A pair, where the first element is True if no error is
                 indicated by the TrivialResponse, which is the second
                 element in the pair.
        """
        tr = self.ask_decode(question)
        ok = isinstance(tr, TrivialResponse) and not tr.error
        return ok,tr

    def cmd_startrun(self, scan, job):
        job_str = MapWOCJSONEncoder().encode(job)
        question = "startrun\r\n%d\r\n%s" % (scan.id, job_str)
        return self.ask_decode(question)

    def cmd_stopsmooth(self):
        return self.ask('stopsmooth')

    def cmd_stoprun(self):
        return self.ask('stoprun')

    def cmd_checkjob(self, visitorconf):
        cmd = "checkjob\r\n%s" % MapWOCJSONEncoder().encode(visitorconf)
        return self.ask_ok_tr(cmd)

    # TODO some of these would do fine with "caching".
    def cmd_get_hcs(self):
        return self.ask_decode('hcs')

    def cmd_get_resultdir(self):
        return self.ask('resultdir')

