# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process to poll a Redirector.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import logging
import socket
from time import sleep

from gnutls.crypto import *
from gnutls.connection import *

from mapwoc.lib.gnutls_util import recvall
from mapwoc.lib.mapwoc_json import mapwoc_from_json_hook as json_decode
from mapwoc.lib.tickprocess import TickProcess

logger = logging.getLogger('master')

class RedirectorProxy(object):

    def __init__(self, host='localhost', port=8228, credentials=None):
        self.host = host
        self.port = port
        self.credentials = credentials

    def ask(self, question):
        """
        Asks a redirector and returns the answer as a string.

        @param question: The Question to ask.
        @return: Answer as string (does not catch exceptions).
        """

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if self.credentials:
            session = ClientSession(sock, self.credentials)
            session.connect((self.host,self.port))
            session.handshake()
            session.sendall(question)
            received = recvall(session)
            try:
                session.bye()
                session.close()
            except Exception as e:
                # Session was likely invalidated by too many recv-calls in
                # recvall.
                pass
        else:
            sock.connect((self.host,self.port))
            sock.sendall(question)
            received = sock.recv(4*32768, socket.MSG_WAITALL)
            sock.close()

        return received

    def ask_decode(self, question):
        """
        Asks a redirector, returns json-decoded answer.

        @param question: The Question to ask.
        @return: Decoded object.
        """
        answer = self.ask(question)
        obj = json.loads(answer, object_hook=json_decode)
        return obj

    def cmd_idle(self):
        """Ask redirector to go idle."""
        return self.ask('go_idle')

    def cmd_fastmode(self, numurls):
        """
        Ask redirector to serve mutliple urls per request.
        @param numurls: Number of urls per request to hand out.
        @return: Answer (string).
        """
        return self.ask('fastmode %d' % numurls)

    def cmd_loaddb(self, listids, dbfile):
        """
        Ask redirector to load urllists (by id) from db (by filepath).
        @param listids: List of rowids of urllists to load from db.
        @param dbfile: Absolute path to db.
        @return: Answer (decoded).
        """
        collections = ""
        for ulid in listids:
            collections += "%d," % ulid

        question = 'loaddb %s,%s' % (collections, dbfile)
        return self.ask_decode(question)

    def ask_ext_state(self):
        """
        Asks redirector for an extended state.
        @return: Extended state (dictionary).
        """
        return self.ask_decode('state_ext 30 10')

class RedirectorPollProcess(TickProcess):
    """
    Ask a redirector every poll_interval seconds about its state.
    Its answer is put in the Queue.
    """

    def __init__(self, queue, poll_interval, redirector):
        """
        Sets up (but doesnt start) the Process to periodically check for
        a redirectors state.

        @param queue: A Queue that will fill up with answers (strings) of the
                      redirector.
        @param poll_interval: The interval in seconds to wait between two
                              communication attempts to the redirector.
        @param redirector: Pre-configured RedirectorProxy.
        """

        TickProcess.__init__(self, queue, poll_interval)
        self.redirector = redirector

    def tick(self):
        """
        Ask redirector about its state, put answer string in queue, sleep
        given number of seconds.
        """

        # TODO try/catch here instead of in ask_redirector
        # TODO proper error handling
        try:
            response = self.redirector.ask_ext_state()
            self.put(response)
        except Exception as e:
            #logger.warning("Error when talking to redirector: %s" % msg)
            logger.warning("RedirectorPollProcess got no answer from %s:%d: %s" % \
                          (self.redirector.host,self.redirector.port, str(e)))
            self.put(e)

