# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process results.
#
# Authors:
# Andre Heinecke <aheinecke@intevation.de>
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
from datetime import datetime
import logging
import os
import shutil
import tarfile
from time import sleep
import traceback

from sqlalchemy.exc import InvalidRequestError, OperationalError

import mapwoc.lib.util as util

from mapwoc.lib.result import Result, Visit, ResultType, ResultFile
from mapwoc.lib.result import static_result_type_id
from mapwoc.lib.result import static_result_type_severity
from mapwoc.lib.scanconfiguration import ScanConfiguration, ScanRun
from mapwocgui.model.meta import Session
from mapwoc.lib.url import URL

logger = logging.getLogger('master')

DEBUGANALYSER = True

class ResultAnalyser(object):
    """ Provides a base class for Result Analysers """

    def __init__(self, result_data, node, tolerance):
        """
        ResultAnalyser class to enter raw result_data into the Database as
        results.

        @param result_data: Qualifier (path to) of the raw data which should be parsed.
        @param node: Name of node on which the scan was running.
        @param tolerance: Required highest severity for a result to store.
        """
        if DEBUGANALYSER:
            logger.debug("ResultAnalyser __init__ with data %s for node %s" %\
                    (result_data, node))
        self.result_data = result_data
        self.visit = Visit()
        self.visit.filepath = result_data
        self.visit.node = node
        Session.autoflush = False
        self.tolerance = tolerance

    def scan_metainformation(self, fptr):
        """
        Scan a Meta Information file.

        @param fptr: Filelike object with meta information about Visit.
        """
        parser = ConfigParser.RawConfigParser()
        parser.readfp(fptr)

        self.visit.scanrun_id = parser.getint("scan", "scanrunid")
        self.visit.date = datetime.fromtimestamp(parser.getfloat("scan", "time"))

        # TODO Clear whether we really need these info.
        #scanrun = Session.query(ScanRun).filter_by(id=self.visit.scanrun_id).first()
        #self.visit.scanrun = scanrun
        if DEBUGANALYSER:
            logger.debug("Metainformation scanrun.id %d" % self.visit.scanrun_id)

    def construct_result(self, url_string, type_name, method, date=None):
        """
        Construct a Result object.

        @param url_string: URL (string) of this result.
        @param type_name: Type of the result as a string.
        @param method: Method with which the result was obtained.
        @return: Result object with valid type but wihout meta information.
        """

        rtype_id = static_result_type_id(type_name)
        if rtype_id == 1:
            logger.warning("Added unknown result type %s " % type_name)
        if DEBUGANALYSER:
            logger.debug("New result %s (%s)" % (type_name, method))
            logger.debug("Result at: %s" % url_string)
        result = Result(url_string, method=method, date_time=date)
        result.result_type_id = rtype_id

        return result

    def commit(self):
        """
        Commits the results and visit of the analyser to the database.
        """
        # Ensure dates are set.
        for result in [r for r in self.visit.results if r.date == None]:
            result.date = self.visit.date

        # Add all bits.
        for result in self.visit.results:
            Session.add(result)
            if result.result_file:
                Session.add(result.result_file)
        self.visit.numresults = len(self.visit.results)
        Session.add(self.visit)
        max_trials = 4
        trials = 0
        while trials < max_trials:
            try:
                Session.commit()
                trials = max_trials + 1
            except Exception as e:
                trials += 1
                logger.warning("Database is locked or transaction issue, retrying (%s)." % str(e))
                if trials == max_trials:
                    traceback.print_exc()
                else:
                    sleep(1)
        if trials == max_trials:
            # TODO this has to be escalated.
            logger.critical("Could not commit result (database is locked or transaction issue).")
            pass

    def is_interesting(self, result):
        """
        Check if the results are interesting.
        Also store severest_result_type_id
        @param result: Result to check
        @return: True if something interesting has been found.
        """
        # TODO Just access a proper index (id->severity) in memory.
        return static_result_type_severity(result.result_type_id) >= self.tolerance

    def generic_analyse(self):
        """
        Runs generic steps of the analysis.
        @return: True if a result has been found, False otherwise.
        """
        if self.visit.results:
            interesting_results = []
            max_sev = 0
            for result in self.visit.results:
                if self.is_interesting(result):
                    interesting_results.append(result)
                    # TODO Just access a proper index (id->severity) in memory.
                    sev = static_result_type_severity(result.result_type_id)
                    if sev > max_sev:
                        max_sev = sev
                        self.visit.severest_result_type_id = result.result_type_id
                else:
                    # Deleting should not be neccessary since they
                    # are not yet added or commited.
                    try:
                        # Without this some results get added
                        # on the commit that should not be added.
                        Session.expunge(result)
                    except:
                        pass
                    del result
            self.visit.is_interesting = len(interesting_results) > 0
            self.visit.results = interesting_results
            if not self.visit.results:
                self.visit.filepath = None
            logger.debug("Added new visit into database with %d results." %\
                           len(self.visit.results))
            self.commit()
            return True

        logger.warning("Sucessfully scanned %s but did not get any results."%\
                        self.result_data)
        return False

