# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process results from an XML representation.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <andre.heinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from xml.etree.ElementTree import ElementTree

from mapwoc.lib.result import Result, Visit, ResultType, ResultFile
from mapwoc.lib.result import static_result_type_id
from mapwoc.lib.scanconfiguration import ScanConfiguration, ScanRun
from mapwocgui.model.meta import Session
from mapwoc.lib.url import URL

import traceback
import logging
import tarfile
import datetime

from mapwoc.masterctrl.resultanalyser import ResultAnalyser

logger = logging.getLogger('master')
DEBUGANALYSER = True

class AnalyserError(Exception):
    """
    Specific Analysis error (with value).
    """
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)

class ResultFileAnalyser(ResultAnalyser):
    """
    Class to analyse result files (xml) provided  by analysis script.
    """

    def __init__(self, result_data, node, tolerance=-1):
        """
        Resultanalyser class to parse and enter results into the Database.

        @param result_data: Qualifier (path to) of the xml data which should be parsed.
        @param node: Name of node on which the scan was running.
        @param tolerance: Required highest severity for a result to store.
        """
        ResultAnalyser.__init__(self, result_data, node, tolerance)

    def analyse_dry(self):
        """
        Method for testing purposes.
        """
        self.__walk_xmlfile(self.result_data)

    def analyse(self):
        """
        Starts the analysis of a result.
        @return: True if at least one result could be entered into
                 the database.
        """

        try:
            if DEBUGANALYSER:
                logger.debug("Parsing result- %s ." % self.result_data)
            tgz = tarfile.open(self.result_data)
            self.scan_metainformation(tgz.extractfile("metainfo"))
            self.__walk_xmlfile(tgz.extractfile("result/Log/Analyse.xml"))
            return self.generic_analyse()
        except:
            traceback.print_exc()
            #logging.error("Invalid tarball %s ignoring result." % \
            #        self.result_data)
            #Session.rollback()
            return False

    def __walk_files(self, filevisits):
        """
        Imports the fileresults into the database.

        @param filevisits: List of file xml etree elements.
        """

        images_in_visit = 0

        result_data = None
        for filevisit in filevisits:
            results_for_file = []
            rf = ResultFile()
            rf.filepath = filevisit.find('extractname').text
            source = None
            if filevisit.find("source") != None:
                source = filevisit.find("source").text
            # Connect the resultfiles to the related connections.
            if source:
                for result in self.visit.results:
                    if ( result.url == source and
                           not result.result_file ):
                        result_time = result.date
                        result.result_file = rf

            # Gather all information gained about the file.
            for finding in filevisit.findall("analysis/method"):
                method = finding.get("name")
                if method == "tool:sha1sum":
                    rf.sha1sum = finding.find("output").text
                    continue
                elif method == "tool:ssdeep":
                    rf.ssdeep = finding.find("output").text
                    continue
                elif method == "tool:file -b":
                    rf.filetype = finding.find("output").text
                    if rf.filetype and rf.filetype.startswith("Mach-O"):
                        source_t = filevisit.find("source").text
                        r = self.construct_result(source_t,
                                                  "MACExecutable", method,
                                                  result_time)
                        r.result_file = rf
                        self.visit.results.append(r)
                    continue
                elif method == "tool:file -bi":
                    rf.mimetype = finding.find("output").text
                    for mimetype in rf.mimetype.split():
                        mimetype = mimetype.strip(";")
                        if mimetype == "application/pdf":
                            r = self.construct_result(
                                    filevisit.find("source").text,
                                                   "PDF", method, result_time)
                            r.result_file = rf
                            self.visit.results.append(r)
                        elif mimetype.startswith("image"):
                            images_in_visit += 1
                        elif mimetype == "application/x-shockwave-flash" or \
                             mimetype == "video/x-flv":
                            r = self.construct_result(
                                    filevisit.find("source").text,
                                                   "Flash", method, result_time)
                            r.result_file = rf
                            self.visit.results.append(r)
                        elif mimetype == "application/vnd.ms-office":
                            r = self.construct_result(
                                    filevisit.find("source").text,
                                                   "MSOfficeDocument",
                                                   method, result_time)
                            r.result_file = rf
                            self.visit.results.append(r)
                    continue
                elif method == "mapwocanalyse.pl/type_classifier":
                    if finding.find("output").text == "Windows Executable/PE":
                        r = self.construct_result(filevisit.find("source").text,
                                                  "WindowsExecutable",
                                                  finding.get("name"), result_time)
                        r.result_file = rf
                        r.text = filevisit.find("filename").text
                        self.visit.results.append(r)
                    continue
                elif method == "tool:clamav":
                    if finding.find("output").text != "no virus":
                        r = self.construct_result(filevisit.find("source").text,
                                                  "VirusFound",
                                                  finding.get("name"), result_time)
                        r.result_file = rf
                        r.text = finding.find("output").text
                        self.visit.results.append(r)
                    continue

                r = self.construct_result(filevisit.find("source").text,
                                          finding.get("name"),
                                          finding.get("name"), result_time)
                r.text = finding.find("output").text

                r.result_file = rf
                self.visit.results.append(r)

        # Result collections
        if images_in_visit:
            if len(self.visit.urls) == 1:
                r = self.construct_result(self.visit.urls[0].url,
                                          "GraphicsFound",
                                          "tool:file -bi")
            else:
                r = self.construct_result("unknown",
                                          "GraphicsFound",
                                          "tool:file -bi")
            r.text = "%s Images downloaded in this visit." % images_in_visit
            self.visit.results.append(r)

    def __walk_connections(self, connections):
        """
        Imports the connections into the database.

        @param connections: List of connection xml etree elements.
        """

        for connection in connections:
            protocol  = connection.find("protocol").text
            if protocol.upper() != "HTTP" and protocol.upper() != "HTTPS":
                r = self.construct_result("unknown", "UnknownProtocol",
                                          "UnknownProtocol")
                r.text = "A connection over an unknown protocol ('%s') was made." % protocol
                self.visit.results.append(r)
                continue
            ip           = connection.find("ip").text
            code         = connection.find("code").text
            result_time  = connection.get("time")
            result_time  = datetime.datetime.strptime(result_time,
                                                  "%Y-%m-%d-%H:%M:%S").time()
            result_time  = self.visit.date + datetime.timedelta(
                                                hours=result_time.hour,
                                                minutes=result_time.minute,
                                                seconds=result_time.second)
            ressource    = connection.find("ressource")
            location     = connection.find("location")
            r            = None
            #If a connection contains a 302 location is the redirected url
            #and ressource is the redirector.

            if ressource != None and ressource.text and ip != "10.0.0.1":
                url = ressource.text
            elif location != None and location.text:
                url = location.text
            else:
                url = ip

            findings = connection.findall('findings/method')
            if not findings:
                if ip == "10.0.0.1": # Redirector
                    continue
                if code == "200":
                    r = self.construct_result(url, "RegularTraffic",
                                              "RegularTraffic", result_time )
                elif code == "404":
                    r = self.construct_result(url, "404", "404", result_time)
                elif code == "204":
                    r = self.construct_result(url, "404", "404", result_time)
                elif code == "302":
                    r = self.construct_result(url, "302", "302", result_time)
                    if location:
                        location_text = location.text
                    else:
                        location_text = "(Empty Location header field)"
                    if ressource:
                        r.text = "Redirect from %s to %s" % \
                                (ressource.text, location_text)
                    else:
                        r.text = "Redirect from %s to %s" % \
                                (ip, location_text)
            else:
                for finding in findings:
                    r = self.construct_result(url,
                                              finding.get("name"),
                                              finding.get("name"), result_time)
                    r.text = finding.find("output").text
            if r:
                self.visit.results.append(r)

    def __walk_xmlfile(self, resultb):
        """
        Imports result data into the fields.

        @param resultb: Path to result file.
        """

        tree = ElementTree()
        tree.parse(resultb)
        root = tree.getroot()
        visitor = root.get("visitor")
        # Remove the 6 tmp chars and last _
        visitor = visitor.replace("-snapshot", "")
        visitor = visitor.replace(".dump", "")
        self.visit.hcsid = visitor[:len(visitor)-7]

        if not root.find("connections"): # No connections were made
            r = self.construct_result("Unknown", "MissingTraffic",
                                      "MissingTraffic")
            self.visit.results.append(r)
            return
        targets = root.findall('target')
        for target in targets:
            # 10.0.0.1 and 10.0.0.2 are known addresses (redirector and proxy).
            if (target.find("ip").text and target.find("ip").text.startswith("10.0.0.")
                and target.find('location') != None):
                # The redirector for single url
                url = target.find('location').text
            else:
                # Fastmode
                url = target.find("url").text
            db_url = Session.query(URL).filter_by(url=url).first()
            if not db_url:
                db_url = URL(url)
            self.visit.urls.append(db_url)
        self.__walk_connections(root.findall('connections/connection'))
        self.__walk_files(root.findall('files/file'))
#        targets = root.findall('target')
#        for target in targets:
#            self.visit.urls.append(URL(target.text))
            # IP missing

