# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Process to poll results from nodes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import datetime
import glob
import logging
import multiprocessing
from os import listdir, path, rmdir, remove, walk, mkdir, makedirs
from os import path
from os.path import isdir, isfile, basename
from Queue import Empty
from shutil import rmtree
from sqlalchemy import *
import subprocess
from subprocess import call
import tarfile
import tempfile
from time import sleep
from traceback import format_exc

from mapwoc.lib.result import Result, ResultType, Visit
from mapwoc.lib.url import URL
import mapwoc.lib.model.result
import mapwoc.lib.model.url
from mapwoc.lib.tickprocess import TickProcess
from mapwoc.masterctrl.resultfileanalyser import ResultFileAnalyser

from mapwocgui.model.meta import Session

logger = logging.getLogger('master')

class ResultPollProcess(TickProcess):
    """
    Process that polls for results (rsnyc) of nodes.
    In case that any results where fetched, these are registered
    in the database.
    """

    def __init__(self, working_dir, nodes, result_queue, credentials_path, db,
                 tolerance):
        """
        Constructor, set up fields to sane values.

        @param working_dir: Directory in which to create temporary dir. for syncing.
        @param nodes: NodeConnection list of nodes with which to sync.
        @param result_queue: Queue in which to write when results were found.
        @param credentials_path: Path to certificates for ssh/rsync
        @param db: Path to database in which to store the results.
        @param tolerance: Required highest severity for a result to store.
        """

        TickProcess.__init__(self, result_queue, 10)
        self.__working_dir = working_dir
        self.__nodes = nodes
        self.credentials_path = credentials_path
        self.db_file = db
        self.tolerance = tolerance

    def __bind_to_db(self):
        """Bind the Session to database."""

        engine = create_engine('sqlite:///%s' % self.db_file)
        c = engine.connect()
        Session.configure(bind=engine)

    def communicate_log_err(self, proc):
        """
        Communicates with process and logs it in error class.
        @param proc: Process to communicate with.
        @return: Stderr of process.
        """
        stdout_value, stderr_value = proc.communicate()
        logger.error("stderr: %s" % stderr_value)
        logger.error("stdout: %s" % stdout_value)
        logger.error("exit code: %d" % proc.returncode)
        return stderr_value

    def tick(self):
        """
        Called upon call to start().
        Starts a Rsync process peridically.
        """

        self.__bind_to_db()

        rsnc_subprocess = None

        # Call that rsync subprocess, sync from there into temp dir.

        for nodekey, node in self.__nodes.items():
            tempd = tempfile.mkdtemp(dir=self.__working_dir,
                                     prefix='resultsync_%s_' % nodekey)
            tempd = path.abspath(tempd)
            if not node.data_path:
                logger.error("Node %s does not have a resultdir set." +\
                             " Is it running?", nodekey)
                continue
            cmd = ['rsync', '-az',
                    '-e',
                    ('"/usr/bin/ssh -oKbdInteractiveAuthentication=no '
                      '-oPasswordAuthentication=no '
                      '-oNumberOfPasswordPrompts=0 -i %s"'
                    ) % path.abspath(self.credentials_path),
                    '--remove-source-files',
                    '%s:%s' % (node.rsync_address, node.data_path),
                    tempd ]
            # TODO avoid shell=True and join()-call.
            p = subprocess.Popen(" ".join(cmd),
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 shell=True)
            p.wait()

            if p.returncode != None:
                if p.returncode == 0:
                    logger.info("Result Sync with node %s successful."
                                % nodekey)
                    # Now, check what came and import it to db.
                else:
                   stderr = self.communicate_log_err(p)
                   logger.error("Could not sync results with node %s: %s"
                                  % (nodekey, stderr))
                   logger.warning("Command was: %s" % cmd)
                   # Escalate this properly (maybe no space left for sync.).
                   # TODO Use TaskIOQueue- mechanism to raise exception over queue.
                   e = Exception("Rsync with node failed: %s" % stderr)
                   self.put(e)
            else:
                logger.warning("Rsync without returncode.")

            # TODO Determine number of files synchronized (len of stdout?),
            #      By globbing through directory?

            # Walk over synced files once, put in db/remove.
            # TODO: Signal the master if something serious is happening.
            if self.__resultdir_to_db(tempd, nodekey):
                self.put('marker')
            #(self.master.update_results)

            # TODO Now, the temporary directory with synchronized files
            #      can be removed, can it?


    def __resultdir_to_db(self, resultdir, node=None):
        """
        Checks a result directory for result tarballs
        and calls the according analayser.
        @param resultdir: Directory in which to look for result tarballs.
        @param node: Name of the node of which the results where synced.
        """

        should_keep = False
        for dirpath, dirnames, filenames in walk(resultdir):
            if not filenames and not dirnames:
                # Remove empty result subdir.
                rmdir(dirpath)
                continue
            for filename in filenames:
                tarpath = path.join(dirpath, filename)
                # TODO Escalate this, maybe put tarball in quarantine.
                if not filename.endswith(".tar.gz"):
                    logger.error("Analysis: Ignoring file: %s ." % tarpath)
                    should_keep = True
                else:
                    analyser = ResultFileAnalyser(tarpath, node, self.tolerance)
                    if not analyser.analyse():
                        logger.error("Could not extract result from %s ." % tarpath)
                        should_keep = True
                    else:
                        # Save screendump.
                        if len(analyser.visit.urls) == 1:
                            # Move screendump to a proper place.
                            db_dir = path.dirname(self.db_file)

                            screendump_dir = path.join(db_dir,
                                "screendumps", str(analyser.visit.urls[0].id))
                            try:
                                makedirs(screendump_dir)
                            except OSError as oe:
                                pass

                            # Now place file.
                            screendump_path_ppm = path.join(screendump_dir,
                                "%s.ppm" % analyser.visit.hcsid)
                            screendump_path_png = path.join(screendump_dir,
                                "%s.png" % analyser.visit.hcsid)
                            self.__extract_screendump(tarpath, screendump_path_ppm)

                            # Convert the PPM to a PNG.
                            if call(["convert", screendump_path_ppm, screendump_path_png]) == 0:
                                remove(screendump_path_ppm)
                            else:
                                logger.warn("Could not convert screendump file.")
                        else:
                            logger.debug("More than one url, not saving screendump")

                        if analyser.visit.results:
                            # Throw away all files extracted from the dump
                            # and only keep the dump to keep tarball sizes
                            # small
                            should_keep = True
                            self.__cleanup_tarball(tarpath, analyser.visit)
                        else:
                            remove(tarpath)
        # If there is nothing interesting to keep from this sync cleanup
        # the empty directory
        if not listdir(resultdir):
            rmtree(resultdir)


    def traverse(self, directory, whitelist = lambda f: True):
        """
        Traverse through a directory tree and return every
        filename that the function whitelist returns as true.

        @param directory: Directory which to traverse.
        @param whitelist: Argument to call on traversed items.
        """

        dirs = [ directory ]
        while dirs:
            mypath = dirs.pop()
            for f in listdir(mypath):
                f = path.join(mypath, f)
                if isdir(f):
                    dirs.append(f)
                elif isfile(f) and whitelist(f):
                    yield f

    def __extract_screendump(self, tarpath, target):
        """
        Extract screendump image file if found.
        @param tarpath Path to the tarball.
        @param target where to unpack.
        """
        tgz = tarfile.open(tarpath)
        zipped_file = tgz.extractfile('result/screendump.ppm')
        outfile = open(target, 'w')
        outfile.write(zipped_file.read())
        outfile.flush()
        outfile.close()
        tgz.close()

    def __cleanup_tarball(self, tarpath, visit):
        """
        Extracts the tarball \param tarpath into a temporary directory, and
        then removes everything but the dump.

        @param tarpath: Path to tarball.
        @param visit: The visit which to look at.
        """

        tempd = tempfile.mkdtemp(dir=self.__working_dir)
        tempd = path.abspath(tempd)
        tgz = tarfile.open(tarpath)
        tgz.extractall(tempd)
        tgz.close()
        ignore_patterns = ["metainfo",
                           "Analyse.xml"]
        #ignore_patterns = [path.join(tempd, "metainfo"),
        #                   path.join(tempd, "result", "Log", "Analyse.xml")]
        for result in visit.results:
            if result.result_file:
                ignore_patterns.append(path.basename(result.result_file.filepath))
        for fname in self.traverse(tempd,
                lambda f: path.basename(f) not in ignore_patterns and \
                            not f.endswith(".ppm") and \
                            not f.endswith(".dump") and \
                            not f.endswith(".proxy")):
            remove(path.join(tempd, fname))

        for dirpath, dirnames, filenames in walk(tempd):
            if not filenames and not dirnames:
                # Remove empty subdirs.
                rmdir(dirpath)
        tgz = tarfile.open(tarpath, "w:gz")
        tgz.add(path.join(tempd, "metainfo"), arcname="metainfo")
        tgz.add(path.join(tempd, "result"), arcname="result")
        tgz.close()
        rmtree(tempd)
