# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Analyser-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import multiprocessing
from os import close, remove, path
from os.path import abspath, basename
from shutil import copy, move, rmtree
import subprocess
import sys
import tarfile
import tempfile
from time import sleep

from mapwoc.lib.result import Result
from mapwoc.lib.taskqueuepair import TaskQueuePair

logger = logging.getLogger("node")

class AnalysisProcess(multiprocessing.Process):
    """
    Analysis Process sits and waits for incoming filepaths in job_queue.
    Once a filepath could be fetched from this queue, the analysis script
    is thrown at it and the path to the results is fed ("back") into the result_queue.
    """


    def __init__(self, io, working_dir, result_dir,
                 analyser = '/usr/bin/mapwocanalyse.pl',
                 conf = '/etc/mapwoc/antivir.cfg'):
        """
        Constructor.
        @param io: The io queue pair to read jobs from and write
                   errors/results to.
        @param working_dir: Working directory (where temp. files will be placed.)
        @param result_dir: Directory where result tarballs will be placed.
        @param analyser: Path to the analyser which should be used.
        @param conf: Path to the configuration file which should be used by the analyser.
        """

        multiprocessing.Process.__init__(self)
        self.io = io
        self.working_dir = abspath(working_dir)
        self.result_dir = abspath(result_dir)
        self.analyser = analyser
        self.conf = conf

    def run(self):
        """
        Invoked on start() - call.
        Polls Queue for (dump-)filenames and analyses these if possible.
        Paths to results are pushed in the result queue.
        """

        while True:
            # Sit on queues get()- call until fed path to dumps.
            job = self.io.get_command(None)

            # Check if a proxy dump is given.
            if type(job) != type(tuple()):
                tcpdump = job
                proxydump = None
            else:
                tcpdump = job[0]
                proxydump = job[1]

            logger.info("Starting analysis of: %s" % tcpdump)

            tempdir = tempfile.mkdtemp(dir=self.working_dir,
                                       prefix='analyser_')
            tempdir = abspath(tempdir)

            # TODO When copying fails, node should go into error state.
            copy(tcpdump, tempdir)
            if proxydump:
                copy(proxydump, tempdir)
            try:
                screendump = tcpdump + ".ppm"
                open(screendump)
                copy(screendump, path.join(tempdir, "screendump.ppm"))
            except IOError as (errno,strerr):
                pass

            # Obtain a temporary working dir.
            cmd = [self.analyser, "-p", abspath(tcpdump), "-a", self.conf, "-b", tempdir]

            if proxydump:
                cmd += ["-j", abspath(proxydump)]

            logger.debug("Starting analysis: %s" % ", ".join(cmd))
            process = subprocess.Popen(cmd, close_fds=True,
                                       stdin=subprocess.PIPE,
                                       stdout=subprocess.PIPE,
                                       stderr=subprocess.PIPE)
            # Wait for process to finish.
            process.wait()

            # Signal if error with analysis happened.
            if process.returncode:
                stdout_value, stderr_value = process.communicate()
                logger.error("stdout: %s" % stdout_value)
                logger.error("stderr: %s" % stderr_value)
                logger.error("exit code: %d" % process.returncode)
                self.io.add_err(Exception("Analysis process died: %s." % stderr_value))
                # TODO Handle this case further.

            # Now bundle the results.
            # TODO watch out leaking file descriptor issue.
            try:
                _fh,fname = tempfile.mkstemp(dir=self.working_dir,
                                             prefix=basename(tempdir),
                                             suffix='.tar')
                tgz = tarfile.open(name=fname, mode='w')
                tgz.add(tempdir, arcname='result')
                tgz.close()
                close(_fh)
            except Exception as e:
                self.io.add_err(e)
                continue

            try:
                screendump = tcpdump + ".ppm"
                remove(screendump)
            except (OSError, IOError) as e:
                pass

            try:
                remove(tcpdump)
                if proxydump:
                    remove(proxydump)
                rmtree(tempdir)
                self.io.add_result(fname)
            except Exception as e:
                self.io.add_err(e)
                continue
