# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Class to represent (one-way) interaction with a kvm-process.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import re

class __PSEUDOHC(object):
    SHORTCUT_ENTER = '[enter]'
    SHORTCUT_HOMEPAGE = '[homepage]'

def eventmap_from_string(string, hc):
    if hc is None:
        hc = __PSEUDOHC()
    eventmap = []
    if string is not None:
        match = re.findall("([0-9]+[a-z])", string)
        for event in match:
            eventmatches = re.search("([0-9]+)([a-z]+)", event).groups()
            if eventmatches[1] == 'e':
                eventmap += [(int(eventmatches[0]), hc.SHORTCUT_ENTER)]
            else:
                eventmap += [(int(eventmatches[0]), hc.SHORTCUT_HOMEPAGE)]

    return eventmap

class HCInteractionScript(object):
    """
    Tuples of times and actions ("eventmap").
    Actions are translated to key-presses that are sent to a VM,
    for example KeyPress.SHORTCUT_HOMEPAGE will resolve to 'alt+home'
    with in turn is the shortcut to go to the home(start)- page in many
    browsers.
    The times are meant in seconds from the script start onwards.
    """

    def __init__(self, script=None):
        """
        Initialize a script.
        @param script: List of tuples, e.g. [(5, ''), (10,'')] .
                       First item is time offset from start, second item is
                       e - for what usually is an ENTER-keypress.
                       everything else - for what is defined as shortcut to
                                         direct the browser to its homepage.
        """

        if script:
            self.script = sorted(script, key=lambda s: s[0])
        else:
            self.script = None

    def needs_action(self, time):
        """
        Figure out whether script wants an action to be done (by) now.
        @return: True if an action should be done at given time or an action
                 should have been done but hasn't yet.
        """
        if self.script is None:
            return False
        if len(self.script) == 0:
            return False
        (at_time,__action) = self.script[0]
        if at_time <= time:
            return True
        return False

    def next_action(self):
        """Next action."""

        try:
            (__t,action) = self.script[0]
            return action
        except:
            return None

    def action_done(self):
        """Mark action as done (remove first list item)."""

        try:
            del(self.script[0])
        except:
            pass


