# -*- coding: utf-8 -*-

# mapWOC
#
# Description: KVM-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import multiprocessing
from os import close, chmod, makedirs, path, remove
from shutil import copy
import socket
import stat
import subprocess
from tempfile import mkstemp
from time import sleep
from Queue import Empty

from mapwoc.lib.hc import HC
from mapwoc.lib.taskqueuepair import TaskQueuePair
from mapwoc.nodectrl.hcinteractionscript import HCInteractionScript
from mapwoc.nodectrl.hcrunprocess import HCRunException, HCRunProcess
from mapwoc.nodectrl.proxy import start_mitmdump

logger = logging.getLogger("node")

# TODO Refactor proxy and io stuff, which is near duplicate in
#      NativeHCRunProcess.
class KVMHCRunProcess(HCRunProcess):
    """
    Process that starts a VM in a subprocess and talks to it via stdin.
    Usually beforehand a second subprocess (a mitmproxy) is started.
    """

    SLEEP_STEP = 1

    def __init__(self, hc, script, timeout, io,
                 work_dir, initial_offset=0, proxy_port=1234,
                 mitmproxy_script=None):
        """
        Constructor.
        @param hc: An HCControl object.
        @param script: A script to be executed (list of pairs (second,action)).
        @param timeout: When the subprocesses will be terminated, in seconds.
        @param io: QueuePair from which to get commands and in which to feed
                   errors and results.
        @param work_dir: Where the dump file will be placed.
        @param initial_offset: Time in seconds to sleep before starting the
                               process (when start() or run() is called).
        @param proxy_port: The port on which the https proxy shoul be started.
        @param mitmproxy_script: The script which mitmproxy should use.
        """

        HCRunProcess.__init__(self, timeout=timeout, hc=hc)

        self.script = HCInteractionScript(script)
        self.vnc_port = None
        self.proxy_port = proxy_port
        self.kvm_subprocess = None
        self.proxy_subprocess = None
        self.work_dir = work_dir
        self.initial_offset = initial_offset
        self.waiting_for_start = True
        self.io = io
        self.mitmproxy_script = mitmproxy_script

    def __del(self):
        """Destructor."""

        # TODO ensure files (from __prepare_kvm_files) are removed.
        pass

    def __prepare_kvm_files(self):
        """Prepare files and paths (e.g. copy image file)."""

        if not path.isdir(self.work_dir):
            makedirs(self.work_dir)

        # TODO Watch out the file descriptor and inheritence issue.
        self.df_handle, self.dump_file = mkstemp(prefix=path.basename('%s_' %
                                          self.hc.hcsid),
                                          suffix='.dump',
                                          dir=self.work_dir)

        self.proxy_dump = self.dump_file.replace(".dump", ".proxy")
        self.cp_img_path = self.hc.image_file

    def __kvm_cmd(self):
        """
        Create KVM Command to start the given HC.
        @return: List to be used with Popen to start kvm process.
        """

        # Networking parameters.
        nup = ["user"]
        # Proxy-forward.
        if self.hc.proxy_ip:
            if self.hc.proxy_port:
                prx_port = self.hc.proxy_port
            else:
                prx_port = 80
            nup.append('guestfwd=tcp:%s:%d-cmd:/bin/nc.traditional localhost %s' % (self.hc.proxy_ip, prx_port, self.proxy_port))
        # Network-related.
        if self.hc.own_ip:
            nup.append('net=%s' % self.hc.own_ip)
        if self.hc.dns_ip:
            nup.append('dns=%s' % self.hc.dns_ip)
        if self.hc.gateway_ip:
            nup.append('host=%s' % self.hc.gateway_ip)
        usermode_net_params = ','.join(nup)

        nic_model_params = ''
        if self.hc.hctype == 'kvm-mac':
            nic_model_params += ',model=e1000'
        elif hasattr(self.hc, 'nic'):
            nic_model_params = ",model=%s" % self.hc.nic

        # Do live migration if not booting.
        migration_cmds = []
        if not self.hc.boots:
            migration_cmds.append('-incoming')
            snapshot_path = self.hc.image_snapshot
            # Is the snapshot (state) compressed?
            if snapshot_path.endswith('gz'):
                migration_cmds.append('exec:gzip -dc  %s' %  snapshot_path)
            else:
                migration_cmds.append('exec:cat %s' % snapshot_path)

        # TODO We could drop privileges (to privs. of other user) with -runas .
        # Maybe kvm should be started as --daemonize.

        cmd = [ "/usr/bin/kvm",
                "-hda", self.cp_img_path,
                "-net",
                # All -net user options need to be made in the following line
                usermode_net_params, # Use user mode network
                "-net", "nic%s" % nic_model_params, # Create Network Interface Card in VLAN 0
                "-net", "dump,file=%s" % self.dump_file # Dump
              ] +  migration_cmds + [ # VM State to load.
                '-snapshot', # Do not alter files.
                '-m', str(self.hc.image_memory),
                '-rtc', "base=localtime",
              ]
        if hasattr(self.hc, 'vmclock') and self.hc.vmclock:
            cmd[-1] += ",clock=vm"

        if self.vnc_port != None:
             cmd += [ "-vnc",  ":%d" % self.vnc_port ]
        else:
             # TODO '-nographic' should save ressources but proved to be
             #      difficult to insert (depends on position in kvm cmd).
             #cmd.append('-nographic')
             cmd += [ "-vnc",  "none" ]
        if self.hc.hctype == 'kvm-mac':
            cmd += [
                    '-M', 'mac',
                    '-osk', 'ourhardworkbythesewordsguardedpleasedontsteal(c)AppleComputerInc',
                    '-no-kvm-irqchip',
                    '-usbdevice', 'keyboard',
                    '-usbdevice', 'mouse',
                    '-rtc', 'clock=vm'
                   ]
        if self.hc.kvmkernel:
            cmd += [
                    '-kernel', self.hc.kvmkernel
                   ]
        cmd += [ "-monitor", "stdio"]

        return cmd

    # TODO resolve duplicate in masterctrl/resultpollprocess
    def communicate_log_err(self, proc):
        """
        Communicates with process and logs it in error class.
        @param proc: Process to communicate with.
        @return: Stderr of process.
        """
        stdout_value, stderr_value = proc.communicate()
        logger.error("stderr: %s" % stderr_value)
        logger.error("stdout: %s" % stdout_value)
        logger.error("exit code: %d" % proc.returncode)
        return stderr_value


    def run(self):
        """
        Invoked on start() - call.
        Starts a KVM subprocess, a man in the middle proxy for ssl traffic
        and a parallel interacts with the kvm process by sending keyboard-press
        commands from a script to the kvm monitor.
        After timeout, abort the processes.
        """

        try:
            self.__prepare_kvm_files()
            cmd = self.__kvm_cmd()
        except Exception as e:
            # TODO Cleanup.
            self.io.add_err(e)
            # return

        if self.initial_offset > 0:
            logger.debug("HC with offset started, sleeping %f" % float(self.initial_offset))
            sleep(self.initial_offset)
        self.waiting_for_start = False

        # Start mitmproxy subprocess.
        use_proxy = self.hc.proxy_ip
        # This will raise an error if something goes wrong.
        if use_proxy:
            (pcmd, self.proxy_subprocess) = start_mitmdump(self.proxy_port,
                                                           self.proxy_dump,
                                                           self.mitmproxy_script)
            logger.debug("Starting mitmdump: %s" % " ".join(pcmd))
            sleep(1) # Give the proxy time to open the port

            # Do initial check.
            self.proxy_subprocess.poll()
            if use_proxy and self.proxy_subprocess.returncode is not None:
                stderr = self.communicate_log_err(self.proxy_subprocess)
                self.io.add_err(Exception("mitmproxy could not be started (%s)." % stderr))
                # TODO Rework the cleanup()-logic. Jump to cleanup() from here.

        # Start kvm subprocess.
        logger.debug("Starting HC with: %s" % " ".join(cmd))
        # TODO try/except?
        self.kvm_subprocess = subprocess.Popen(cmd, stdin=subprocess.PIPE,
                                               stdout=subprocess.PIPE,
                                               stderr=subprocess.PIPE,
                                               close_fds=True)

        self.kvm_subprocess.poll()
        if self.kvm_subprocess.returncode is not None:
            stderr = self.communicate_log_err(self.kvm_subprocess)
            self.io.add_err(Exception("kvm could not be started (%s)." % stderr))
            # TODO Rework the cleanup()-logic. Jump to cleanup() from here.

        # Init time.
        self.start_time()

        while not self.time_over():
            self.kvm_subprocess.poll()
            if use_proxy:
                self.proxy_subprocess.poll()
            if use_proxy and self.proxy_subprocess.returncode is not None:
                stderr = self.communicate_log_err(self.proxy_subprocess)
                self.io.add_err(Exception("mitmproxy process died. (%s)" % stderr))
                break

            if self.kvm_subprocess.returncode is not None:
                stderr = self.communicate_log_err(self.kvm_subprocess)
                self.io.add_err(Exception("KVM process died. (%s)" % stderr))
                break
            try:
                cmd = self.io.get_command()
                # If command is given, this is the command to stop.
                break
            except Empty:
                pass

            while self.script.needs_action(self.elapsed_time()):
                action = self.script.next_action()
                self.kvm_subprocess.stdin.write('sendkey %s\n\r' % action)
                self.kvm_subprocess.stdin.flush()
                self.script.action_done()
                logger.debug("Sent keypress (%s) to VM." % action)

            sleep(KVMHCRunProcess.SLEEP_STEP)

        # Save a screenshot.
        self.kvm_subprocess.stdin.write('stop\n\r')
        self.kvm_subprocess.stdin.write('screendump %s\n\r' % (self.dump_file + ".ppm"))
        self.kvm_subprocess.stdin.flush()
        # Let it some time.
        sleep(KVMHCRunProcess.SLEEP_STEP*2)

        try:
            close(self.df_handle)
            # OSErrors if e.g. already closed.
        except:
            pass

        # Terminate the Processes.
        if self.hc.proxy_ip:
            self.__terminate(self.proxy_subprocess)
        if self.__terminate(self.kvm_subprocess) == None:
            # TODO Ensure that the process is joined.
            pass
        self.cleanup()
        if self.hc.proxy_ip:
            self.io.add_result((self.dump_file, self.proxy_dump))
        else:
            self.io.add_result(self.dump_file)


    def __terminate(self, process):
        """
        Terminates the kvm subprocess and waits for it to shut down.
        Blocks for up to 10 seconds before sending sigkill.

        @return: The subprocesses return code.
        """

        state = process.poll()
        # TODO If can, send quit command to process.
        #process.stdin.write('quit\n\r')
        if state == None:
            process.terminate()
        waited = 0
        while state == None and waited < 10:
            state = process.poll()
            if state:
                break
            sleep(1)
            waited += 1
        if state == None:
            logger.warning("subprocess not terminated after 10 seconds.")
            process.kill()
        return state

    def stop_processes(self):
        """Big red button."""
        # TODO examine issue why this doesnt work
        # (probably related to process/forking)

        self.io.add_task('stop now')
        if self.kvm_subprocess is not None:
            self.kvm_subprocess.terminate()
        if self.kvm_subprocess is not None:
            self.kvm_subprocess.kill()
        if self.proxy_subprocess is not None:
            self.proxy_subprocess.terminate()
        if self.proxy_subprocess is not None:
            self.proxy_subprocess.kill()

    def stop(self):
        """Stop processes directly."""

        self.stop_processes()

    def get_result(self):
        """
        Get result(s) from queue.

        @return: Pair with dumpfiles if proxy was configured, single dump
                 file path otherwise.
        """

        return self.io.get_result()

    def cleanup(self):
        """Any cleanups to do (none, currently)."""

        #remove(self.dump_file)
        #remove(self.proxy_dump)

        logger.debug('Cleaning up kvm process, leaving dump file.')

