# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Native HC-Process related classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import multiprocessing
from os import close, chmod, makedirs, path, remove
from Queue import Empty
from shutil import copy
import socket
import stat
import subprocess
from tempfile import mkstemp
from time import sleep

from mapwoc.lib.hc import HC
from mapwoc.nodectrl.hcinteractionscript import HCInteractionScript
from mapwoc.nodectrl.hcrunprocess import HCRunProcess
from mapwoc.nodectrl.proxy import start_mitmdump

logger = logging.getLogger("node")

class NativeHCRunProcess(HCRunProcess):
    """
    Process that keeps network access and dumper for a native client alive.

    Native HCs are bound to a vlan interface.
    To "start" a native HC means to allow network access via the interface
    and record the traffic.

    As a regular user does not have the permissions to (de-)configure
    interfaces record traffic from it, these steps are underdone by a
    bash-script.
    """

    SLEEP_STEP = 1

    def __init__(self, native_hc_program, hc, timeout,
                 io, work_dir, mitmproxy_script=None):
        """
        Constructor.
        @param native_hc_program: A program to (dis-)allow and record traffic
                                  of native HCs, will be sudo-executed.
        @param hc: An HCControl object.
        @param io: QueuePair from which to get commands and in which to feed
                   errors and results.
        @param timeout: When the subprocesses will be terminated, in seconds.
        @param work_dir: Where the dump file will be placed.
        """

        HCRunProcess.__init__(self, timeout=abs(timeout), hc=hc)

        self.native_hc_program = native_hc_program
        self.dump_file = ''
        self.dump_subprocess = None
        self.work_dir = work_dir
        self.io = io
        self.mitmproxy_script = mitmproxy_script

    def __del(self):
        """Destructor."""

        # TODO ensure files (from __prepare_*) are removed.
        pass

    def __prepare_control_handle(self):
        """Create the file which stops the client, if deleted."""

        if not path.isdir(self.work_dir):
            makedirs(self.work_dir)
        # TODO Watch out this one for leaking/inheriting open descriptors.
        # TODO Close the ctrl handle.
        hdl, fl = mkstemp(prefix=path.basename('%s_' %
                                                     self.hc.hcsid),
                                              suffix='.survivor',
                                              dir=self.work_dir)
        self.ctrl_handle, self.survivor_file = hdl, fl

    def __prepare_dump_file(self):
        """Prepare the (temporary) dump file."""

        if not path.isdir(self.work_dir):
            makedirs(self.work_dir)

        # TODO Watch out this one for leaking/inheriting open descriptors.
        # TODO Close the dump handle.
        self.dump_handle, self.dump_file = mkstemp(prefix=path.basename('%s_' %
                                                   self.hc.hcsid),
                                                   suffix='.dump',
                                                   dir=self.work_dir)

    def __proxy_setup_cmd(self):
        """Prepare the setup to let traffic go to and through proxy."""
        pass

    def __dumper_cmd(self):
        """
        Creates command for Popen to set up native client communication and dumps.
        @return: List to be used with Popen.
        """

        cmd = [ "sudo",
                self.native_hc_program,
                self.hc.interface,
                self.survivor_file, # Cut native connection if this file is deleted.
                self.dump_file      # Expect the dump in this file.
              ]
        return cmd

    def run(self):
        """
        Invoked on start() - call.
        Starts a script that allows a native HC to talk to the net and
        records its conversations.
        After timeout, remove a file which will cause the script to end.
        """

        try:
            self.__prepare_control_handle()
            self.__prepare_dump_file()
            cmd = self.__dumper_cmd()
        except Exception as e:
            logger.critical("Could not initialize run of Native HC: %s" % str(e))
            self.io.add_err(e)
            # Close any handles
            #self.dump_hanlde, self.ctrl_handle
            return

        logger.debug("Allowing %d seconds of traffic for native HC with: %s" % (self.timeout, cmd))

        # Start dumper subprocess.
        try:
            self.dump_subprocess = subprocess.Popen(cmd,
                                                stdin=subprocess.PIPE,
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE,
                                                close_fds=True)
        except Exception as e:
            logger.critical("Dumper process could not be started.")
            self.io.add_err(e)
            # Close any handles
            return

        sleep(0.8)

        if hasattr(self.hc, 'proxy_ip') and self.hc.proxy_ip \
           and hasattr(self.hc, 'proxy_port') and self.hc.proxy_port:
            self.proxy_dump = self.dump_file.replace(".dump", ".proxy")
            (pcmd, self.proxy_subprocess) = start_mitmdump(self.hc.proxy_port,
                                                           self.proxy_dump,
                                                           self.mitmproxy_script,
                                                           self.hc.proxy_ip)
            logger.debug("Starting mitmdump: %s" % " ".join(pcmd))
            sleep(1) # Give the proxy time to open the port
        else:
            self.proxy_dump = None

        # Init time.
        self.start_time()

        while not self.time_over():
            self.dump_subprocess.poll()
            # TODO Also poll and check proxy process, if any.
            if self.dump_subprocess.returncode is not None:
                # TODO escalate this
                logger.error("Dumper Process seemed to have died all by itself.")
                logger.warning("Exit code: %d" % self.dump_subprocess.returncode)
                self.io.add_err(Exception("Native Script died."))
                break
            try:
                cmd = self.io.get_command()
                # If command is given, this is the command to stop.
                break
            except Empty:
                pass

            sleep(NativeHCRunProcess.SLEEP_STEP)

        logger.info("Dumper Process: Timeout hit.")

        # Script shall stop when survivor_file is removed.
        # (Termination of processes is not possible, due to sudo/permissions).
        remove(self.survivor_file)
        if self.proxy_dump:
            if self.proxy_subprocess:
                self.proxy_subprocess.terminate()
        sleep(0.5)
        if self.proxy_dump:
            if self.proxy_subprocess:
                self.proxy_subprocess.kill()

        self.cleanup()
        if self.proxy_dump:
            self.io.add_result((self.dump_file, self.proxy_dump))
        else:
            self.io.add_result(self.dump_file)
        try:
            close(self.dump_handle)
        except:
            pass
        try:
            close(self.control_handle)
        except:
            pass

    def __terminate(self, process):
        """
        @param process: The process to terminate.
        @return: The subprocesses return code.
        """
        # TODO Rework this for the "native hc" case

        state = process.poll()
        # TODO Send quit command to process.
        #process.stdin.write('quit\n\r')
        return
        if state == None:
            process.terminate()
        waited = 0
        while state == None and waited < 10:
            state = process.poll()
            if state:
                break
            sleep(1)
            waited += 1
        if state == None:
            logger.warning("Dumper subprocess not terminated after 10 seconds.")
            process.kill()
        return state

    def get_result(self):
        """Get result(s) from queue."""
        return self.io.get_result()

    def stop(self):
        """Immediately stop all subprocesses."""
        self.io.add_task("stop")
        self.stop_processes()

    def stop_processes(self):
        """Big red button."""
        # TODO Rework this for the "native hc" case

        # Doesnt have this attribute from caller side.
        try:
            self.io.add_task("stop")
            remove(self.survivor_file)
        except:
            pass
        return
        if self.dump_subprocess is not None:
            self.dump_subprocess.terminate()
        if self.dump_subprocess is not None:
            self.dump_subprocess.kill()
            #join()?

    def cleanup(self):
        """Remove control file (but not dump)."""

        #if path.isfile(self.survivor_file):
        #    remove(self.survivor_file)
        logger.debug('Cleaning up native hcrun process, leaving dump file.')

