# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Node- centric classes.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
# Andre Heinecke <aheinecke@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import ConfigParser
import logging
import multiprocessing
from os import path, listdir, remove, makedirs
import shutil
from Queue import Empty
from shutil import copy
import socket
import subprocess
import time
import tarfile
import traceback
import gzip

from mapwoc.nodectrl.analyserprocess import AnalysisProcess
from mapwoc.nodectrl.hcinteractionscript import eventmap_from_string
from mapwoc.nodectrl.kvmhcrunprocess import KVMHCRunProcess
from mapwoc.nodectrl.nativehcrunprocess import NativeHCRunProcess

from mapwoc.lib.hc import HC
from mapwoc.lib.hccontrol import HCControl
from mapwoc.lib.scanconfiguration import VisitorConfiguration, VisitorBundle
from mapwoc.lib.taskqueuepair import TaskQueuePair
from mapwoc.lib.result import Result
from traceback import format_exc

def handleError(self, record):
  raise
logging.Handler.handleError = handleError

logger = logging.getLogger("node")

class NodeState(object):
    """Representing a state of a Node as communicated to other components."""

    IDLE,RUNNING,MANUAL,ANALYSING,GRACEFUL_STOPPING,ABORTING,ERROR = range(1,8)
    strstate = {IDLE:'idle',
                RUNNING:'running',
                MANUAL:'manual',
                ANALYSING:'analysis',
                GRACEFUL_STOPPING:'stopping',
                ABORTING:'aborting',
                ERROR:'error'}

    def __init__(self):
        """Inits state to IDLE, empty runlist."""

        self.state = self.IDLE
        self.runlist = []
        self.err_msg = ''

    def set(self, new_state, err_msg = ''):
        """
        Sets state to new state.
        @param new_state: New state.
        @param err_msg: Error message (if state is ERROR).
        """

        self.state = new_state
        self.err_msg = err_msg

    def export(self):
        """
        Export state such that it can be easily send through wire as json.
        @return: Dictionary describing the state as expected by the master
                 module.
        """
        if self.state != NodeState.ERROR:
            return {'' : self.strstate[self.state],
                    'runlist' : self.runlist }
        else:
            return {'' : self.strstate[self.state],
                    'runlist' : self.runlist,
                    'err_msg' : self.err_msg }

class Node:
    """The NodeCtrl."""

    VERSION = "1.3.3"
    SPAWN_POLICY_DIRECT = 'direct'
    SPAWN_POLICY_SPREAD = 'spread'

    def __init__ (self, hcsdir, work_dir, result_dir, num_analyser,
                  max_gb_ram, analyser_program, analyser_conf, native_gate_script,
                  proxy_ports, mitmproxy_script=None):
        """
        Initializes an empty Node, read in config and hc config.
        @param hcsdir: Directory in which to look for HCControl files.
        @param work_dir: Directory in which to put temporary results.
        @param result_dir: Directory in which to put results.
        @param num_analyser: Number of parallel analyser-processes to use (untested).
        @param max_gb_ram: For feasibility-of-scan checks, the maximum sum of
                           RAM allowed to be consumed by HCs.
        @param analyser_program: Path to program used as analyser.
        @param analyser_conf: Path to configuration file for analyser.
        @param native_gate_script: Path to script to use as native HC gatekeeper.
        @param proxy_ports: Range of ports (e.g. 10000-20000) that the proxy can use.
        @param mitmproxy_script: (request-)Script to pass to the mitmproxy.
        """

        self.__hcsdir = hcsdir
        self.__work_dir = work_dir
        self.__result_dir = result_dir
        self.__num_analyser_processes = num_analyser
        self.__ressource_mb_ram = 1024 * max_gb_ram
        self.__analyser_program = analyser_program
        self.__native_hc_program = native_gate_script
        self.scanrunid = -1
        try:
            self.proxy_min = int(proxy_ports.split("-")[0])
            self.proxy_max = int(proxy_ports.split("-")[1])
            if (self.proxy_max - self.proxy_min) <= 0:
                raise Exception()
        except:
            raise ValueError("'proxy_ports' config option is in the wrong format"
                             "\n format is: <min_port>-<max_port>")
        self.next_proxy_port = self.proxy_min

        self.__hcs = [] # Available honeyclients, HCControls
        self.__hc_instances = [] # Running (kvm) child processes

        # State
        self.state = NodeState()

        self.__load_hcs()
        self.base_vnc_port = 1
        self.next_vnc_port = self.base_vnc_port # For debuggin purposes, vnc port.
        self.nr_vnc_ports = 99 # Limit to max. 99 open vnc ports.

        self.scan_configuration = None
        self.visitor_configuration = None # Part of the scanconfiguration
        self.__analyser_io = TaskQueuePair()

        # Start as many analyser-processes as specified in the config file.
        self.__analyser_instances = []
        for i in range(0, self.__num_analyser_processes):
            self.__analyser_instances.append(AnalysisProcess(
                                                  self.__analyser_io,
                                                  self.__work_dir,
                                                  self.__result_dir,
                                                  self.__analyser_program,
                                                  analyser_conf))
        for analyser in self.__analyser_instances:
            analyser.start()

        if not path.isdir(self.__work_dir):
            makedirs(self.__work_dir)
        if not path.isdir(self.__result_dir):
            makedirs(self.__result_dir)

        for _file in listdir(self.__work_dir):
            logger.warning("Remaining %s in working directory (%s): ignoring." % (_file,
                path.abspath(self.__work_dir)))

        for _file in listdir(self.__result_dir):
            logger.warning("Remaining %s in result directory (%s): ignoring." % (_file,
                path.abspath(self.__result_dir)))
        self.spawn_policy = self.SPAWN_POLICY_SPREAD
        self.mitmproxy_script = mitmproxy_script

    def __state(self):
        return self.state.state

    def get_hcs(self):
        """
        Accessor to hcs (HCControl).
        @return: HCControls.
        """

        return self.__hcs

    def get_hc(self, hcsid):
        """
        Accessor for a HCControl with given hcsid.
        @return: The HCControl, or None if not found.
        """

        for h in self.__hcs:
            if h.hcsid == hcsid:
                return h

        return None

    def get_available_ram(self):
        """Accessor for __ressource_mb_ram."""
        return self.__ressource_mb_ram

    def num_running_hcsid(self, hcsid):
        """
        Count how many HC with a certain hcsid are running.
        @param hcsid: The hcsid to search for.
        @return: Number of HCs with given hcsid.
        """

        # Equivalent to
        #sum([1 for hcrunproc in self.__hc_instances if hcrunproc.hc.hcsid == hcsid])

        cnt = 0
        for hcrunproc in self.__hc_instances:
            if hcrunproc.hc.hcsid == hcsid:
                cnt += 1
        return cnt

    def advance_proxy_port(self):
        """Sets next_proxy_port to the next available proxy port."""
        if self.next_proxy_port >= self.proxy_max:
            self.next_proxy_port = self.proxy_min
        else:
            self.next_proxy_port += 1

    def advance_vnc_port(self):
        """Sets next_vnc_port to the next available proxy port."""
        if self.next_vnc_port >= self.base_vnc_port + self.nr_vnc_ports:
            self.next_vnc_port = self.base_vnc_port
        else:
            self.next_vnc_port += 1

    def start_hc(self, hc, visitorconfig, offset=0):
        """
        Starts a KVM HC and returns its process (or, in case of native
        HCs, allow traffic generation and record it).
        Increases the next vnc port.

        @param hc: The HCControl needed to start the process.
        @param visitorconfig: The VisitorConfig, needed to determine timeout.
        @param offset: The initial offset (seconds to wait until to start
                       subprocess, for kvm only).
        @return: A started (KVM or Native)HCRunProcess.
        """

        if hc.max_parallel > 0:
            curr = self.num_running_hcsid(hc.hcsid)
            if curr >= hc.max_parallel:
                logger.error("Cannot start HC: max_parallel constraint violation.")
                raise Exception("Cannot start HC (max_parallell constraint violation.")

        if hc.hctype == 'kvm' or hc.hctype == 'kvm-mac':
            eventmap = eventmap_from_string(visitorconfig.eventmap, hc)
            kvm_process = KVMHCRunProcess(hc,
                                          eventmap,
                                          abs(visitorconfig.active_window),
                                          TaskQueuePair(),
                                          self.__work_dir,
                                          initial_offset=offset,
                                          proxy_port=self.next_proxy_port,
                                          mitmproxy_script=self.mitmproxy_script)
            if self.scan_configuration.enable_vnc:
                kvm_process.vnc_port = self.next_vnc_port
                self.advance_vnc_port()
            self.advance_proxy_port()

            kvm_process.start()
            self.__hc_instances += [kvm_process]
            return kvm_process
        elif hc.hctype == 'native':
            dumper_process = NativeHCRunProcess(self.__native_hc_program,
                                                hc,
                                                visitorconfig.active_window,
                                                TaskQueuePair(),
                                                self.__work_dir,
                                                mitmproxy_script=self.mitmproxy_script)
            dumper_process.start()
            self.__hc_instances += [dumper_process]
            return dumper_process
        else:
            logger.error("Could not start HC of unknown type %s." % hc.hctype)
            return None

    def stop_all(self):
        """Kill all processes, the hard way."""

        self.state.set(NodeState.ABORTING)
        while len(self.__hc_instances) != 0:
            instance = self.__hc_instances.pop()
            instance.stop()
            instance.cleanup()
            # Do we need to join these processes?

            # TODO We now have dumps that should be saved and analysed somewhere.
            # Create an analyser instance / put in analysers job queue
        self.visitor_configuration = None
        self.state.set(NodeState.ANALYSING)
        self.state.runlist = []


    # TODO merge job_feasible
    def ram_requirements(self, visitorconfig):
        ram_used = 0
        for vb in visitorconfig.visitors:
            hc = self.get_hc(vb.hcsid)
            if hc == None:
                logger.warning("While checking ram requirement, could not"
                               " find hc %s." % vb.hcsid)
                continue

            # Check RAM usage.
            ram_used += int(hc.image_memory) * int(vb.number)
        return ram_used


    def job_feasible(self, visitorconfig):
        """
        Checks whether the given job is feasible for this node.
        Currently three measures are checked:
         - The sum of ram requirements of the given HCs has to be is less
           than the configured maxram in the configuration file.
         - The total numbers of HCs to start that are configured to use a
           proxy has to be lower than the number of available ports for
           proxies.
         - The max_parallel field of HCs cannot be disrespected.

        @param visitorconfig: The Configuration to check for feasibility.
        @return: RAM left if job would be started (negative if not feasible):
                 Additionally, 'symbolic' negative returns for misconditions.
        """

        ram_used = 0
        nr_visitors = 0

        # TODO Raise exceptions instead of symbolic negative returns.
        for vb in visitorconfig.visitors:
            hc = self.get_hc(vb.hcsid)
            if hc == None:
                logger.warning("While checking jobs feasibility, could not"
                               " find hc %s." % vb.hcsid)
                continue

            # Check RAM usage.
            ram_used += int(hc.image_memory) * int(vb.number)
            # Accumulate numbers for checking available proxy ports.
            if hasattr(hc, 'proxy_ip') and  hc.proxy_ip:
                nr_visitors += int(vb.number)
            # Check whether max_parallel field of HC is respected.
            if int(hc.max_parallel) > 0 and int(hc.max_parallel) > int(vb.number):
                logger.warning("Job not feasible, due to violation"
                               " of max_parallel constraint of HC %s." % hc.hcsid)
                return -2
        logger.debug("Feasibility-check of job: takes %d, have %d RAM." % (
                     ram_used, self.__ressource_mb_ram))
        if nr_visitors > self.proxy_max - self.proxy_min:
            logger.warning("Job has too many visitors (have fewer ports for proxies.")
            return -1
        return  self.__ressource_mb_ram - ram_used

    def start_job(self, scanconfig):
        """
        Starts a job, running given numbers of given hcsid with given
        livespan until stop is called.

        @param scanconfig: The scanconfiguration to use.
        @return: True if job was started, False otherwise.
        """

        # If already scanning, cannot start a new one.
        if self.visitor_configuration:
            logger.debug("Cannot start new scan: already scanning.")
            return False

        next_off = 0.0

        if scanconfig.visitor_configuration is None:
            logger.debug("Cannot start new scan: No VisitorConfiguration given.")
            return False
        self.scan_configuration = scanconfig
        self.visitor_configuration = scanconfig.visitor_configuration
        for vb in self.visitor_configuration.visitors:
            logger.info("Starting job with hcsid: %s" % vb.hcsid)
            hc = self.get_hc(vb.hcsid)
            if hc == None:
                logger.warning("Could not find HC (%s), part of a job, to start." % vb.hcsid)
                continue
            try:
                for i in range(vb.number):
                    hcrun_process = self.start_hc(hc,
                                                  self.visitor_configuration,
                                                  next_off)

                    # Trivial approach to SPAWN_POLICY_SPREAD: Spawn with 2 secs. pause.
                    if self.spawn_policy == self.SPAWN_POLICY_SPREAD:
                        next_off += 2.0
            except:
                logger.warning("Error while starting a hcrun-process.")
                # TODO Cleanup, ERROR state etc.
                traceback.print_exc()
                return False

        self.state.set(NodeState.RUNNING)
        self.__update_runlist()
        return True

    def stop_run(self, graceful=False):
        """
        Stops a current running job.
        @param graceful: If True, wait for HCs to time out.
        """

        # Remove all conf, cleanup, set state, etc.
        logger.info("Will stop run.")
        if self.state.state is not NodeState.IDLE:
            if graceful:
                logger.debug("Graceful stopping.")
                self.state.set(NodeState.GRACEFUL_STOPPING)
            else:
                self.state.set(NodeState.ABORTING)

        if not graceful:
            for hcrunprocess in self.__hc_instances:
                hcrunprocess.stop()
            # TODO cleanup
        self.visitor_configuration = None

    def get_result_dir(self):
        """Returns the location where the results are stored."""
        return path.abspath(self.__result_dir)

    def __load_hcs(self):
        """Add HCs for which a config file is found in config path."""

        if not path.exists(self.__hcsdir):
            logger.warning("The HC directory %s does not exist." % self.__hcsdir)
            return

        for rel_hc_file in listdir(self.__hcsdir):
            abs_hc_file = path.normpath(path.join(self.__hcsdir, rel_hc_file))
            if not abs_hc_file.endswith(".hc"):
                logger.warning(
                            "HCControl file does not follow naming scheme: %s"
                             % abs_hc_file)

            try:
                new_hc = HCControl(abs_hc_file)
            except Exception as e:
                logger.critical("Could not load hc from file %s: %s"
                                % (abs_hc_file, e.args[0]))
                exit(6)
            if len([hc for hc in self.__hcs if hc.hcsid == new_hc.hcsid]) > 0:
                logger.critical("Two configurations for hcsid %s "
                                "in HC directory %s found, exiting." %
                                (new_hc.hcsid, self.__hcsdir))
                exit(5)
            self.__hcs += [new_hc]


    def get_hcsids(self):
        """
        Access list with hcsids of available (startable) honeyclients.
        @return: List with hcsids of available (startable) honeyclients.
        """

        answer = [hc.hcsid for hc in self.__hcs]
        return answer

    def __runs_timed_scan(self):
        if not self.scan_configuration:
            return False
        return self.scan_configuration.visitor_configuration.active_window < 0

    def poll_hcs(self):
        """
        Look for results of hcrun processes and look at the processes itself.
        Repopulate state.

        Restart finished HCs, if they stopped and own state is d'accord.
        """

        # TODO What if the instances "die out" (e.g. become zombies)?
        # if len(self.__hc_instances) != aim ...
        logger.debug("Have %d instances running." % len(self.__hc_instances))

        # Iterate over processes, reverse order, and remove finished ones.
        cnt_unfinisheds = 0
        restart_hcs = self.__state() != NodeState.GRACEFUL_STOPPING\
                      and self.__state() != NodeState.ABORTING\
                      and self.__state() != NodeState.ERROR\
                      and not self.__runs_timed_scan()

        for i in range(len(self.__hc_instances)-1, -1, -1):
            hcrun_process = self.__hc_instances[i]
            hc = hcrun_process.hc
            try:
                # TODO Process might die without putting stuff in result queue.
                #      Check processes time to live and state, too.
                # Check for content result_queue (raises Empty, if not).
                r = hcrun_process.get_result()

                # Tell it to clean up, store result, start analysis.
                # If we had a timed scan, now go over to stop,analyse,idle

                if isinstance(r, tuple):
                    logger.info("HC %s finished. Dump: %s ." % (hc.hcsid, r[0]))
                else:
                    logger.info("HC %s finished. Dump: %s ." % (hc.hcsid, r))
                hcrun_process.stop()
                hcrun_process.cleanup()
                self.__analyser_io.add_task(r)

                # Restart HC if it should be restarted.
                if restart_hcs:
                    # TODO Re-respect the spawn policy and offsets.
                    n_hc_process = self.start_hc(hc, self.visitor_configuration)
                    logger.info("Restarting a HC.")
                else:
                    logger.info("Not Restarting a HC because stopping/aborting or timed scan.")
                del self.__hc_instances[i]
            except Empty:
                cnt_unfinisheds += 1
                if self.__state() == NodeState.ABORTING:
                    hcrun_process.stop()
            except Exception as e:
                # KVMHCRun ran into an error. Report it and go to error state.
                logger.critical("One HC has a serious issue: %s" % e)
                self.stop_all()
                self.state.set(NodeState.ERROR)
                self.state.err_msg = str(e)

        # Log running HCs.
        if cnt_unfinisheds == 1:
            logger.debug("One HC did not yet deliver results.")
        elif cnt_unfinisheds > 1:
            logger.debug("%d HCs did not yet deliver results." % cnt_unfinisheds)

        # TODO Put analyser-result-fetching in own method.
        # Pull and pack as many results as possible.
        try:
            results = self.__analyser_io.get_all_results()
            for result in results:
                self.pack_result(result)
            if len(results):
                logger.info("Have %d results." % len(results))
        except Empty:
            logger.debug("No Analysis results.")
        except Exception as e:
            logger.critical("Error during analysis: %s" % str(e))
            self.state.set(NodeState.ERROR)
            self.state.err_msg = str(e)

        self.__update_runlist()

    def pack_result(self, result):
        """
        Adds additional meta information to a result and
        moves it into a directory for publishing.

        @param result: Tarball that should be extended and published.
        """

        # It is not possible to append to gzip'ed tarballs in python.
        tgz = tarfile.open(name=result, mode='a')
        metafile = path.join(self.__work_dir, "metainfo")
        if path.isfile(metafile):
            logger.warning("Metafile %s already exists, deleting...")
            remove(metafile)
        with open(metafile, "w") as fptr:
            parser = ConfigParser.RawConfigParser()
            parser.add_section("scan")
            parser.set("scan", "time", time.time())
            parser.set("scan", "scanrunid", self.scanrunid)
            parser.write(fptr)

        tgz.add(metafile, arcname="metainfo")
        tgz.close()
        f_in = open(result, 'rb')
        f_out = gzip.open(result + ".gz", 'wb')
        f_out.writelines(f_in)
        f_out.close()
        f_in.close()
        remove(result)
        result = result + ".gz"
        remove(metafile)
        shutil.move(result, result.replace(self.__work_dir, self.__result_dir))

    def __update_runlist(self):
        """
        Updates runlist in state dictionary.
        Also, updates general state.
        """

        self.state.runlist = [[kvm_proc.hc.hcsid, kvm_proc.vnc_port] for kvm_proc in self.__hc_instances]
        if len(self.__hc_instances) == 0 and (self.visitor_configuration == None or self.scan_configuration.visitor_configuration.active_window < 0):
            # Set state to idle if no analysis running.
            logger.debug("Runlist update.")
            if self.__analyser_io.all_empty():
                # TODO Implement a way to find out how many things still to
                #      analyse, how many finished etc.
                self.visitor_configuration == None
                if self.__state() != NodeState.ERROR:
                    self.state.set(NodeState.IDLE)
            else:
                if self.__state() != NodeState.ERROR:
                    self.state.set(NodeState.ANALYSING)

    def __print_hcs(self):
        """Prints available and running HCs to stdout."""

        print "Available HCs:"
        for h in self.__hcs:
            print "  " + h.hcsid + " (" + h.name + ")"
        print
        print "Running HCs:"
        for h in self.__hc_instances:
            print "  " + h.hcsid + " (" + h.name + ")"
