# -*- coding: utf-8 -*-

# mapWOC
#
# Description: A NodeCtrl with twisted.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from ConfigParser import NoSectionError
import json
import logging
import traceback

from twisted.internet import reactor, task
from twisted.internet.protocol import Factory, Protocol
from twisted.protocols.basic import LineReceiver

from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook
from mapwoc.lib.protocolutil import json_trsuccess, json_trerror
from mapwoc.lib.toolutil import MapwocOptionParser

from mapwoc.nodectrl.node import Node

# TODO There is severe overlap between this and the master implemenation.
#      Eventually, it might be refactored/merged.

logger = logging.getLogger('nodectrlsrv')

class Requests:
    """Request type 'enum'."""
    STATE = 1002
    HCS = 1003
    START_RUN = 1004
    STOP_RUN = 1005
    RESULT_DIR = 1006
    VERSION = 1010
    CHECK_RESSOURCES = 1011
    STOP_SMOOTH = 1012

cmd_request_map = {'state':Requests.STATE,
                   'hcs':Requests.HCS,
                   'startrun':Requests.START_RUN,
                   'stoprun':Requests.STOP_RUN,
                   'stopsmooth':Requests.STOP_SMOOTH,
                   'version':Requests.VERSION,
                   'checkjob':Requests.CHECK_RESSOURCES,
                   'resultdir':Requests.RESULT_DIR
                  }


def which_cmd(text):
    """Returns Requests from string."""

    return cmd_request_map.get(text)


class NodeRequest(LineReceiver):
    """Handle one (control or info) request."""

    def connectionMade(self):
        """Just prepare for data coming."""

        self.text = ''
        self.cmd = None
        self.nr_lines_received = 0

        session = self.transport.socket
        try:
            peer_name = session.peer_certificate.subject
        except AttributeError as err:
            peer_name = 'Unknown'
        try:
            if hasattr(session, "verify_peer"):
                session.verify_peer()
        except Exception as e:
            logger.warning("Could not verify peers certificate: %s" % str(e))
        logger.debug('Connection to node from peer "%s" (%s).' % (
                                                 peer_name, self.transport.getPeer()))

    def lineReceived(self, line):
        """
        Callback, received a line.
        React upon it and answer, or wait for further data (lines).
        @param line: The line of text (part of request) received.
        """
        self.text += line
        self.nr_lines_received = self.nr_lines_received + 1

        if not self.cmd:
            self.cmd = which_cmd(self.text.strip())

            if self.cmd is Requests.STATE:
                answer = self.factory.node.state.export()
                self.__answer_json(answer)
            elif self.cmd is Requests.HCS:
                answer = self.factory.node.get_hcs()
                self.__answer_json(answer)
            elif self.cmd is Requests.START_RUN:
                self.text = ''
            elif self.cmd is Requests.CHECK_RESSOURCES:
                self.text = ''
            elif self.cmd is Requests.STOP_SMOOTH:
                self.factory.node.stop_run(graceful=True)
                #TODO Return estimated time till done.
                self.__answer("ok, will take soandso many seconds")
            elif self.cmd is Requests.STOP_RUN:
                self.factory.node.stop_run()
                #TODO Return something meaningful
                self.__answer("Meaningless")
            elif self.cmd is Requests.VERSION:
                self.__answer(self.factory.node.VERSION)
            elif self.cmd is Requests.RESULT_DIR:
                self.__answer(self.factory.node.get_result_dir())
            else:
                logger.warning("Unknown command: " + self.text)
                self.transport.loseConnection()
        else:
            # More data is coming
            if self.cmd is Requests.START_RUN:
                # Expect two lines: First is integer (scanrun.id), second
                # line is json ScanConfiguration.
                if self.nr_lines_received == 2:
                    self.factory.node.scanrunid = int(self.text)
                    self.text = ''
                else:
                    scan_config = json.loads(self.text, object_hook=mapwoc_from_json_hook)
                    if self.factory.node.start_job(scan_config):
                        self.__answer(json_trsuccess())
                    else:
                        self.factory.node.scanrunid = -1
                        self.__answer(json_trerror(500, "Node did not start the job"))
            elif self.cmd is Requests.CHECK_RESSOURCES:
                visitorconf = json.loads(self.text, object_hook=mapwoc_from_json_hook)
                ram_left = self.factory.node.job_feasible(visitorconf)

                if ram_left >= 0:
                    self.__answer(json_trsuccess())
                else:
                    # TODO Use exception instead.
                    if ram_left == -2:
                        self.__answer(json_trerror(102, "max_parallel violation"))
                    elif ram_left == -1:
                        self.__answer(json_trerror(103, "not enough proxy ports"))
                    else:
                        self.__answer(json_trerror(101,
                                "RAM: %dMB/%dMB (%dMB missing)" % (
                               self.factory.node.ram_requirements(visitorconf),
                               self.factory.node.get_available_ram(),
                               -ram_left)))
            else:
                logger.error("Unexpected data.")
                self.transport.loseConnection()

    def connectionLost(self, reason):
        """
        Callback when connection lost.
        @param reason: Ignored.
        """
        pass

    def __answer(self, answer):
        """
        Write string answer to wire and close connection.
        @param answer: Answer to send as is.
        """

        self.transport.write(answer)
        self.transport.loseConnection()

    def __answer_json(self, answer):
        """
        Sends json encoded answer through the wire and closes the
        connection afterwards.
        @param answer: Answer to encode and send.
        """

        json_str = MapWOCJSONEncoder().encode(answer)
        self.__answer(json_str)


class NodeResponseFactory(Factory):
    """Factory producing NodeRequest-protocols."""

    protocol = NodeRequest

    def __init__(self, node):
        """
        @param node: The Node object.
        """
        self.node = node

