# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Classes related to processes that can time out.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import multiprocessing
import subprocess
import logging
from Queue import Empty

import time
from time import sleep

logger = logging.getLogger("node")

class TimedProcess(multiprocessing.Process):
    """
    Base class for a Process that keeps track of time with some convenience
    methods to access (time) progression.
    """

    def __init__(self, timeout):
        """
        TimedProcess has to be initialized with a call to start_time().
        @param timeout: The timeout in seconds.
        """

        if timeout < 0:
            raise Exception("Timeout for TimedProcess must be positive.")

        multiprocessing.Process.__init__(self)
        self.__start_time = None
        self.timeout = timeout

    def elapsed_time(self):
        """Calculates the elapsed time since start."""

        if not self.__start_time:
            raise Exception("TimedProcess had not been started via start_time().")

        return time.time() - self.__start_time

    def time_to_live(self):
        """Calculates the time (left) to live."""

        return self.timeout - self.elapsed_time()

    def time_over(self):
        """
        @return: True if timeout is hit.
        """

        return self.time_to_live() < 0

    def start_time(self):
        """Starts the internal timer."""

        if self.__start_time:
            raise Exception("start_time can only be called once")

        self.__start_time = time.time()

