# -*- coding: utf-8 -*-

# Description: This is the implementation of a mapwoc-Redirector.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt fur Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

# Using Threads and the servers as done here is a bad design decision, but ok for prototype.
# Use multiprocessing in the future, and/or a twisted-like approach.

from copy import deepcopy
import logging
import socket
import threading
import time
import traceback

from sqlalchemy import *
from sqlalchemy.orm import create_session

from mapwoc.lib.url import URL, URLList, validate_URL
from mapwoc.lib.redirectorvisit import RedirectorVisit
from mapwocgui.model.meta import Session
import mapwoc.lib.model.url

from mapwoc.redirector.redirectorserver import RedirectorServer
from mapwoc.redirector.redirectorctrlserver import RedirectorCtrlServer

logger = logging.getLogger('redirector')

class RedirectorState:
    """State 'enum'."""

    IDLE = 0 # Not serving URLs
    RUNNING = 200 # Serving URLs
    DONE = 201 # If looping, serving URLs, served each already at least once.
    LOADING = 300 # Loading URL List
    ERROR = 400 # Error occurred


class RedirectorMode:
    """Operation-mode 'enum'."""

    # Note: COMPREHENSIVE mode not implemented.
    COMPREHENSIVE = 100 # One list per hcsid.
    SAMPLE = 200        # All hcsids draw from the same list.

def mapping_from_db(parameters):
    """
    Return a mapping from database.
    @param parameters: Comma-separated list with of database-ids of URL-lists to
                       load and path to database (last).
    @return: Sample-mode mapping. None in case of errors.
    """

    parameters = parameters.split(",")
    database = parameters.pop()

    try:
        logger.debug("Opening database %s" % database)
        engine = create_engine('sqlite:///%s' % database)
        c = engine.connect()
        dbsession = create_session(bind=engine)
    except:
        logger.warning("Unable to open database %s " % database)
        traceback.print_exc() 
        return None

    mapping = {}
    mlist = []

    hitter = 0
    for urllistid in parameters:
        if not urllistid: continue
        logger.debug("Grabbing URLs for urllist id: %d" % int(urllistid))
        urllist = dbsession.execute("SELECT url FROM url, urllist_url "
                                     "WHERE urllist_url.url_id = url.id"
                                     " AND urllist_url.urllist_id = %d;" % int(urllistid))
        for url in urllist:
            mlist.append(url[0])
            hitter +=1
            if hitter % 10000 == 0:
                logger.info("Hit %d (%s)" % (hitter, url[0]))

    dbsession.close_all()

    if not len(mlist):
        logger.warning("Could not load any urllists")
        return None

    mlist.reverse()

    mapping[''] = mlist
    return mapping


class Redirector:
    """Redirector, serves redirects and listens on controll-socket."""

    VERSION = "1.3.3"

    def __init__(self, redirector_interface=None, redirector_port=8338,
                 ctrl_interface='localhost', ctrl_port=8228, cred=None):
        """
        Prepares the servers to listen for http and control requests.

        @param redirector_interface: Interface on which to redirect.
        @param redirector_port: Port on which to listen for http-requests and
                                serve http-redirects.
        @param ctrl_interface: Interface on which to listen for commands.
        @param ctrl_port: Port on which ControlServer listens for commands.
        """

        self.mapping = {}
        self.serve_history = []
        self.redirect_port = redirector_port
        self.ctrl_port = ctrl_port
        r_iface = redirector_interface
        if r_iface is None:
            r_iface = socket.gethostname()
        self.redirectorsrv = RedirectorServer(self, r_iface,
                                              self.redirect_port)
        self.ctrlsrv = RedirectorCtrlServer(self, (ctrl_interface, self.ctrl_port), cred)
        logger.info('Redirector-Interface: %s, Ctrl-Interface: %s' % (r_iface, ctrl_interface))
        self.mode = RedirectorMode.SAMPLE
        self.state = RedirectorState.IDLE
        self.do_loop = True
        self.mapping_to_loop = deepcopy(self.mapping)
        self.__num_urls_per_request = 1

    def is_fast_mode(self):
        """Returns True if one url should be given per request."""
        return self.__num_urls_per_request != 1

    def set_num_urls(self, numurls):
        """Set the number of URLs answered per request."""
        self.__num_urls_per_request = numurls

    def get_num_urls(self):
        """Return the number of URLs per request."""
        return self.__num_urls_per_request

    def start(self, mapping=None) :
        """Starts the server with given mapping (blocks)."""

        self.mapping = mapping
        self.mapping_to_loop = deepcopy(self.mapping)
        self.__update_mode()

        try:
            redirectorsrv_thread = threading.Thread(
                                       target=self.redirectorsrv.serve_forever)
            redirectorsrv_thread.setDaemon(True)
            redirectorsrv_thread.start()
            logger.info("Redirector listening at port %d." % self.redirect_port)
            redirectorctrlsrv_thread = threading.Thread(
                                        target=self.ctrlsrv.serve_forever)
            redirectorctrlsrv_thread.setDaemon(True)
            redirectorctrlsrv_thread.start()
            logger.info("Redirector-Ctrl listening at port %d." % self.ctrl_port)
        except KeyboardInterrupt:
            print "kbint"
            pass

        redirectorsrv_thread.join()
        self.ctrlsrv.shutdown()
        redirectorctrlsrv_thread.join()


    def get_next_url(self, hcsid, peer):
        """Get the next URL for a given hcsid, None if none left."""
      
        # Allow queries that are identifiable by hcsid also in sample mode.
        if self.mode == RedirectorMode.SAMPLE:
            query_hcsid = ''
        else:
            query_hcsid = hcsid

        url = None
        try:
            url = self.mapping[query_hcsid].pop()
        except KeyError:
            self.state = RedirectorState.DONE
        except IndexError:
            if (self.mode == RedirectorMode.SAMPLE) and (self.do_loop == True):
                self.restart_with_mapping(self.mapping_to_loop)
                # Agreed that this can be done much nicer.
                url_l = self.mapping.get(query_hcsid)
                if url_l:
                    url = url_l.pop()
                self.state = RedirectorState.DONE

        # TODO That means an IDLE Redirector will become DONE after
        #      a request. Fix that.
        if self.mode == RedirectorMode.SAMPLE:
            remains = self.mapping.get('')
            if remains is None or len(remains) == 0:
                self.state = RedirectorState.DONE

        self.log_visit(hcsid, url, peer)
        if url is None:
            return None
        return url.strip()


    def peek_next_url(self, hcsid, n=1):
        """Look at the next URL for a given hcsid, empty list in case of errors."""

        try:
            if len(self.mapping) > 0:
                upcoming_urls = self.mapping[hcsid]
                if upcoming_urls is None:
                    return []
                last = upcoming_urls[-n:] 
                last.reverse()
                return last
            else:
                return []
        except KeyError:
            return []

    def stop(self):
        """Attempt to stop redirector server."""
        
        self.redirectorsrv.shutdown()


    def log_visit(self, hcsid, url, peer):
        """Adds an entry to the serve_history."""

        # Only log URLs.
        if url is not None:
            self.serve_history += [RedirectorVisit(hcsid, url, peer)]
            logger.info("Visit of %s: %s" % (hcsid, url))
        else:
            logger.debug("Non-Visit by %s." % hcsid)

    def go_idle(self):
        """Loads empty mapping and set own state to idle."""
        self.restart_with_mapping({'':[]})
        self.state = RedirectorState.IDLE

    def str_state(self):
        """Return a simple textual representation of current state."""

        if self.state == RedirectorState.IDLE:
            return 'idle'
        elif self.state == RedirectorState.RUNNING:
            return 'running'
        elif self.state == RedirectorState.LOADING:
            return 'loading'
        elif self.state == RedirectorState.DONE:
            return 'done'
        else:
            return 'error'


    def load_url_file(self, filepath):
        """
        Load an url list from file.

        @return: True if successfull (False otherwise).
        """

        self.state = RedirectorState.LOADING
        try:
            mapping = mapping_from_file(filepath)
            self.restart_with_mapping(mapping)
            self.state = RedirectorState.RUNNING
            return True
        except IOError:
            self.state = RedirectorState.ERROR
            return False


    def restart_with_mapping(self, mapping):
        """Sets a new mapping and state to RUNNING."""
        
        self.state = RedirectorState.LOADING
        self.mapping = mapping
        self.__update_mode()
        self.mapping_to_loop = deepcopy(mapping)
        self.state = RedirectorState.RUNNING
        # Reset the history.
        self.serve_history = []


    def __update_mode(self):
        """Sets mode variable according to the mapping."""

        if len(self.mapping.keys()) == 1 and '' in self.mapping.keys():
            self.mode = RedirectorMode.SAMPLE
        else:
            self.mode = RedirectorMode.COMPREHENSIVE

    def load_url_db(self, parameters):
        """
        Load URL(-List)s from a database.

        @param parameters: IDs of URLs to load from database, colon-separated.
                           Last element needs to be the database from which
                           to load those URLs.
        @return: True if successfull, False otherwise.
        """

        logger.debug("Parameters %s" % parameters)
        ul = mapping_from_db(parameters)
        if ul is not None:
            self.restart_with_mapping(ul)
        else:
            return False
        logger.warning("Restarted with mapping.")
        return True

    def get_state(self, number_last_urls=1, number_next_urls=1):
        """
        Access the current state of the redirector.

        @param number_last_urls: Number of last served URLs to include in
                                 the state.
        @param number_next_urls: Number of next URLs to serve to include in
                                 the state.
        @return: Returns a dictionary that represents the current state of
                 the redirector.
        """

        next_urls = []
        for nexturl in self.peek_next_url('', number_next_urls):
            if nexturl is not None:
                next_urls.append(nexturl)

        if self.mapping_to_loop and self.mapping_to_loop.get(''):
            n_total_urls = len(self.mapping_to_loop.get(''))
        else:
            n_total_urls = 0
        if self.mapping and self.mapping.get(''):
            n_left_urls = len(self.mapping.get(''))
        else:
            n_left_urls = 0
        return {
                'version' : self.VERSION,
                'n_urls' : n_total_urls,
                'n_urls_served' : n_total_urls - n_left_urls,
                'state' : self.str_state(),
                'time' : time.time(),
                'n_last_urls' : number_last_urls,
                'last_urls' : self.serve_history[-number_last_urls:],
                'n_next_urls' : number_next_urls,
                'next_urls' : next_urls,
                'mode' : ''
               }

