# -*- coding: utf-8 -*-

# Description: This is the implementation of the server for a redirector.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt fur Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import SocketServer
import logging
import re

from gnutls.crypto import *
from gnutls.connection import *

from mapwoc.lib.protocolutil import json_trerror, json_trsuccess
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder

logger = logging.getLogger('redirector')

# TODO Merge with the other request handlers, use commands and regexps.
class RedirectorCtrlServerReqHandler(SocketServer.BaseRequestHandler):
    """Handles control requests to the redirector."""

    def handle(self):
        """Handle a single request."""
        logger.debug("Redirector CTRL-Request")

        answer = "unknown command"
        if self.server.credentials:
            # Make it a server session.
            session = ServerSession(self.request, self.server.credentials)
            sock = session
            sock.handshake()
            session.verify_peer()
            self.server.credentials.check_certificate(session.peer_certificate,
                                                      cert_name='peer certificate')
        else:
            sock = self.request

        question = sock.recv(1024).strip()

        if question == 'next':
            answer = self.server.redirector.peek_next_url("")
            if answer == None:
                answer = ""
        elif question == 'state':
            answer = self.server.redirector.str_state()
        elif question == 'last':
            answer = ""
            for entry in self.server.redirector.serve_history:
                answer += "%s %s %s\n" % (entry.hcsid, entry.url, entry.date)
        elif question.startswith('loaddb'):
            parameters = question[7:]
            if self.server.redirector.load_url_db(parameters):
                answer = json_trsuccess()
            else:
                answer = json_trerror(400, "Could not load URL-List from DB.")
            
        elif question == 'shutdown':
            answer = "attemping shutdown"
        # TODO Use regexp here.
        elif (question.startswith('next ')) and (len(question) > len('next ')):
            answer = ""
            num = int(question[5:])
            for nexturl in self.server.redirector.peek_next_url('', num):
                if nexturl is not None:
                    answer += "%s\n" % nexturl
        elif (question.startswith('last ')) and (len(question) > len('last ')):
            num = int(question[5:])
            last_urls = self.server.redirector.serve_history[-num:] 
            answer = MapWOCJSONEncoder().encode(last_urls)
        elif question.startswith('load '):
            # Load an url file
            filepath = question[5:]
            if self.server.redirector.load_url_file(filepath):
                anser = "ok"
            else:
                answer = "error loading file"
        elif question.startswith('urllist'):
            # Clear the list of redirector
            # Refill the list for redirector
            urls = question.split("\n")
            urls.pop(0)
            urls.reverse()
            answer = "ok, x urls found."
            mapping = {'':urls}
            self.server.redirector.restart_with_mapping(mapping)
        elif question == "nrremaining":
            try:
                answer = str(len(self.server.redirector.mapping['']))
            except:
                answer = "0"
        elif question == 'percentage':
            total = self.server.redirector.mapping_to_loop.get('')
            left = self.server.redirector.mapping.get('')
            lentotal = len(total)
            if lentotal == 0:
                answer = "n/a"
            elif total is not None:
                if left is None:
                    left = []
                answer = str( 100 * (1.0 - float(len(left)) / len(total)))
            else:
                answer = "n/a"
        elif question.startswith('fastmode '):
            numurls = int(question[9:])
            self.server.redirector.set_num_urls(numurls)
            answer = 'ok'
        elif question == 'go_idle':
            self.server.redirector.go_idle()
            answer = 'ok'
        elif question == 'version':
            answer = self.server.redirector.VERSION
        elif question.startswith('state_ext'):
            rm = re.match("^state_ext (\d+) (\d+)$", question)
            if rm is None or len(rm.groups()) != 2:
                answer = MapWOCJSONEncoder().encode(
                                     self.server.redirector.get_state())
            else:
                answer = MapWOCJSONEncoder().encode(
                     self.server.redirector.get_state(int(rm.group(1)), int(rm.group(2))))

        sock.sendall(answer)
        if self.server.credentials:
            try:
                sock.shutdown()
            except:
                pass
            sock.close()

        # Aftermath (can probably be resolved earlier)
        if question == 'shutdown':
            self.server.redirector.stop()


class RedirectorCtrlServer(SocketServer.TCPServer):
    """Handles control requests to the redirector."""

    allow_reuse_address = 1

    def __init__(self, redirector, address, credentials=None):
        """
        Starts a SocketServer on given address with given redirector.
        @param redirector: Redirector object to allow requesthandler access to.
        @param address: The address to listen at.
        @param credentials: Credentials to us (None is fine, too).
        """

        SocketServer.TCPServer.__init__(self, address, RedirectorCtrlServerReqHandler)
        self.redirector = redirector
        self.allow_reuse_address = True
        self.credentials = credentials

