# -*- coding: utf-8 -*-

# Description: This is the implementation of a server for a redirector.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt fur Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import BaseHTTPServer
import logging

logger = logging.getLogger('redirector')

class RedirectHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """A RequestHandler that answers with redirects, taken from its server,
       based on the surfed location."""

    server_version = '(hidden)?'
    sys_version = '(hidden)!'
    
    # Template for the fastmode html/javascript. Substitute comma-separated
    # list of URLs ( FASTMODEPAGE % '"http://url.org", http://urls.org"' )
    FASTMODEPAGE="""
<html><head><script type="text/javascript">
function openWindowNoReferer (url) {
 w = window.open("", "_blank", "toolbar=yes, menubar=yes, scrollbars=yes, resizable=yes, status=yes, location=yes");
 w.document.open();
 w.document.writeln('<html><head><meta http-equiv="refresh" content="0; URL=' + url + '"></head><body>Loading: '+ url + '</body></html>');
 w.document.close(); }
 var urls = new Array(%(urlstring)s);
 for (n in urls) openWindowNoReferer(urls[n]);
</script></head>
<body>
%(urlmarkers)s
</body></html>"""
    #Instead of the http-equiv refresh method, we could try to issue
    # "document.location.href=" directly.

    def multi_redirect(self):
        """Dummy redirect for now. Serves what is stored in FASTMODEPAGE."""
        
        hcsid = self.path[1:]
        url = self.server.get_next_url(hcsid, self.client_address[0])
        if url == None:
            self.send_response(204)
            self.end_headers()
        else:
            numurls = self.server.redirector.get_num_urls()
            # Fetch other URLs and bring them into desired form
            urlstring = '"%s"' % url
            urls = []
            for i in range (1, numurls):
                urls.append('"%s"' % self.server.get_next_url(hcsid,
                                             self.client_address[0]))
            urlstring = "%s,%s" % (urlstring, ",".join(urls))
            urlmarkers = "MAPWOC_FASTMODE_URL %s\n" % urlstring
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            self.wfile.write('\n')
            self.wfile.write(self.FASTMODEPAGE % dict(urlstring=urlstring,
                                                      urlmarkers=urlmarkers))
            self.wfile.write('\n')


    def redirect(self):
        """
        Redirects (HTTP 302) to the next url for a given request
        which includes the hcsid. Responds with (HTTP 204) if no
        next URL for this hcsid known.
        """

        # Defend against favicon requests.
        if self.path[1:] == "favicon.ico":
            self.send_response(404)
            return

        if self.server.redirector.is_fast_mode():
            self.multi_redirect()
            return

        # Possible HTTP states
        # 410 - Gone, 412 - Precondition Failure, 423 - Locked
        # 204 - No Content

        # Path is "/<hcsid>".
        # First character in self.path is '/'
        hcsid = self.path[1:]

        url = self.server.get_next_url(hcsid, self.client_address[0])
        if url == None:
            self.send_response(204)
            # Apparently, sometimes pipe breaks here.
            self.end_headers()
        else:
            self.send_response(302)
            self.send_header("Location", url)
            self.end_headers()

        self.wfile.write('\n')
        self.wfile.close()

    def do_GET(self):
        self.redirect()

    def do_POST(self):
        self.redirect()

class RedirectorServer(BaseHTTPServer.HTTPServer):
    """Small HTTP-Server that serves URLs from a Redirectors mapping."""

    allow_reuse_address = 1

    def __init__(self, redirector, host='localhost', port=8338):
        """Intitializes (but does not start) a redirector."""

        BaseHTTPServer.HTTPServer.__init__(self, (host,port), RedirectHandler)
        self.redirector = redirector

    def get_next_url(self, hcsid, peer):
        """Gets next URL from servers redirector."""

        return self.redirector.get_next_url(hcsid, peer)

