import logging
from codecs import getdecoder

from pylons import request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect
from pylons.i18n.translation import _, set_lang
from pylons.decorators import validate
import mapwocgui.lib.helpers as h

import formencode

from pylons.controllers import WSGIController
from authkit.permissions import ValidAuthKitUser
from authkit.authorize.pylons_adaptors import authorized
import authkit.users

from mapwocgui.lib.base import render
from mapwocgui.model import Session
import mapwocgui.model.masterproxy as master
from mapwoc.lib.model.users import UserData
from mapwocgui.lib.auth import checkAccess, user_set_single_role
from mapwocgui.lib.forms import ChangeSettings, ChangePassword, CreateNewValidator

log = logging.getLogger(__name__)


class AccountController(WSGIController):
    def __before__ (self, action, **params):
        c.messages = master.get_messages()
        if 'lang' in session:
            set_lang(session['lang'])

    def signin(self):
        if not request.environ.get('REMOTE_USER'):
            abort(401)
        else:
            session['REMOTE_USER'] = request.environ.get('REMOTE_USER')
            auth_name = request.environ.get('REMOTE_USER').lower()
            if auth_name:
                user_data = \
                        Session.query(UserData).filter_by(user_name=auth_name).first()
                session['lang'] = user_data.language
                set_lang(session['lang'])
            session.save()
            return render('/account/signedin.mako')

    def signout(self):
        request.environ['paste.auth_tkt.logout_user']()
        session.clear()
        session.save()
        return render('/account/signedout.mako')

    @checkAccess("admin")
    def manage(self):
        c.users = Session.query(UserData).all()
        return render('/account/manage.mako')

    def set_lang(self, id):
        session['lang'] = id
        set_lang(session['lang'])
        auth_name = request.environ.get('REMOTE_USER').lower()
        if auth_name:
            user_data = \
                    Session.query(UserData).filter_by(user_name=auth_name).first()
            user_data.language = id
            Session.commit()
        session.save()
        response.status_int = 302
        response.headers['location'] = h.url_back(0)
        return "Moved temporarily"

    @validate(schema=ChangePassword(), form='settings', post_only=True)
    def change_pw(self):
        if h.authorized(h.is_admin) and self.form_result.get("username"):
            # Admin may change every password
            auth_name = self.form_result["username"]
        else:
            auth_name = request.environ.get('REMOTE_USER').lower()
        Session.flush() # < necessary for dear authkit
        request.environ['authkit.users'].user_set_password(auth_name,
                self.form_result['password'])
        Session.commit()
        h.good_flash(_(u"Password changed."))
        return self.settings(auth_name)

    @validate(schema=ChangeSettings(), form='settings', post_only=True)
    def change_settings(self):
        if h.authorized(h.is_admin) and self.form_result.get("username"):
            auth_name = self.form_result["username"]
        else:
            auth_name = request.environ.get('REMOTE_USER').lower()
        user_data = \
                Session.query(UserData).filter_by(user_name=auth_name).first()
        user_data.notify_on_success = \
                self.form_result.get('notify_on_success') or False
        user_data.notify_on_error = \
                self.form_result.get('notify_on_error') or False
        if auth_name == request.environ.get('REMOTE_USER').lower():
            session['lang'] = self.form_result['preflang']
            set_lang(session['lang'])
            session.save()
        user_data.language = self.form_result['preflang']
        user_data.e_mail = self.form_result['mailaddress']
        if auth_name != request.environ.get('REMOTE_USER').lower():
            # Admin is changing settings for another user
            new_privileges = self.form_result['privileges']
            user_set_single_role(auth_name, new_privileges)
            h.good_flash(_(u"Settings changed for user %s." % auth_name))
        else:
            h.good_flash(_(u"Settings changed."))
        Session.commit()
        return self.settings(auth_name)

    @checkAccess("admin")
    def change_another(self, environ, start_response):
        for k in request.params.keys():
            if k.startswith('submit'):
                username = k.split("_", 1)[1]
                # Settings action checks authorization
                return self.settings(username)

        for k in request.params.keys():
            if k.startswith('delete'):
                username = k.split("_", 1)[1]
                if (h.authorized(h.is_admin) and
                    request.environ['authkit.users'].user_exists(username) and
                    request.environ['REMOTE_USER'].lower() != username):
                    request.environ['authkit.users'].user_delete(username)
                    user_data = \
                            Session.query(UserData).filter_by(
                                    user_name=username).first()
                    Session.delete(user_data)
                    Session.commit()
                    h.good_flash(_(u"Account %s has been deleted" % username))
                    return self.manage()

    @checkAccess("admin")
    @validate(schema=CreateNewValidator(), form='manage', post_only=True)
    def new(self, environ, start_response):
        username = self.form_result['username'].strip()
        try:
            request.environ['authkit.users'].user_create(
                username, self.form_result['password'])
        except authkit.users.AuthKitError:
            # This should only happen if a user is created
            # after the validation happend.
            h.flash(_("User creation failed"))
            return self.manage()
        user_data = UserData(username.lower())
        Session.add(user_data)
        Session.commit()
        h.good_flash(_(u"Account %s created." % username))
        return self.settings(username)

    def settings(self, id = None):
        if not request.environ.get('REMOTE_USER'):
            abort(401)
            return

        if id != None:
            id = id.lower()
        if ( id != None and h.authorized(h.is_admin) ):
            auth_name = id
            if id != request.environ.get('REMOTE_USER').lower():
                # Changing_another bypasses setting restrictions
                # make sure to only make this available for Admins
                c.changing_another = True
        else:
            auth_name = request.environ.get('REMOTE_USER').lower()
        user_data = \
                Session.query(UserData).filter_by(user_name=auth_name).first()

        # Improvise because the select helper is too unflexible
        c.selected_admin = ""
        c.selected_privileged = ""
        c.selected_none = ""
        authkit_users = request.environ['authkit.users']

        if authkit_users.user_has_role(auth_name, "admin"):
            c.selected_admin = "selected"
        elif authkit_users.user_has_role(auth_name, "privileged"):
            c.selected_privileged = "selected"
        else:
            c.selected_none = "selected"

        if user_data == None:
            # Someone is trying to cheat!
            return "User does not exist!"
        c.user_data = user_data
        if user_data.language == "de":
            c.selected_de = "selected"
            c.selected_en = ""
        else:
            c.selected_de = ""
            c.selected_en = "selected"
        c.e_mail = user_data.e_mail
        c.notify_on_success = user_data.notify_on_success
        c.notify_on_error = user_data.notify_on_error
        return render('/account/settings.mako')
