import datetime
import logging
from os import path
import tarfile
import tempfile
import time

import glob

from paste.fileapp import FileApp

from pylons import config, request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect

from pylons.decorators import validate
from pylons.decorators.rest import restrict
from pylons.i18n.translation import _

import formencode
from sqlalchemy import desc, or_
from sqlalchemy import func, select
import webhelpers.paginate as paginate

from mapwocgui.lib.base import BaseController, render
from mapwocgui.model import Session
import mapwocgui.model.masterproxy as master
import mapwocgui.lib.helpers as h

from mapwoc.lib.result import Visit, Result, ResultType, ResultFile
from mapwoc.lib.scanconfiguration import ScanRun
from mapwoc.lib.url import URL
import mapwoc.lib.model.result
from mapwoc.lib.model.result import *
from mapwocgui.lib.forms import FilterForm
from mapwocgui.controllers.url import screendump, screendumps

log = logging.getLogger(__name__)

def sort_visits(visit_query):
    """
    Create the correct query to sort the visits.

    @param visit_query: sqlalchemy query for filtered visits.
    @returns: sqlalchemy query for visits with the correct sort order.
    """

    sorting = session.get("sort_visits")
    if sorting == None:
        # Default
        sorting = "date_desc"

    if sorting == "date_desc":
        visit_query = visit_query.order_by(desc(Visit.date))
    elif sorting == "date_asc":
        visit_query = visit_query.order_by(Visit.date)
    elif sorting == "hcsid_desc":
        visit_query = visit_query.order_by(desc(Visit.hcsid))
    elif sorting == "hcsid_asc":
        visit_query = visit_query.order_by(Visit.hcsid)
    elif "severity" in sorting:
        rt=Session.query(ResultType, ResultType.severity.label('type_sev')).subquery()
        visit_query = visit_query.outerjoin((rt,Visit.severest_result_type_id==rt.c.id))
        if sorting == "severity_desc":
            visit_query = visit_query.order_by(desc(rt.c.type_sev))
        else:
            visit_query = visit_query.order_by(rt.c.type_sev)
    elif sorting == "results_desc":
        visit_query = visit_query.order_by(desc(Visit.numresults))
    elif sorting == "results_asc":
        visit_query = visit_query.order_by(Visit.numresults)

    return visit_query

class ResultController(BaseController):

    def index(self):
        visits = sort_visits(Session.query(Visit))
        (_v,item_count) = Session.query(Visit, func.count()).one()
        c.visit_paginator = paginate.Page(visits,
                                          item_count=item_count,
                                          page=int(
                                               request.params.get('page', 1)),
                                          items_per_page=30,)
        c.resulttypes = Session.query(ResultType).all()
        c.dtnow = datetime.datetime.now()
        srs = Session.query(ScanRun)
        c.scanruns = srs.order_by(desc(ScanRun.starttime)).limit(20).all()
        if c.scanruns:
            c.dtfirst = min([scan.starttime for scan in c.scanruns])
        if not c.dtfirst:
            c.dtfirst = c.dtnow
        return render('/result/index_result.mako')

    @validate(schema=FilterForm(), form='index')
    def filter(self):
        # Initialize filter values.
        if not hasattr(self, 'form_result'):
            self.form_result = session['filter_form_result']
        searchmask = self.form_result.get('searchmask')
        types = self.form_result.get('resultType')
        usedate = self.form_result.get('datefilter')
        query = Session.query(Visit)
        scan_id = self.form_result.get('scanrun')
        date_to = self.form_result.get('dateto')
        date_from = self.form_result.get('datefrom')

        for k in self.form_result.keys():
            if k.startswith("sort_"):
                session["sort_visits"] = k[5:len(k)-2]
                session.save()
                break

        # Initialize filter settings.
        c.selected_scan_id = scan_id
        c.resulttypes = Session.query(ResultType).all()
        srs = Session.query(ScanRun)
        c.scanruns = srs.order_by(desc(ScanRun.starttime)).limit(20).all()
        c.dtnow = datetime.datetime.now()
        if c.scanruns:
            c.dtfirst = min([scan.starttime for scan in c.scanruns])
        if not c.dtfirst:
            c.dtfirst = c.dtnow

        if searchmask:
            query = query.filter(Visit.urls.any(
                URL.url.like("%%%s%%" % searchmask)))

        if isinstance(types, list):
            clauses = []
            for rtype in types:
                clauses.append(Visit.results.any(result_type_id=int(rtype)))
            query = query.filter(or_(*clauses))
        elif types and len(types):
            query = query.filter(Visit.results.any(result_type_id=int(types)))
        if scan_id and scan_id != "any":
            query = query.filter_by(scanrun_id = scan_id)

        if usedate:
            datefrom = time.strptime(date_from, "%d.%m.%Y")
            dateto = time.strptime(date_to, "%d.%m.%Y")
            query = query.filter(Visit.date.between(datefrom, dateto))
            return datefrom

        # Store the form to be used with other pages from paginator.
        session['filter_form_result'] = self.form_result

        visits = sort_visits(query)
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        html = render('/result/index_result.mako')
        return formencode.htmlfill.render(html, defaults = self.form_result)

    def by_type(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter(
                Visit.results.any(result_type_id = int(type)))
        visits = sort_visits(visits)
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        return render('/result/index_result.mako')

    def by_scanrun(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter_by(scanrun_id=id)
        visits = sort_visits(visits)
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        return render('/result/index_result.mako')

    def by_url(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter(Visit.urls.any(id=id))
        visits = sort_visits(visits)
        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)

        c.urlid = id

        # Get the list of hcs with screendumps.
        c.screendump_hcs = screendumps(id)

        return render('/result/index_result.mako')

    def visits_by_file(self, id):
        c.filtered_view = True
        visits = Session.query(Visit).filter(
                Visit.results.any(result_file_id = int(id)))
        visits = sort_visits(visits)

        c.visit_paginator = paginate.Page(visits,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        return render('/result/index_result.mako')

    def by_file(self, id):
        c.file_id = int(id)
        results = Session.query(Result).filter_by(result_file_id=int(id))
        #results = results_by(desc(Result.date))

        c.result_paginator = paginate.Page(results,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)

        first_result = results.first()
        sim_results = Session.query(Result).filter_by(url=first_result.url)
        c.sim_result_paginator = paginate.Page(sim_results,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        sha1_results = []#Session.query(Result).filter(Result.result_file.sha1sum=="abc")
        c.sha1sum_result_paginator = paginate.Page(sha1_results,
                                   page=int(request.params.get('page', 1)),
                                   items_per_page=30,)
        return render('/result/view_by_file.mako')

    def view_visit(self, id, dst="unset"):
        c.visit = Session.query(Visit).filter_by(id=id).first()
        if c.visit is None:
            abort(404)
        c.scanrun = c.visit.scan_run
        c.sc = c.scanrun.scanconfiguration
        c.types_found = []
        # TODO: Localize
        c.firstDate = c.visit.date.strftime(_("%d.%m.%Y - %H:%M:%S"))
        for result in c.visit.results:
            if not result.result_type in c.types_found:
                c.types_found.append(result.result_type)
        sorted_results = c.visit.results

        sortkey = dst
        if sortkey == "typ":
           sorted_results.sort(key=lambda result:result.result_type)
        elif sortkey == "url":
            sorted_results.sort(key=lambda result:result.url)
        elif sortkey == "sev":
            sorted_results.sort(key=lambda result:result.result_type.severity)
        elif sortkey == "dat":
            sorted_results.sort(key=lambda result:result.date)
        else:
            sorted_results.sort(key=lambda result:result.date)

        c.result_paginator = paginate.Page(sorted_results,
                                 page=int(request.params.get('page', 1)),
                                 items_per_page=30,)
        if len(c.types_found):
            c.highest_severity = max([type.severity for type in c.types_found])
        else:
            c.highest_severity = 0
        # TODO what if c.visit.hcsid is not known?
        c.hc = master.get_hc(c.visit.hcsid)

        if (len(c.visit.urls) > 0):
            c.screendump = screendump(c.visit.urls[0].id, c.visit.hcsid)
        else:
            c.screendump = [None]

        return render('/result/view_visit.mako')

    def send_dump(self, id):
        """ Sends the results dump """
        # TODO gzip
        visit = Session.query(Visit).filter_by(id=id).first()
        if visit is None:
            return "Visit id not found"
        if not visit.filepath:
            return "Visit has no saved files"
        tgz = tarfile.open(visit.filepath)
        dumpname = ""
        for name in tgz.getnames():
            if name.endswith(".dump"):
                dumpname = name
                break
        else:
            return "Visit has no dump"
        tfile = tgz.extractfile(dumpname)
        user_filename = "%s_%s.dump" % (visit.hcsid, id)

        response.headers = [('Content-Disposition', 'attachment; filename=\"' + \
                        user_filename + '\"'),
                   ('Content-Type', 'text/plain'),
                   ('Content-Length', str(tfile.size))]
        def file_streamer():
            while True:
                block = tfile.read(4096)
                if not block:
                    break
                yield block
        response.app_iter = file_streamer()
      #  return tmpfile.read()

    def send_tarball(self, id):
        """ Sends the result tarball. """
        visit = Session.query(Visit).filter_by(id=id).first()
        if visit is None:
            return "Visit id not found"

        if not visit.filepath:
            return "Visit has no saved files"
        # TODO If not os.path.isfile ...

        # TODO Middleware seems to notify client that stream is gzipped,
        #      Thus filesize is not correct.
        user_filename = "Result_%s.tar" % id
        file_size = path.getsize(visit.filepath)

        headers = [('Content-Disposition', 'attachment; filename=\"' + \
                        user_filename + '\"'),
                   ('Content-Type', 'text/plain'),
                   ('Content-Length', str(file_size))]
        fapp = FileApp(visit.filepath, headers=headers)
        return fapp(request.environ, self.start_response)

    def view_result(self, id):
        # Find out whether good screendump exists
        c.result = Session.query(Result).filter_by(id=id).first()
        if c.result is None:
            abort(404)
        c.visit =  c.result.visit
        # TODO What if c.visit.hcsid is not known?
        c.hc = master.get_hc(c.visit.hcsid)
        c.sc = c.visit.scan_run

        # Get the list of hcs with screendumps.
        c.screendump_hcs = screendumps(id)

        return render('/result/view_result.mako')

