import logging

import gzip
import StringIO

from pylons import request, response, session, tmpl_context as c, url
from pylons.controllers.util import abort, redirect
from pylons.i18n.translation import _, ungettext

import formencode

from pylons.decorators import validate
from pylons.decorators.rest import restrict
from sqlalchemy import func
import webhelpers.paginate as paginate

from mapwocgui.lib.base import BaseController, render
import mapwocgui.lib.helpers as h 

from mapwocgui.lib.auth import checkAccess
from mapwocgui.model import Session
from mapwoc.lib.url import URLList, URL, create_URL_List, separate_valid_urls
from mapwoc.lib.model.url import url_table, urllist_url, urllist_table, insert_list
from mapwocgui.lib.forms import NewURLListForm, UploadURLListForm

log = logging.getLogger(__name__)


class UrllistController(BaseController):

    def index(self):

        q = Session.query(URLList).filter(URLList.deleted==False)
        q = q.order_by(URLList.name)

        c.urllist_paginator = paginate.Page(q,
                                   page=int(request.params.get('page',1)),
                                   items_per_page=15,)
        return render('/urllist/index_urllist.mako')

    @checkAccess("privileged")
    def new(self):
        return render('/urllist/index_urllist.mako')

    def __handle_invalids(self, invalids):
        """ Sets a flash alert for invalid URLs. """

        if not invalids:
            return
        if len(invalids) < 20:
            h.flash(_(u"URL list contained invalid URLs:"))
            for invalid in invalids:
                h.flash(invalid)
        else:
            h.flash(_(u"More then 20 URL's in the list were invalid."))
            h.flash(_(u"Please check the created list."))

    @checkAccess("privileged")
    @restrict('POST')
    @validate(schema=UploadURLListForm(), form='new')
    def upload(self):
        """ Creates an URLList from a file. """

        filecontent = request.POST['URLListFile'].value

        # Handle gzipped content.
        sio = StringIO.StringIO(filecontent)
        try:
            gio = gzip.GzipFile("", 'rb', fileobj=sio)
            filecontent = gio.read()
            gio.close()
        except Exception as e:
            pass

        # TODO Limit upload size / length of uploaded url list.
        print "Length of uploaded file: %d " % len(filecontent)

        # TODO Count number of duplicates.
        valids,invalids = separate_valid_urls(filecontent)
        self.__handle_invalids(invalids)

        if len(valids) == 0:
            response.status_int = 302
            response.headers['location'] = h.url_for(controller='urllist',
                                                     action='index')
            h.flash(_("No valid URL in list."))
            return "Moved temporarily"

        # Create from request.
        urllist_name = self.form_result['URLListUploadName']
        urllist_comment = self.form_result['URLListUploadComment'] or ''
        # Make list unique, preserve order.
        seen = {}
        result = []
        for item in valids:
            if item in seen: continue
            seen[item] = 1
            result.append(item)

        # TODO detect and respect fail cases.
        uid = insert_list(Session, result, urllist_name, urllist_comment)

        Session.flush()
        Session.commit()

        h.good_flash(_("URL-List created."))
        response.status_int = 302
        response.headers['location'] = h.url_for(controller='urllist',
                                                 action='index')
        return "Moved temporarily"


    @checkAccess("privileged")
    @restrict('POST')
    @validate(schema=NewURLListForm(),form='new')
    def create(self):
        """Creates an URLList from a request."""

        # Create from request
        urllist = URLList(self.form_result['URLListName'])
        urllist.comment = self.form_result['URLListComment'] or ''

        urllist.numurls = 0
        valids,invalids = separate_valid_urls(self.form_result['URLsText'])
        self.__handle_invalids(invalids)

        if len(valids) == 0:
            response.status_int = 302
            response.headers['location'] = h.url_for(controller='urllist',
                                                     action='index',
                                                     id=urllist.id)
            h.flash(_("No valid URL in list."))
            return "Moved temporarily"

        # Avoid duplicate URLs in database.
        # TODO Note optimization potential here.
        for url_str in valids:
            url = Session.query(URL).filter_by(url=url_str).first()
            if not url:
                # URL not yet in database.
                url = URL(url_str)
                Session.add(url)
                Session.flush()
            urllist.urls.append(url)

        urllist.numurls = len(urllist.urls)
        # Add to store
        Session.add(urllist)
        Session.commit()

        # Redirect
        # TODO use redirect_to from pylons.controller.utils ?
        response.status_int = 302
        response.headers['location'] = h.url_for(controller='urllist', action='index', id=urllist.id)
        return "Moved temporarily" 

    def view(self, id):
        query = Session.query(URLList)
        c.urllist = query.get(int(id))

        # TODO Increase performance by query with LIMIT and proper pagination.
        # Use query like in Redirector to preserve order.
        urls = Session.execute("SELECT url, id FROM url, urllist_url "
                               "WHERE urllist_url.url_id = url.id"
                               " AND urllist_url.urllist_id = %d;" % int(id))
        urllist = []
        for url in urls:
            urllist.append((url[0], url[1]))

        c.url_paginator = paginate.Page(urllist,
                        page=int(request.params.get('page',1)),
                        items_per_page=100,)
        return render('/urllist/view_urllist.mako')

    #TODO check whether the url list is used in any scans.
    @checkAccess("privileged")
    def delete(self, id):
        if id is None:
            abort(404)
        urllist = Session.query(URLList).get(id)
        if urllist is None:
            abort(404)
        urllist.delete()
        Session.commit()
        response.status_int = 302
        response.headers['location'] = h.url_for(controller='urllist',action='index',id=urllist.id)
        return "Moved temporarily"        

