# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Forms (GUI).
#
# Authors:
# Andre Heinecke <andre.heinecke@intevation.de>
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011,2012 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import time

from pylons.i18n.translation import _, ungettext
from pylons import tmpl_context as c, request

import formencode

import mapwocgui.lib.helpers as h
from mapwocgui.lib.validators import QuickScanFormURLValidator
from mapwocgui.lib.validators import HCSelectionValidator
from mapwocgui.lib.validators import RessourceCheckValidator
from mapwocgui.lib.validators import PasswordValidator, UsernameValidator

# TODO While it would be nice to validate this, the master will complain
#      if a scan is running, anyways. It brings some moral complications to
#      import the masterproxy here. So following function is broken (and
#      uncommented where it should be called).
def check_running_scan(self, values, c):
    c.current_scan_name = master.get_state_set().get('scanname')
    if c.current_scan_name:
        h.flash(
            _(u"Quickscan can not be started while %s is still running.") % \
                c.current_scan_name)
        raise formencode.validators.Invalid(self.message("no_message", c),
               values, c)

# TODO Refactor shared convenienve methods.

class MapwocSchema(formencode.Schema):
    def create_invalid(self, msg, values, c):
        return formencode.validators.Invalid(self.message(msg, c),values, c)


# Scan Forms


class QuickScanForm(formencode.Schema):
    """Schema for the form to create and start the QuickScan."""

    messages = {'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False
    quickurl = formencode.All(formencode.validators.String(not_empty=True),
                              QuickScanFormURLValidator())

    # TODO Validate VM.
    def validate_python(self, values, c):
        # TODO see above for check_running_scan
        #check_running_scan(self, values, c)
        pass

class ManualScanForm(formencode.Schema):
    """Schema for the form to create and start the ManualScan."""

    messages = {'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False

    # TODO Validate VM.
    def validate_python(self, values, c):
        # TODO see above for check_running_scan
        #check_running_scan(self, values, c)
        pass

class TimedScanForm(MapwocSchema):
    """Schema for the form to create and start the TimedScan."""

    messages = {'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False
    time_frame = formencode.validators.Int(not_empty=True, min=1, max=3600)

    def validate_python(self, values, c):
        """Check that there is mimimum one HC selected."""
        hckeys = [key for key in values.keys() if key.startswith('hc') and values[key] == '1']
        if len(hckeys) < 1:
            h.flash(_(u"At least one HoneyClient should be selected for timed scan!"))
            raise self.create_invalid("no_message", values, c)

class ScanForm(MapwocSchema):
    """Schema for the form to create a Scan."""

    chained_validators = [HCSelectionValidator, RessourceCheckValidator]
    allow_extra_fields = True
    filter_extra_fields = False
    name = formencode.All(formencode.validators.String(not_empty=True))
    comment = formencode.validators.String()
    active_window = formencode.validators.Int(not_empty=True, min=1, max=600)
    urllist = formencode.validators.NotEmpty()
    messages = {'no_message' : " "}

    def validate_python(self, values, c):
        reserved_names = ["QuickScan", "ManualScan"]
        if values["name"] in reserved_names:
            h.flash(_(u"Names ManualScan and QuickScan is reserved"))
            raise self.create_invalid("no_message", values, c)

# User Management Forms


class ChangeSettings(MapwocSchema):
    """ Schema and valadiation for a setting change request."""
    messages = {'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False
    mailaddress = formencode.validators.Email(not_empty=False)

    def validate_python(self, values, c):
        if ( values.get("username") != request.environ.get('REMOTE_USER').lower() and
                h.authorized(h.is_admin) ):
            if values.get('privileges') and not\
                    values['privileges'] in ['admin', 'privileged', 'none']:
                h.flash(u"Invalid data in privileges request")
                raise self.create_invalid("no_message", values, c)
            if values.get('username') and \
                not request.environ['authkit.users'].user_exists(
                    values["username"]):
                h.flash(u"User does not exist")
                raise self.create_invalid("no_message", values, c)

class ChangePassword(MapwocSchema):
    """ Schema and valadiation for a password change request."""
    messages = {'no_message' : " "}
    chained_validators = [PasswordValidator]
    allow_extra_fields = True
    filter_extra_fields = False

    def validate_python(self, values, c):
        if (values.get("username") and
             values["username"] != request.environ.get('REMOTE_USER').lower()\
             and h.authorized(h.is_admin) ):
            # Admin may change the password for all other users
            return True
        from authkit.authenticate import valid_password
        if not valid_password(request.environ,
                              request.environ.get('REMOTE_USER'),
                              values['current']):
            h.flash(_(u"The current password is invalid."))
            raise self.create_invalid("no_message", values, c)

class CreateNewValidator(formencode.Schema):
    """ Schema and valadiation for a new user creation."""
    messages = {'no_message' : " "}
    chained_validators = [PasswordValidator, UsernameValidator]
    allow_extra_fields = True
    filter_extra_fields = False


# URL-List Forms


class NewURLListForm(formencode.Schema):
    """URLList Validation, dummy."""

    # Extra field is submit button, right now
    allow_extra_fields = True
    filter_extra_fields = True
    URLListName = formencode.validators.String(not_empty=True)
    URLListComment = formencode.validators.String()
    URLsText = formencode.validators.String(not_empty=True)

class UploadURLListForm(formencode.Schema):
    """URLList upload validation."""

    # Extra field is submit button, right now
    allow_extra_fields = True
    filter_extra_fields = True
    URLListUploadName = formencode.validators.String(not_empty=True)
    URLListUploadComment= formencode.validators.String()
    URLListFile = formencode.validators.String(not_empty=True)


# Filter Form


class FilterForm(MapwocSchema):

    messages = {'badFrom' : _(u"Invalid date format in field from:"),
                'badTo' :   _(u"Invalid date format in field to:"),
                'no_message' : " "}
    allow_extra_fields = True
    filter_extra_fields = False
    searchmask = formencode.validators.String()

    def validate_python(self, values, c):
        if not values.get("datefilter"):
            return
        try:
            date_from = time.strptime(values['datefrom'], "%d.%m.%Y")
        except:
            h.flash(self.message("badFrom", c))
            raise formencode.validators.Invalid(
                      self.message("no_message", c), values, c)
        try:
            date_to = time.strptime(values['dateto'], "%d.%m.%Y")
        except:
            h.flash(self.message("badTo", c))
            raise create_invalid("no_message", values, c)
