# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Validators for Forms (GUI).
#
# Authors:
# Andre Heinecke <andre.heinecke@intevation.de>
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from codecs import getdecoder

from pylons.i18n.translation import _, ungettext
from pylons import tmpl_context as c, request

import formencode

from mapwoc.lib.url import URLList, URL, create_URL_List
from mapwoc.lib.scanconfiguration import TransportableSC

import mapwocgui.lib.helpers as h
from mapwocgui.model.meta import Session, Base, metadata
import mapwocgui.model.masterproxy as master
from mapwocgui.lib.form_transformers import scanconfig_from_form

# TODO consolidate, refactor with lib/forms
class MapwocValidator(formencode.validators.FancyValidator):
    def create_invalid(self, msg, values, c):
        return formencode.validators.Invalid(self.message(msg, c),values, c)

class HCSelectionValidator(MapwocValidator):
    """ Validates the HoneyClient list. """

    messages = {'empty' : _(u"You must at least select one HoneyClient."),
                'negativeNumber' : _(u"Cannot start a negative number of HoneyClients."),
                'noNode' : _(u"One ore more HoneyClients do"
                              " not appear to be on any Node."),
                'resourceCheckFail' : _(u"This configuration did not pass"
                              " the ressource check on one or more nodes."),
                'not_int' : _("Number for HC has to be an integer."),
                'no_message' : " "}

    def validate_python(self, values, c):
        atleastone = False
        visitors = {}

        hckeys = [key for key in values.keys() if key.startswith('hc')]
        for key in hckeys:
            try:
                if int(values[key]) < 0:
                    h.flash(self.message('negativeNumber' ,c ))
                    raise self.create_invalid("no_message", values, c)
                if int(values[key]):
                    atleastone = True
                if len(key.split(",")) < 2:
                    h.flash(self.message("noNode", c))
                    raise self.create_invalid("no_message", values, c)
            except ValueError:
                h.flash(self.message("not_int", c))
                raise self.create_invalid("no_message", values, c)

        if not atleastone:
            # Workaround to make the error message look consistent
            h.flash(self.message("empty", c))
            raise self.create_invalid("no_message", values, c)

class RessourceCheckValidator(MapwocValidator):
    """
    Ask the master to asks its node whether they have the ressources to start
    this scan.
    """

    messages = {
                'no_message' : " ",
                'not_enough_ressources' : _(u"Ressource Check Failed:"
                     " One or more Nodes do not have sufficient ressources."),
                'not_enough_ram' : _(u"Ressource Check Failed:"
                     " One or more Nodes do not have sufficient memory."),
                'not_enough_proxy_ports': _(u"Ressource Check Failed:"
                     " One or more Nodes do not have sufficient proxy ports."),
                'not_enough_vnc_ports': _(u"Ressource Check Failed:"
                     " VNC can only be enabled for less then 100 clients per node."),
                'not_max_parallel': _(u"Check Failed:"
                     " Violation of max_parallel constraint."),
               }

    def validate_python(self, values, c):
        """
        @param values: Form values.
        @param c: The context.
        """
        sc = scanconfig_from_form(values)

        # The URL-Lists from this scanconfig are created from a form
        # and not read from database, thus not connected to db-Session.
        # Establish this connection now. 
        Session.add(sc)
        dburllists = []
        for ul in sc.urlcollection:
            dbul = Session.query(URLList).filter_by(name=ul.name).first()
            dburllists.append(dbul)
        sc.urlcollection = dburllists

        if sc.enable_vnc:
            needed_displays = {}
            for visitor in sc.visitor_configuration.visitors:
                needed_displays[visitor.node] = needed_displays.get(visitor.node, 0) + \
                                                visitor.number
            for number in needed_displays.values():
                if number > 99:
                    h.flash(self.message('not_enough_vnc_ports', c))
                    raise self.create_invalid("no_message", values, c)

        tr = master.check_scan(TransportableSC(sc))
        Session.rollback()
        if not tr.error:
            pass
        else:
            # TODO Choose message epending on error code.
            if tr.error == 101:
                # TODO parse the ram message, if successfull. localize it.
                h.flash(self.message('not_enough_ram', c) + tr.message)
            elif tr.error == 102:
                h.flash(self.message('not_enough_proxy_ports', c) + tr.message)
            elif tr.error == 103:
                h.flash(self.message('not_max_parallel', c) + tr.message)
            else:
                h.flash(self.message('not_enough_ressources', c) + tr.message)
            raise self.create_invalid("no_message", values, c)


class QuickScanFormURLValidator(MapwocValidator):
    """Validates the URL entered in the QuickScan Form."""

    messages = {'no_message' : " "}

    def validate_python(self, values, c):
        urllist,invalids = create_URL_List("tmp",
                                    values, None)
        for invalid in invalids:
            h.flash(_(u"Invalid URL: %s") % invalid)
        if invalids:
            raise self.create_invalid("no_message", values, c)
        return values

class PasswordValidator(MapwocValidator):
    messages = {'no_message' : " "}

    def validate_python(self, values, c):

        if not values['password'] == values['password2']:
            h.flash(_(u"New passwords do not match."))
            raise self.create_invalid("no_message", values, c)

        if not len(values['password']) > 5 :
            h.flash(_(u"The password needs to have at least 6 Characters."))
            raise self.create_invalid("no_message", values, c)

        try:
            getdecoder("ascii")(values['password'])
        except (UnicodeEncodeError, UnicodeDecodeError):
            h.flash(_(u"Only ASCII characters are allowed in your Password"))
            raise self.create_invalid("no_message", values, c)

class UsernameValidator(MapwocValidator):
    messages = {'no_message' : " "}

    def validate_python(self, values, c):
        username = values['username'].strip()
        if not len(username) > 1:
            h.flash(_(u"A username must have at least 1 Character."))
            raise self.create_invalid("no_message", values, c)

        if len(username.split()) != 1:
            h.flash(_(u"A username may not contain whitespace characters."))
            raise self.create_invalid("no_message", values, c)

        try:
            getdecoder("ascii")(username)
        except (UnicodeEncodeError, UnicodeDecodeError):
            h.flash(_(u"Only ASCII characters are allowed in your Password"))
            raise self.create_invalid("no_message", values, c)

        if request.environ['authkit.users'].user_exists(username):
            h.flash(_(u"A user with that name already exists."))
            raise self.create_invalid("no_message", values, c)

