# -*- coding: utf-8 -*-

# mapWOC
#
# Description: Convenience to query a mapwoc-master.
#
# Authors:
# Felix Wolfsteller <felix.wolfsteller@intevation.de>
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import socket

import gnutls.crypto
from gnutls.crypto import *
from gnutls.connection import *

from pylons import config

import mapwoc.lib.scanconfiguration
from mapwoc.lib.gnutls_util import load_cred, recvall
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook as json_decode
from mapwoc.lib.trivialresponse import TrivialResponse

"""
Global credentials to be used within this module.
"""
gnutls_credentials = None

def __load_credentials():
    """
    Load credentials, if specified in config file.
    """
    global gnutls_credentials

    if gnutls_credentials:
        return

    sslconfig = config.get('master.sslconfig')
    if sslconfig:
        gnutls_credentials = load_cred(sslconfig)

def __ask_for_string_response(question):
    """
    Connect to masterctrlserver, send a string and
    receive and return max 1024.
    """

    global gnutls_credentials

    HOST, PORT = config['master.host'], int(config['master.port'])

    try:
        # Create a socket (SOCK_STREAM means a TCP socket)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        __load_credentials()
        if gnutls_credentials:
            session = ClientSession(sock, gnutls_credentials)
            session.connect((HOST, PORT))
            session.handshake()
            session.sendall(question + "\r\n")
            received = recvall(session)
            try:
                session.bye()
                session.close()
            except:
                # Session likely invalidated by recvall call.
                pass
        else:
            sock.connect((HOST, PORT))

            # Connect to server and send data
            sock.sendall(question + "\r\n")

            # TODO handle all the odds of sockets, timeouts, exceptions, etc
            # TODO lessen assumption about amount of data transmitted

            # Receive data from the server and shut down
            received = sock.recv(4*40480, socket.MSG_WAITALL)
            sock.close()

    except socket.error as msg:
        print "Exception: "
        print type(msg)
        print msg
        received = "Connection Error: " + str(msg)

    return received
    
def __ask_for_trivial_response(question):
    """Invokes a command that returns a trivial response."""

    string_answer = __ask_for_string_response(question)
    trivial_response = json.loads(string_answer, object_hook=json_decode)
    if isinstance(trivial_response, TrivialResponse):
        return trivial_response
    else:
        return None

def get_state_set():
    """Connects to MasterCtrl and queries its state."""
    
    answer = __ask_for_string_response("state")
    return json.loads(answer, object_hook=json_decode)

def get_n_messages(num):
    """Connects to MasterCtrl and queries for a list of messages."""
    answer = __ask_for_string_response("messages %d\n"%num)
    # Might raise a ValueError (no JSON object could be decoded)
    return json.loads(answer, object_hook=json_decode)

def get_messages():
    """Connects to MasterCtrl and queries for a list of messages."""
    answer = __ask_for_string_response("messages\n")
    # Might raise a ValueError (no JSON object could be decoded)
    return json.loads(answer, object_hook=json_decode)

def get_node(nodename):
    """
    Connects to MasterCtrl and queries for a node by it's name
    @return A Node object
    """
    answer = __ask_for_string_response("node " + nodename)
    return json.loads(answer, object_hook=json_decode)

def get_hcs():
    """
    Connects to MasterCtrl and queries for a list of hcsids, then
    subsequently ask MasterCtrl for HC instances.
    @return: A list of HCs.
    """
    # TODO: Formulate *one* question for this, instead of one querstion
    #       per hc.
    answer =  __ask_for_string_response("hcs")
    hcs = []
    for hcsid in json.loads(answer).keys():
        hcs.append(get_hc(hcsid))
    return hcs

def get_node_states():
    """Connects to MasterCtrl and queries for a list of states of nodes."""
    answer = __ask_for_string_response("nodestates")
    return json.loads(answer)

def get_hcs_of(node):
    """Connects to MasterCtrl and queries for a list of HCs of a node."""
    return __ask_for_string_response("hcs " + node)

def get_node_state(node):
    """Connects to MasterCtrl and queries for a state of a node."""
    return __ask_for_string_response("nodestate " + node)

def get_hc(hcsid):
    """Connects to MasterCtrl and queries for details of an hc."""
    answer = __ask_for_string_response("hc " + hcsid)
    return json.loads(answer, object_hook=json_decode)

def get_node_confs():
    """Connects to MasterCtrl and queries for config file for all nodes."""
    return __ask_for_string_response("nodeconfs")

def get_node_conf(hcsid):
    """Connects to MasterCtrl and queries for config file for a node."""
    return __ask_for_string_response("nodeconf " + hcsid)

def start_hc(hcsid, node):
    """Connects to MasterCtrl and asks to start a given hc."""
    return __ask_for_string_response('nodectrl ' + node + ' starthc ' + hcsid)

def stop_hc(hcsid, node):
    """Connects to MasterCtrl and asks to stop a given hc."""
    return __ask_for_string_response('nodectrl ' + node + ' stophc ' + hcsid)

def get_node_runlist(node):
    """Connects to MasterCtrl and asks to for list of running hcs."""
    return __ask_for_string_response('nodectrl ' + node + ' runlist')

def start_scan(scanconf):
    """Connects to MasterCtrl and asks to start a scan."""
    return __ask_for_trivial_response("startscan %d\r\n" % scanconf.id)

def check_scan(scanconf):
    """Connects to MasterCtrl and asks whether scan is ok."""
    json_question = MapWOCJSONEncoder().encode(scanconf)
    return __ask_for_trivial_response("checkscan\r\n" + json_question + "\r\n")

def stop_smooth():
    """Connects to MasterCtrl and asks to stop scan."""
    return __ask_for_trivial_response("stopsmooth")

def stop_scan():
    """Connects to MasterCtrl and asks to stop scan."""
    return __ask_for_trivial_response("stopscan")

def get_version():
    """Connects to MasterCtrl and asks for its version."""
    return __ask_for_string_response("version")
